@@ -64,8 +64,10 @@ char *s;
 int phork()
 {				/**/
     int pid;
+
 #ifdef RLIM_INFINITY
     int t0;
+
 #endif
 
     if (thisjob >= MAXJOB - 1) {
@@ -205,7 +207,7 @@ int dash;
     }
     argv = makecline(args);
     unblockchld();
-    if ((int) strlen(arg0) > MAXPATHLEN) {
+    if ((int)strlen(arg0) > MAXPATHLEN) {
 	zerr("command too long: %s", arg0, 0);
 	_exit(1);
     }
@@ -223,24 +225,27 @@ int dash;
     if (cn && ISEXCMD(cn->flags)) {
 	char nn[MAXPATHLEN];
 
-	for (pp = path; pp < cn->u.name; pp++)
-	    if (**pp == '.' && (*pp)[1] == '\0') {
-		ee = zexecve(arg0, argv);
-		if (ee != ENOENT)
-		    eno = ee;
-	    } else if (**pp != '/') {
-		z = buf;
-		strucpy(&z, *pp);
-		*z++ = '/';
-		strcpy(z, arg0);
-		ee = zexecve(buf, argv);
-		if (ee != ENOENT)
-		    eno = ee;
-	    }
-	strcpy(nn, cn->u.name ? *(cn->u.name) : "");
-	strcat(nn, "/");
-	strcat(nn, cn->nam);
-
+	if (cn->flags & BUILTIN)
+	    strcpy(nn, cn->u.cmd);
+	else {
+	    for (pp = path; pp < cn->u.name; pp++)
+		if (**pp == '.' && (*pp)[1] == '\0') {
+		    ee = zexecve(arg0, argv);
+		    if (ee != ENOENT)
+			eno = ee;
+		} else if (**pp != '/') {
+		    z = buf;
+		    strucpy(&z, *pp);
+		    *z++ = '/';
+		    strcpy(z, arg0);
+		    ee = zexecve(buf, argv);
+		    if (ee != ENOENT)
+			eno = ee;
+		}
+	    strcpy(nn, cn->u.name ? *(cn->u.name) : "");
+	    strcat(nn, "/");
+	    strcat(nn, cn->nam);
+	}
 	ee = zexecve(nn, argv);
 
 	if (ee != ENOENT)
@@ -283,7 +288,7 @@ char *arg0;
 	cn = hashcmd(arg0, path);
     if (cn && cn->flags & DISABLED)
 	cn = NULL;
-    if ((int) strlen(arg0) > MAXPATHLEN)
+    if ((int)strlen(arg0) > MAXPATHLEN)
 	return NULL;
     for (s = arg0; *s; s++)
 	if (*s == '/') {
@@ -296,19 +301,21 @@ char *arg0;
     if (cn && ISEXCMD(cn->flags)) {
 	char nn[MAXPATHLEN];
 
-	for (pp = path; pp < cn->u.name; pp++)
-	    if (**pp != '/') {
-		z = buf;
-		strucpy(&z, *pp);
-		*z++ = '/';
-		strcpy(z, arg0);
-		try(buf);
-	    }
-
-	strcpy(nn, cn->u.name ? *(cn->u.name) : "");
-	strcat(nn, "/");
-	strcat(nn, cn->nam);
-
+	if (cn->flags & BUILTIN)
+	    strcpy(nn, cn->u.cmd);
+	else {
+	    for (pp = path; pp < cn->u.name; pp++)
+		if (**pp != '/') {
+		    z = buf;
+		    strucpy(&z, *pp);
+		    *z++ = '/';
+		    strcpy(z, arg0);
+		    try(buf);
+		}
+	    strcpy(nn, cn->u.name ? *(cn->u.name) : "");
+	    strcat(nn, "/");
+	    strcat(nn, cn->nam);
+	}
 	try(nn);
     }
     for (pp = path; *pp; pp++) {
@@ -404,10 +411,13 @@ void fullhash()
 void execlist(list)		/**/
 List list;
 {
+    static int donetrap;
+
     if (breaks)
 	return;
     if (!list || list == &dummy_list)
 	return;
+    donetrap = 0;
     simplifyright(list);
     switch (list->type) {
     case SYNC:
@@ -415,12 +425,16 @@ List list;
 	execlist2(list->left, list->type, !list->right);
 	if (sigtrapped[SIGDEBUG])
 	    dotrap(SIGDEBUG);
-	if (sigtrapped[SIGZERR] && lastval)
-	    dotrap(SIGZERR);
-	if (lastval && isset(ERREXIT) && sourcelevel < 32768) {
-	    if (sigtrapped[SIGEXIT])
-		dotrap(SIGEXIT);
-	    exit(lastval);
+	if (sourcelevel < 32768 && !donetrap) {
+	    if (sigtrapped[SIGZERR] && lastval) {
+		dotrap(SIGZERR);
+		donetrap = 1;
+	    }
+	    if (lastval && isset(ERREXIT)) {
+		if (sigtrapped[SIGEXIT])
+		    dotrap(SIGEXIT);
+		exit(lastval);
+	    }
 	}
 	if (list->right && !retflag) {
 	/* errflag = 0; */
@@ -478,8 +492,8 @@ int last1;
     ipipe[0] = ipipe[1] = opipe[0] = opipe[1] = 0;
     if (in_pipe) {
 	execpline2(l->left, how, opipe[0], ipipe[1], last1);
-	return lastval;	/* if we're in a pipe don't change lastval; it'll be */
-    }			/* done by the first instance of execpline */
+	return lastval;		/* if we're in a pipe don't change lastval; it'll be */
+    }				/* done by the first instance of execpline */
     blockchld();
     if ((thisjob = getfreejob()) == -1)
 	return 1;
@@ -524,17 +538,20 @@ int last1;
 {
     int pid;
     int pipes[2];
+    int oldlineno = lineno;
 
     if (breaks)
 	return;
     if (!pline)
 	return;
+    lineno = pline->left->lineno;
     if (pline->type == END) {
 	in_pipe = 0;
 	execcmd(pline->left, input, output, how == ASYNC, last1);
 	pline->left = NULL;
     } else {
 	int old_pipe = in_pipe;
+
 	in_pipe = 0;
 	mpipe(pipes);
 	if (pline->left->type >= CURSH && how == SYNC) {
@@ -581,6 +598,8 @@ int last1;
 	    close(pipes[0]);
 	}
     }
+
+    lineno = oldlineno;
 }
 
 /* make the argv array */
@@ -774,7 +793,8 @@ int export;
 		val = ztrdup(ugetnode(vl));
 	    }
 	    pm = setsparam(v->name, ztrdup(val));
-	    if (errflag) return;
+	    if (errflag)
+		return;
 	    if (export && !(pm->flags & PMFLAG_x))
 		addenv(v->name, val);
 	    zsfree(val);
@@ -793,7 +813,8 @@ int export;
 	}
 	*ptr = NULL;
 	setaparam(v->name, arr);
-	if (errflag) return;
+	if (errflag)
+	    return;
     }
 }
 
@@ -860,24 +881,24 @@ int last1;
 
 	if (!strcmp(cmdarg, "builtin") && args->first->next &&
 	    (cn2 = (Cmdnam) gethnode(cmdarg, cmdnamtab)) &&
-	    (cn2->flags & BUILTIN)) {
+	    (cn2->flags & BUILTIN) &&
+	    !(cn2->flags & EXCMD)) {
 	    builtin = 1;
 	    cmdarg = (char *)args->first->next->dat;
 	}
 	cn1 = (Cmdnam) gethnode(cmdarg, cmdnamtab);
-	if (cn1 && (cn1->flags & BUILTIN))
+	if (cn1 && (cn1->flags & BUILTIN) && !(cn1->flags & EXCMD))
 	    assign = istypeset(cn1, NULL);
 	else if (builtin)
 	    assign = istypeset(NULL, cmdarg);
 	if (builtin)
 	    cn1 = cn2;
     }
-
-    /* do prefork substitutions */
+ /* do prefork substitutions */
     prefork(args, (((type == CCASE) ? 010 : 0)
 		   | (assign ? 02 : isset(MAGICEQUALSUBST))));
 
-    /* warn about "rm *" */
+ /* warn about "rm *" */
     if (unset(RMSTARSILENT) && interact && isset(SHINSTDIN) &&
 	type == SIMPLE && full(args) && nextnode(firstnode(args)) &&
 	!strcmp(peekfirst(args), "rm") &&
@@ -934,7 +955,8 @@ int last1;
 /* this is nonzero if cmd is a current shell procedure */
 
     iscursh = (type >= CURSH) || (type == SIMPLE && cn1 &&
-				  (cn1->flags & (BUILTIN | SHFUNC)));
+				  (cn1->flags & (BUILTIN | SHFUNC)) &&
+				  !(cn1->flags & EXCMD));
 
 /* if this command is backgrounded or (this is an external
 		command and we are not exec'ing it) or this is a builtin
@@ -943,9 +965,11 @@ int last1;
 		the rest of the function think we forked. */
 
     if (bkg || !(iscursh || (cmd->flags & CFLAG_EXEC)) ||
-	(cn1 && (cn1->flags & (BUILTIN | SHFUNC)) && output)) {
+	(cn1 && (cn1->flags & (BUILTIN | SHFUNC)) &&
+	 !(cn1->flags & EXCMD) && output)) {
 	int synch[2];
 
+	blockchld();
 	pipe(synch);
 	pid = (last1 && execok())? 0 : phork();
 	if (pid == -1) {
@@ -964,8 +988,8 @@ int last1;
 		    lastpid = pid;
 		else if (!jobtab[thisjob].stty_in_env && full(cmd->vars))
 		    while (full(cmd->vars))	/* search for STTY=... */
-			if (!strcmp(((Varasg)ugetnode(cmd->vars))->name,
-				    "STTY")) { 
+			if (!strcmp(((Varasg) ugetnode(cmd->vars))->name,
+				    "STTY")) {
 			    jobtab[thisjob].stty_in_env = 1;
 			    break;
 			}
@@ -977,7 +1001,12 @@ int last1;
 	entersubsh(bkg);
 	close(synch[1]);
 	forked = 1;
-    }
+
+	if (sigtrapped[SIGINT] == 2)
+	    holdintr();
+    } else if ((cmd->flags & CFLAG_EXEC) && !nullexec)
+	entersubsh(bkg);
+
     if (bkg && isset(BGNICE))
 	nice(5);
 
@@ -1116,13 +1145,17 @@ int last1;
 		lastval = execbin(args, cn1);
 		if (isset(PRINTEXITVALUE) && isset(SHINSTDIN) &&
 		    lastval && !subsh) {
-		    fprintf(stderr, "zsh: exit %ld\n", (long) lastval);
+		    fprintf(stderr, "zsh: exit %ld\n", (long)lastval);
 		}
 		fflush(stdout);
-		if (ferror(stdout)) {
-		    zerr("write error: %e", NULL, errno);
+		if (save[1] == -1) {
+		    if (ferror(stdout)) {
+			zerr("write error: %e", NULL, errno);
+			clearerr(stdout);
+			errflag = 0;
+		    }
+		} else
 		    clearerr(stdout);
-		}
 	    }
 	} else {
 	    if (cmd->flags & CFLAG_EXEC)
@@ -1205,6 +1238,7 @@ int bkg;
     if (sigtrapped[SIGQUIT] != 2)
 	sig_default(SIGQUIT);
     opts[MONITOR] = OPT_UNSET;
+    opts[USEZLE] = OPT_UNSET;
     clearjobtab();
 }
 
@@ -1436,7 +1470,7 @@ int qt;
 	popheap();
 	close(pipes[1]);
 	retval = readoutput(pipes[0], qt);
-	chldsuspend();		/* unblocks */
+	chldsuspend(0);		/* unblocks */
 	lastval = cmdoutval;
 	cmdoutval = 0;
 	return retval;
@@ -1486,6 +1520,8 @@ int qt;
 	if (!qt && isep(c)) {
 	    if (cnt) {
 		*ptr = '\0';
+		if (isset(GLOBSUBST))
+		    tokenize(buf);
 		addnode(ret, buf);
 		ptr = buf = (char *)ncalloc(bsiz = 64);
 		cnt = 0;
@@ -1503,6 +1539,8 @@ int qt;
 	ptr[-1] = '\0';
     else
 	*ptr = '\0';
+    if (!qt && isset(GLOBSUBST))
+	tokenize(buf);
     if (cnt)
 	addnode(ret, buf);
     fclose(fin);
@@ -1532,12 +1570,17 @@ char *cmd;
     addnode(jobtab[thisjob].filelist, ztrdup(nam));
     heapalloc();
     blockchld();
-    if ((pid = phork())) {
+    if ((pid = cmdoutpid = phork())) {
 	popheap();
 	if (pid < 0)
 	    unblockchld();
-	else
-	    chldsuspend();
+	else {
+	    int os = jobtab[thisjob].stat;
+
+	    waitforpid(pid);
+	    cmdoutval = 0;
+	    jobtab[thisjob].stat = os;
+	}
 	return nam;
     }
     subsh = 1;
@@ -1759,8 +1802,8 @@ char *zgetwd()
     struct stat sbuf;
     struct dirent *de;
     DIR *dir;
-    ino_t ino, rootino = (ino_t) ~0;
-    dev_t dev, rootdev = (dev_t) ~0;
+    ino_t ino, rootino = (ino_t) ~ 0;
+    dev_t dev, rootdev = (dev_t) ~ 0;
 
     holdintr();
     buf2[0] = '\0';
@@ -1806,7 +1849,8 @@ char *zgetwd()
 		if (sbuf.st_dev == dev)
 		    goto match;
 	    }
-	rewinddir(dir);
+	closedir(dir);
+	dir = opendir(".");
 	readdir(dir);
 	readdir(dir);
 	while ((de = readdir(dir))) {
@@ -1975,8 +2019,8 @@ int flags;
 	opts[XTRACE] = OPT_SET;
     opts[PRINTEXITVALUE] = OPT_UNSET;
     if (doshargs) {
-	pparams = x = 
-	  (char **)zcalloc(((sizeof *x) * (1 + countnodes(doshargs))));
+	pparams = x =
+	    (char **)zcalloc(((sizeof *x) * (1 + countnodes(doshargs))));
 	argzero = ztrdup(ugetnode(doshargs));
 	while ((*x = (char *)ugetnode(doshargs)))
 	    *x = ztrdup(*x), x++;
@@ -1999,6 +2043,7 @@ int flags;
 	zsfree(s);
     }
     zfree(locallist, sizeof(struct lklist));
+
     locallist = olist;
     breaks = retflag = 0;
     freearray(pparams);
@@ -2062,7 +2107,7 @@ char *cancd(s)			/**/
 char *s;
 {
     int nocdpath = s[0] == '.' &&
-	(s[1] == '/' || !s[1] || (s[1] == '.' && (s[2] == '/' || !s[1])));
+    (s[1] == '/' || !s[1] || (s[1] == '.' && (s[2] == '/' || !s[1])));
     char *t;
 
     if (*s != '/') {
@@ -2077,17 +2122,14 @@ char *s;
 		sprintf(sbuf, "%s/%s", *cp, s);
 		if (cancd2(sbuf)) {
 		    doprintdir = -1;
-		    zsfree(s);
 		    return dupstring(sbuf);
 		}
 	    }
 	if ((t = cd_able_vars(s))) {
 	    if (cancd2(t)) {
 		doprintdir = -1;
-		zsfree(s);
 		return t;
 	    }
-	    zsfree(t);
 	}
 	return NULL;
     }
