@@ -33,8 +33,10 @@ static int matchct;		/* number of matches found */
 static char pathbuf[MAXPATHLEN];/* pathname buffer */
 static char **matchbuf;		/* array of matches */
 static char **matchptr;		/* &matchbuf[matchct] */
-static Comp exclude;		/* pattern to exclude */
 static char *colonmod;		/* colon modifiers in qualifier list */
+static ino_t old_ino;
+static dev_t old_dev;
+static int old_pos;
 
 #ifdef ULTRIX
 typedef struct stat *Statptr;	/* This makes the Ultrix compiler happy.  Go figure. */
@@ -75,11 +77,23 @@ struct complist {
     int closure;		/* 1 if this is a (foo/)# */
 };
 struct comp {
-    Comp left, right, next;
+    Comp left, right, next, exclude;
     char *str;
-    int closure, last;
+    int stat;
 };
 
+#define C_ONEHASH	1
+#define C_TWOHASH	2
+#define C_CLOSURE	(C_ONEHASH|C_TWOHASH)
+#define C_LAST		4
+#define C_PATHADD	8
+
+#define CLOSUREP(c)	(c->stat & C_CLOSURE)
+#define ONEHASHP(c)	(c->stat & C_ONEHASH)
+#define TWOHASHP(c)	(c->stat & C_TWOHASH)
+#define LASTP(c)	(c->stat & C_LAST)
+#define PATHADDP(c)	(c->stat & C_PATHADD)
+
 void glob(list, np)		/**/
 Lklist list;
 Lknode *np;
@@ -120,7 +134,8 @@ Lknode *np;
 #endif
 
 	for (s = str + sl - 2; s != str; s--)
-	    if (*s == Bar || *s == Outpar || *s == Inpar)
+	    if (*s == Bar || *s == Outpar || *s == Inpar
+		|| (isset(EXTENDEDGLOB) && *s == Tilde))
 		break;
 	if (*s == Inpar) {
 	    *s++ = '\0';
@@ -138,7 +153,7 @@ Lknode *np;
 		} else if (*s == ',' || *s == Comma) {
 		    s++;
 		    if (qualct) {
-			qn = (struct qual *) hcalloc(sizeof *qn);
+			qn = (struct qual *)hcalloc(sizeof *qn);
 			qo->or = qn;
 			qo = qn;
 			qualorct++;
@@ -356,7 +371,7 @@ Lknode *np;
 		    }
 		if (func) {
 		    if (!qn)
-			qn = (struct qual *) hcalloc(sizeof *qn);
+			qn = (struct qual *)hcalloc(sizeof *qn);
 		    if (ql)
 			ql->next = qn;
 		    ql = qn;
@@ -375,17 +390,16 @@ Lknode *np;
 		    return;
 	    }
 	}
-    } else if ((str[sl - 1] == '/') && 
+    } else if ((str[sl - 1] == '/') &&
 	       !((str[sl - 2] == Star) &&
 		 (str[sl - 3] == Star))) {	/* foo/ == foo(/) */
 	str[sl - 1] = '\0';
-	quals = (struct qual *) hcalloc(sizeof *quals);
+	quals = (struct qual *)hcalloc(sizeof *quals);
 	quals->func = qualmode;
 	quals->data = S_IFDIR;
 	quals->sense = 0;
 	qualct = 1;
     }
-
     if (*str == '/') {		/* pattern has absolute path */
 	str++;
 	pathbuf[0] = '/';
@@ -405,6 +419,9 @@ Lknode *np;
     matchptr = matchbuf = (char **)zalloc((matchsz = 16) * sizeof(char *));
 
     matchct = 0;
+    old_ino = (ino_t) 0;
+    old_dev = (dev_t) 0;
+    old_pos = -1;
     scanner(q);			/* do the globbing */
     if (matchct)
 	badcshglob |= 2;
@@ -500,8 +517,6 @@ char *s;
     struct stat buf, buf2, *bp;
     int statted = 0;
 
-    if (exclude && domatch(s, exclude, gf_noglobdots))
-	return;
     if (gf_listtypes || gf_markdirs) {
 	statted = 1;
 	if (!lstat(s, &buf) && (gf_listtypes || S_ISDIR(buf.st_mode))) {
@@ -562,30 +577,30 @@ char *s;
 char file_type(mode_t filemode)
 {
 #else
-char file_type(filemode)    /**/
+char file_type(filemode)	/**/
 mode_t filemode;
 {
 #endif
-    switch (filemode & S_IFMT) {	/* screw POSIX */
-      case S_IFDIR:
+    switch (filemode & S_IFMT) {/* screw POSIX */
+    case S_IFDIR:
 	return '/';
 #ifdef S_IFIFO
-      case S_IFIFO:
+    case S_IFIFO:
 	return '|';
 #endif
-      case S_IFCHR:
+    case S_IFCHR:
 	return '%';
-      case S_IFBLK:
+    case S_IFBLK:
 	return '#';
 #ifdef S_IFLNK
-      case S_IFLNK:
+    case S_IFLNK:
 	return /* (access(pbuf, F_OK) == -1) ? '&' :*/ '@';
 #endif
 #ifdef S_IFSOCK
-      case S_IFSOCK:
+    case S_IFSOCK:
 	return '=';
 #endif
-      default:
+    default:
 	if (filemode & 0111)
 	    return '*';
 	else
@@ -754,8 +769,8 @@ Lknode *np;
 		c1 = ztokens[c1 - STOUC(Pound)];
 	    if (itok(c2 = *p))
 		c2 = ztokens[c2 - STOUC(Pound)];
-	    if (c1 == '-' && lastch && p < str2 && (int) lastch <= (int) c2) {
-		while ((int) lastch < (int) c2)
+	    if (c1 == '-' && lastch && p < str2 && (int)lastch <= (int)c2) {
+		while ((int)lastch < (int)c2)
 		    ccl[lastch++] = 1;
 		lastch = 0;
 	    } else
@@ -877,7 +892,7 @@ char *b;
 /* do the ${foo%%bar}, ${foo#bar} stuff */
 /* please do not laugh at this code. */
 
-char *get_match_ret(s, b, e, fl)	/**/
+char *get_match_ret(s, b, e, fl)/**/
 char *s;
 int b;
 int e;
@@ -983,9 +998,9 @@ int n;
 	break;
 
     case 3:
-	for (i = 1, t = s; i < l; i++, t++) {
+	for (i = 0, t = s; i < l; i++, t++) {
 	    if (domatch(t, c, 0) && !--n) {
-		*sp = get_match_ret(*sp, i - 1, l, fl);
+		*sp = get_match_ret(*sp, i, l, fl);
 		return;
 	    }
 	}
@@ -1056,12 +1071,12 @@ int n;
 
 /* add a component to pathbuf */
 
-SPROTO(int addpath, (char * s));
+SPROTO(int addpath, (char *s));
 
 static int addpath(s)
 char *s;
 {
-    if ((int) strlen(s) + pathpos >= MAXPATHLEN)
+    if ((int)strlen(s) + pathpos >= MAXPATHLEN)
 	return 0;
     while ((pathbuf[pathpos++] = *s++));
     pathbuf[pathpos - 1] = '/';
@@ -1086,16 +1101,27 @@ Complist q;
 {
     Comp c;
     int closure;
+    struct stat st;
 
     if (!q)
 	return;
+
+    if (q->closure && old_pos != pathpos &&
+	stat((*pathbuf) ? pathbuf : ".", &st) != -1)
+	if (st.st_ino == old_ino && st.st_dev == old_dev)
+	    return;
+	else {
+	    old_pos = pathpos;
+	    old_ino = st.st_ino;
+	    old_dev = st.st_dev;
+	}
     if ((closure = q->closure))	/* (foo/)# */
 	if (q->closure == 2)	/* (foo/)## */
 	    q->closure = 1;
 	else
 	    scanner(q->next);
     if ((c = q->comp)) {
-	if (!(c->next || c->left) && !haswilds(c->str))
+	if (!(c->next || c->left) && !haswilds(c->str)) {
 	    if (q->next) {
 		int oppos = pathpos;
 
@@ -1113,6 +1139,7 @@ Complist q;
 
 		if (exists(s = getfullpath(c->str)))
 		    insert(dupstring(s));
+	    }
 	} else {
 	    char *fn;
 	    int dirs = !!q->next;
@@ -1199,6 +1226,26 @@ int fist;
 
 #define untok(C)  (itok(C) ? ztokens[(C) - Pound] : (C))
 
+/* See if pattern has a matching exclusion (~...) part */
+
+int excluded(c, eptr, efst)	/**/
+Comp c;
+char *eptr;
+int efst;
+{
+    char *saves = pptr;
+    int savei = first, ret;
+
+    first = efst;
+    pptr = (PATHADDP(c) && pathpos) ? getfullpath(eptr) : eptr;
+
+    ret = doesmatch(c->exclude);
+
+    pptr = saves;
+    first = savei;
+
+    return ret;
+}
 
 /* see if current pattern matches c */
 
@@ -1206,9 +1253,10 @@ int doesmatch(c)		/**/
 Comp c;
 {
     char *pat = c->str;
+    int done = 0;
 
- tailrec:
-    if (c->closure == 1) {
+  tailrec:
+    if (ONEHASHP(c) || (done && TWOHASHP(c))) {
 	char *saves = pptr;
 
 	if (first && *pptr == '.')
@@ -1218,6 +1266,7 @@ Comp c;
 	pptr = saves;
 	first = 0;
     }
+    done++;
     for (;;) {
 	if (!pat || !*pat) {
 	    char *saves;
@@ -1228,7 +1277,8 @@ Comp c;
 	    saves = pptr;
 	    savei = first;
 	    if (c->left || c->right)
-		if (!doesmatch(c->left))
+		if (!doesmatch(c->left) ||
+		    (c->exclude && excluded(c, saves, savei)))
 		    if (c->right) {
 			pptr = saves;
 			first = savei;
@@ -1236,13 +1286,14 @@ Comp c;
 			    return 0;
 		    } else
 			return 0;
-	    if (c->closure) {
+	    if (*pptr && CLOSUREP(c)) {
 		pat = c->str;
 		goto tailrec;
 	    }
 	    if (!c->next)
-		return (!c->last || !*pptr);
+		return (!LASTP(c) || !*pptr);
 	    c = c->next;
+	    done = 0;
 	    pat = c->str;
 	    goto tailrec;
 	}
@@ -1345,21 +1396,6 @@ Comp c;
 Complist parsepat(str)		/**/
 char *str;
 {
-    char *s;
-
-    exclude = NULL;
-    if (isset(EXTENDEDGLOB)) {
-	s = str + strlen(str);
-	while (s-- > str) {
-	    if (*s == Tilde && s[1]) {
-		*s++ = '\0';
-		exclude = parsereg(s);
-		if (!exclude)
-		    return NULL;
-		break;
-	    }
-	}
-    }
     mode = 0;
     pptr = str;
     return parsecomplist();
@@ -1370,7 +1406,7 @@ char *str;
 {
     mode = 1;
     pptr = str;
-    return parsecompsw();
+    return parsecompsw(0);
 }
 
 Complist parsecomplist()
@@ -1385,12 +1421,12 @@ Complist parsecomplist()
 	if (pptr[-1] == Star)
 	    pptr += 2;
 	p1 = (Complist) alloc(sizeof *p1);
-	if ( (p1->next = parsecomplist()) == NULL ) {
+	if ((p1->next = parsecomplist()) == NULL) {
 	    errflag = 1;
 	    return NULL;
 	}
 	p1->comp = (Comp) alloc(sizeof *p1->comp);
-	p1->comp->last = 1;
+	p1->comp->stat |= C_LAST;
 	p1->comp->str = dupstring("*");
 	*p1->comp->str = Star;
 	p1->closure = 1;
@@ -1408,7 +1444,7 @@ Complist parsecomplist()
 	if (str[0] != Pound || str[-1] != Outpar || str[-2] != '/')
 	    goto kludge;
 	pptr++;
-	if (!(c1 = parsecompsw()))
+	if (!(c1 = parsecompsw(0)))
 	    return NULL;
 	if (pptr[0] == '/' && pptr[1] == Outpar && pptr[2] == Pound) {
 	    int pdflag = 0;
@@ -1426,7 +1462,7 @@ Complist parsecomplist()
 	}
     } else {
       kludge:
-	if (!(c1 = parsecompsw()))
+	if (!(c1 = parsecompsw(1)))
 	    return NULL;
 	if (*pptr == '/' || !*pptr) {
 	    int ef = *pptr == '/';
@@ -1434,7 +1470,7 @@ Complist parsecomplist()
 	    p1 = (Complist) alloc(sizeof *p1);
 	    p1->comp = c1;
 	    p1->closure = 0;
-	    p1->next = (*pptr == '/') ? (pptr++, parsecomplist()) : NULL;
+	    p1->next = ef ? (pptr++, parsecomplist()) : NULL;
 	    return (ef && !p1->next) ? NULL : p1;
 	}
     }
@@ -1450,6 +1486,8 @@ Comp parsecomp()
     c->next = tail;
 
     while (*pptr && (mode || *pptr != '/') && *pptr != Bar &&
+	   (!isset(EXTENDEDGLOB) || *pptr != Tilde ||
+	    !pptr[1] || pptr[1] == Outpar || pptr[1] == Bar) &&
 	   *pptr != Outpar) {
 	if (*pptr == Hat) {
 	    *s++ = Hat;
@@ -1459,12 +1497,15 @@ Comp parsecomp()
 		return NULL;
 	    return c;
 	}
-	if (*pptr == Star && pptr[1] && (mode || pptr[1] != '/')) {
+	if (*pptr == Star && pptr[1] &&
+	    (!isset(EXTENDEDGLOB) || pptr[1] != Tilde || !pptr[2] ||
+	     pptr[2] == Bar ||
+	     pptr[2] == Outpar) && (mode || pptr[1] != '/')) {
 	    *s++ = '\0';
 	    pptr++;
 	    c1 = (Comp) alloc(sizeof *c1);
 	    *(c1->str = dupstring("?")) = Quest;
-	    c1->closure = 1;
+	    c1->stat |= C_ONEHASH;
 	    if (!(c2 = parsecomp()))
 		return NULL;
 	    c1->next = c2;
@@ -1496,15 +1537,16 @@ Comp parsecomp()
 	    }
 	    if (!(c1 = parsecomp()))
 		return NULL;
-	    tail = c1;
+	    tail = dpnd ? NULL : c1;
 	    endp = pptr;
 	    pptr = startp;
 	    pptr++;
 	    *s++ = '\0';
 	    c->next = (Comp) alloc(sizeof *c);
-	    c->next->left = parsecompsw();
-	    c->next->closure = dpnd;
-	    c->next->next = (Comp) alloc(sizeof *c);
+	    c->next->left = parsecompsw(0);
+	    if (dpnd)
+		c->next->stat |= (dpnd == 2) ? C_TWOHASH : C_ONEHASH;
+	    c->next->next = dpnd ? c1 : (Comp) alloc(sizeof *c);
 	    pptr = endp;
 	    tail = stail;
 	    return c;
@@ -1522,7 +1564,7 @@ Comp parsecomp()
 		c1 = c;
 	    c1->next = c2 = (Comp) alloc(sizeof *c);
 	    c2->str = dupstring(ls);
-	    c2->closure = 1;
+	    c2->stat |= C_ONEHASH;
 	    c2->next = parsecomp();
 	    if (!c2->next)
 		return NULL;
@@ -1552,27 +1594,47 @@ Comp parsecomp()
 	    *s++ = *pptr++;
     }
     if (*pptr == '/' || !*pptr)
-	c->last = 1;
+	c->stat |= C_LAST;
     *s++ = '\0';
     return c;
 }
 
-Comp parsecompsw()
-{				/**/
-    Comp c1, c2, c3;
+Comp parsecompsw(pathadd)	/**/
+int pathadd;
+{
+    Comp c1, c2, c3, excl = NULL;
 
     c1 = parsecomp();
     if (!c1)
 	return NULL;
-    if (*pptr == Bar) {
-	c2 = (Comp) alloc(sizeof *c2);
+    if (isset(EXTENDEDGLOB) && *pptr == Tilde) {
+	int oldmode = mode;
+
+	mode = 1;
 	pptr++;
-	c3 = parsecompsw();
-	if (!c3)
+	excl = parsecomp();
+	mode = oldmode;
+	if (!excl)
 	    return NULL;
+    }
+    if (*pptr == Bar || excl) {
+	c2 = (Comp) alloc(sizeof *c2);
+	if (*pptr == Bar) {
+	    pptr++;
+	    c3 = parsecompsw(pathadd);
+	    if (!c3)
+		return NULL;
+	} else {
+	    if (!*pptr || *pptr == '/')
+		c2->stat |= C_LAST;
+	    c3 = NULL;
+	}
 	c2->str = dupstring("");
 	c2->left = c1;
 	c2->right = c3;
+	c2->exclude = excl;
+	if (pathadd)
+	    c2->stat |= C_PATHADD;
 	return c2;
     }
     return c1;
