@@ -40,7 +40,8 @@ char **envp;
 	zshname = argv[0];
     else
 	zshname++;
-    if ( *zshname == '-' ) zshname++;
+    if (*zshname == '-')
+	zshname++;
     setflags(zshname);
     parseargs(argv);
     setmoreflags();
@@ -126,9 +127,9 @@ const char *zshname;
 {
     int c;
 
-    for (c = 0; c != 42; c++)
+    for (c = 0; c != 43; c++)
 	opts[c] = OPT_UNSET;
-    for (c = 42; c != 128; c++)
+    for (c = 43; c != 128; c++)
 	opts[c] = OPT_INVALID;
     for (c = 'a'; c <= 'z'; c++)
 	opts[c] = opts[c - 'a' + 'A'] = OPT_UNSET;
@@ -156,6 +157,7 @@ const char *zshname;
     if (strcmp(zshname, "sh") == 0 || strcmp(zshname, "ksh") == 0) {
 	opts['K'] = opts['k'] = opts['I'] = opts['3'] = OPT_SET;
 	opts['H'] = opts['y'] = opts[KSHOPTIONPRINT] = OPT_SET;
+	opts[GLOBSUBST] = OPT_SET;
     }
 }
 
@@ -275,7 +277,7 @@ void setmoreflags()
 #if defined(TIOCSETD) && defined(NTTYDISC)
 	    int ldisc = NTTYDISC;
 
-	    ioctl(SHTTY, TIOCSETD, (char *) &ldisc);
+	    ioctl(SHTTY, TIOCSETD, (char *)&ldisc);
 #endif
 	    gettyinfo(&shttyinfo);	/* get tty state */
 #if defined(__sgi)
@@ -468,13 +470,6 @@ const char *zshname;
     fceditparam = ztrdup(DEFFCEDIT);
     tmpprefix = ztrdup(DEFTMPPREFIX);
     postedit = ztrdup("");
-    if (ispwd(home))
-	pwd = ztrdup(home);
-    else if ((ptr = zgetenv("PWD")) && ispwd(ptr))
-	pwd = ztrdup(ptr);
-    else
-	pwd = zgetwd();
-    oldpwd = ztrdup(pwd);
     hostnam = (char *)zalloc(256);
     underscore = ztrdup("");
     gethostname(hostnam, 256);
@@ -506,12 +501,12 @@ const char *zshname;
 #ifdef SYSV
 	!(username = cuserid(NULL))
 #else
-	  !(username = getlogin())
+	!(username = getlogin())
 #endif
-	  ) {
+	) {
 	if ((pswd = getpwuid(getuid()))) {
 	    username = ztrdup(pswd->pw_name);
-	    if(!home)
+	    if (!home)
 		home = ztrdup(pswd->pw_dir);
 	} else {
 	    username = ztrdup("");
@@ -524,6 +519,14 @@ const char *zshname;
     else
 	zlogname = ztrdup(username);
 
+    if (ispwd(home))
+	pwd = ztrdup(home);
+    else if ((ptr = zgetenv("PWD")) && ispwd(ptr))
+	pwd = ztrdup(ptr);
+    else
+	pwd = zgetwd();
+    oldpwd = ztrdup(pwd);
+
     setparams();
 
     inittyptab();
@@ -537,7 +540,7 @@ const char *zshname;
 void compctlsetup()
 {				/**/
     static char
-	*os[] =
+        *os[] =
     {"setopt", "unsetopt", NULL}, *vs[] =
     {"export", "typeset", "vared", "unset", NULL}, *cs[] =
     {"which", "builtin", NULL}, *bs[] =
@@ -560,6 +563,7 @@ void initialize()
 {				/**/
 #ifdef RLIM_INFINITY
     int t0;
+
 #endif
 
     breaks = loops = 0;
@@ -611,15 +615,15 @@ void initialize()
 #ifdef SIGNAL_MASKS
     if (islogin) {
 	sig_setmask(sig_mask(0));
-    }
-    else if (interact) {
+    } else if (interact) {
 	sigset_t set;
+
 	sigemptyset(&set);
 	sigaddset(&set, SIGINT);
 	sigaddset(&set, SIGQUIT);
 	sig_unblock(set);
     }
-#endif /* SIGNAL_MASKS */
+#endif				/* SIGNAL_MASKS */
 }
 
 void addreswords()
@@ -653,7 +657,10 @@ char *zshname;
 	    source("/etc/profile");
 	if (islogin)
 	    sourcehome(".profile");
-	source(getsparam("ENV"));
+	if (!isset(KSHPRIV))
+	    source(getsparam("ENV"));
+	else
+	    source("/etc/suid_profile");
     } else {
 #ifdef GLOBALZSHENV
 	source(GLOBALZSHENV);
@@ -680,6 +687,7 @@ char *zshname;
 	    }
 	}
     }
+    sourcelevel = 0;
 
     if (isset('c')) {
 	if (SHIN >= 10)
@@ -696,8 +704,6 @@ char *zshname;
 #endif
     if (isset('t') && opti == OPT_UNSET)
 	prompt = ztrdup("");
-
-    sourcelevel = 0;
 }
 
 void ainit()
