@@ -22,13 +22,13 @@
 #include <setjmp.h>
 
 #if defined(POSIX) && !defined(__386BSD__) && !defined(__NetBSD__) && !defined(__FreeBSD__)
-# define JMP_BUF       sigjmp_buf
-# define SETJMP(b)     sigsetjmp((b), 1)
-# define LONGJMP(b,n)  siglongjmp((b), (n))
+#define JMP_BUF       sigjmp_buf
+#define SETJMP(b)     sigsetjmp((b), 1)
+#define LONGJMP(b,n)  siglongjmp((b), (n))
 #else
-# define JMP_BUF       jmp_buf
-# define SETJMP(b)     setjmp(b)
-# define LONGJMP(b,n)  longjmp((b), (n))
+#define JMP_BUF       jmp_buf
+#define SETJMP(b)     setjmp(b)
+#define LONGJMP(b,n)  longjmp((b), (n))
 #endif
 
 #if defined(RESETHANDNEEDED) && !defined(POSIX)
@@ -41,7 +41,7 @@
 
 /* empty job structure for quick clearing of jobtab entries */
 
-static struct job zero; /* static variables are initialized to zero */
+static struct job zero;		/* static variables are initialized to zero */
 
 struct timeval dtimeval, now;
 struct timezone dummy_tz;
@@ -69,15 +69,16 @@ static struct z_jmp_buf {
     JMP_BUF jbuf;
 } foil_race;
 
-void chldhandler DCLPROTO((struct z_jmp_buf *jump));
+void chldhandler DCLPROTO((struct z_jmp_buf * jump));
 
-void chldsuspend()
-{				/**/
+void chldsuspend(sig)		/**/
+int sig;
+{
 /* assumes blockchld() is in effect */
     if (SETJMP(foil_race.jbuf) == 0) {
 	chld_longjmp = 1;
 	unblockchld();
-	chldpause();
+	chldpause(sig);
     }
     chld_longjmp = 0;
 }
@@ -114,7 +115,6 @@ int sig;
 	if (do_jump)
 	    jump_to = foil_race;/* copy foil_race */
     }
-
     fast_unblock(&heldsigs);	/* Signal traps OK again (foil_race copied) */
 
     switch (sig) {
@@ -289,8 +289,9 @@ Job jn;
 		0200 | WTERMSIG(pn->statusp) : WEXITSTATUS(pn->statusp);
 	if (pn->pid == jn->gleader) {
 	    statusp = pn->statusp;
-	    if (pgrp == 0 || pn->pid == pgrp)
-	      inforeground = 1;
+	    if (pgrp == 0 || pn->pid == pgrp || 
+		(pgrp > 1 && kill(-pgrp, 0) == -1))
+		inforeground = 1;
 	}
     }
     if (!notrunning)
@@ -298,6 +299,7 @@ Job jn;
     if (somestopped) {
 	if (jn->stty_in_env && !jn->ty) {
 	    jn->ty = (struct ttyinfo *)zalloc(sizeof(struct ttyinfo));
+
 	    gettyinfo(jn->ty);
 	}
 	if (jn->stat & STAT_STOPPED)
@@ -305,6 +307,8 @@ Job jn;
     }
     if (alldone && job == thisjob)
 	lastval = val;
+    if (alldone)
+	lastval2 = val;
     if (inforeground && !ttyfrozen && !val && !jn->stty_in_env)
 	gettyinfo(&shttyinfo);
 #ifdef TIOCGWINSZ
@@ -324,7 +328,7 @@ Job jn;
     if (sigtrapped[SIGCHLD] && job != thisjob)
 	dotrap(SIGCHLD);
 
-    /* If the foreground job got a signal, pretend we got it, too. */
+ /* If the foreground job got a signal, pretend we got it, too. */
     if (inforeground && WIFSIGNALED(statusp)) {
 	if (sigtrapped[WTERMSIG(statusp)]) {
 	    dotrap(WTERMSIG(statusp));
@@ -395,7 +399,7 @@ int lng;
 		    doputnl = 1;
 	    } else if (WIFSTOPPED(pn->statusp)) {
 		sig = WSTOPSIG(pn->statusp);
-		if ((int) strlen(sigmsg[sig]) > len)
+		if ((int)strlen(sigmsg[sig]) > len)
 		    len = strlen(sigmsg[sig]);
 		if (job == thisjob && sig == SIGTSTP)
 		    doputnl = 1;
@@ -422,7 +426,7 @@ int lng;
 		for (qn = pn->next; qn; qn = qn->next) {
 		    if (qn->statusp != pn->statusp)
 			break;
-		    if ((int) strlen(qn->text) + len2 + ((qn->next) ? 3 : 0) > lineleng)
+		    if ((int)strlen(qn->text) + len2 + ((qn->next) ? 3 : 0) > lineleng)
 			break;
 		    len2 += strlen(qn->text) + 2;
 		}
@@ -431,7 +435,7 @@ int lng;
 		    fprintf(stderr, "[%ld]  %c ",
 			    (long)(jn - jobtab),
 			    (job == curjob) ? '+'
-			     : (job == prevjob) ? '-' : ' ');
+			    : (job == prevjob) ? '-' : ' ');
 		else
 		    fprintf(stderr, (job > 9) ? "        " : "       ");
 	    else
@@ -523,6 +527,7 @@ Job jn;
     }
     if (jn->ty)
 	zfree(jn->ty, sizeof(struct ttyinfo));
+
     *jn = zero;
 }
 
@@ -554,7 +559,7 @@ void initjob()
 
 /* add a process to the current job */
 
-void addproc(pid, text)	/**/
+void addproc(pid, text)		/**/
 long pid;
 char *text;
 {
@@ -604,7 +609,7 @@ long pid;
 /* blockchld() around this loop in case #ifndef WNOHANG */
     blockchld();		/* unblocked in chldsuspend() */
     while (!errflag && (kill(pid, 0) >= 0 || errno != ESRCH)) {
-	chldsuspend();
+	chldsuspend(SIGINT);
 	blockchld();
     }
     unblockchld();
@@ -612,21 +617,21 @@ long pid;
 
 /* wait for a job to finish */
 
-void waitjob(job)		/**/
+void waitjob(job, sig)		/**/
 int job;
+int sig;
 {
     Job jn = jobtab + job;
 
     blockchld();		/* unblocked during chldsuspend() */
-    if (jn->procs) {	/* if any forks were done */
+    if (jn->procs) {		/* if any forks were done */
 	jn->stat |= STAT_LOCKED;
 	if (jn->stat & STAT_CHANGED)
 	    printjob(jobtab + job, !!isset(LONGLISTJOBS));
-	while (jn->stat &&
+	while (!errflag && jn->stat &&
 	       !(jn->stat & STAT_DONE) &&
-	       !(interact && (jn->stat & STAT_STOPPED)))
-	{
-	    chldsuspend();
+	       !(interact && (jn->stat & STAT_STOPPED))) {
+	    chldsuspend(sig);
 	    blockchld();
 	}
     } else
@@ -638,7 +643,7 @@ int job;
 
 void waitjobs()
 {				/**/
-    waitjob(thisjob);
+    waitjob(thisjob, 0);
     thisjob = -1;
 }
 
@@ -653,6 +658,7 @@ void clearjobtab()
 	    zsfree(jobtab[t0].pwd);
 	if (jobtab[t0].ty)
 	    zfree(jobtab[t0].ty, sizeof(struct ttyinfo));
+
 	jobtab[t0] = zero;
     }
 }
@@ -752,7 +758,7 @@ char *desc;
 	    switch (s++, *s) {
 	    case 'E':
 		fprintf(stderr, "%ld.%03lds",
-			(long) real->tv_sec, (long) real->tv_usec / 1000);
+			(long)real->tv_sec, (long)real->tv_usec / 1000);
 		break;
 #ifndef HAS_RUSAGE
 	    case 'U':
@@ -764,42 +770,40 @@ char *desc;
 			ti->st / HZ, ti->st * 1000 / HZ % 1000);
 		break;
 	    case 'P':
-      if ( real->tv_sec > 21000 ) {
-         real100 = (real->tv_sec + 99)/100;
+		if (real->tv_sec > 21000) {
+		    real100 = (real->tv_sec + 99) / 100;
 		    fprintf(stderr, "%d%%",
-                 (int)((ti->ut + ti->st) / HZ)/real100);
-      }
-      else {
-         if (( real100 = real->tv_sec * 1000 + real->tv_usec / 1000 ))
-            fprintf(stderr, "%d%%",
-                    (int)(100000 * ((ti->ut + ti->st) / HZ)) / real100);
-      }
+			    (int)((ti->ut + ti->st) / HZ) / real100);
+		} else {
+		    if ((real100 = real->tv_sec * 1000 + real->tv_usec / 1000))
+			fprintf(stderr, "%d%%",
+				(int)(100000 * ((ti->ut + ti->st) / HZ)) / real100);
+		}
 		break;
 #else
 	    case 'U':
 		fprintf(stderr, "%ld.%03lds",
-			(long) ru->ru_utime.tv_sec,
-			(long) ru->ru_utime.tv_usec / 1000);
+			(long)ru->ru_utime.tv_sec,
+			(long)ru->ru_utime.tv_usec / 1000);
 		break;
 	    case 'S':
 		fprintf(stderr, "%ld.%03lds",
-			(long) ru->ru_stime.tv_sec,
-			(long) ru->ru_stime.tv_usec / 1000);
+			(long)ru->ru_stime.tv_sec,
+			(long)ru->ru_stime.tv_usec / 1000);
 		break;
 	    case 'P':
-      if ( real->tv_sec > 21000 ) {
-         real100 = (real->tv_sec + 99)/100;
+		if (real->tv_sec > 21000) {
+		    real100 = (real->tv_sec + 99) / 100;
 		    fprintf(stderr, "%ld%%",
-                 (ru->ru_utime.tv_sec + ru->ru_stime.tv_sec)/real100);
-      }
-      else {
-         if (( real100 = real->tv_sec * 1000 + real->tv_usec / 1000 ))
-            fprintf(stderr, "%ld%%",
-                    (100000 * (ru->ru_utime.tv_sec + ru->ru_stime.tv_sec)
-			     + 
-                     (ru->ru_utime.tv_usec + ru->ru_stime.tv_usec) / 10)
-                    / real100);
-      }
+			    (ru->ru_utime.tv_sec + ru->ru_stime.tv_sec) / real100);
+		} else {
+		    if ((real100 = real->tv_sec * 1000 + real->tv_usec / 1000))
+			fprintf(stderr, "%ld%%",
+				(100000 * (ru->ru_utime.tv_sec + ru->ru_stime.tv_sec)
+				 +
+				 (ru->ru_utime.tv_usec + ru->ru_stime.tv_usec) / 10)
+				/ real100);
+		}
 		break;
 	    case 'W':
 		fprintf(stderr, "%ld", ru->ru_nswap);
