@@ -22,7 +22,9 @@
 
 /* lexical state */
 
-static int dbparens;
+static int dbparens, in_brace_param = 0;
+int len = 0, bsiz = 256;
+char *bptr;
 
 struct lexstack {
     struct lexstack *next;
@@ -31,6 +33,7 @@ struct lexstack {
     int incond;
     int incasepat;
     int dbparens;
+    int in_brace_param;
     int alstat;
     char *hlastw;
     int isfirstln;
@@ -44,6 +47,9 @@ struct lexstack {
     int hlinesz;
     char *hline;
     char *hptr;
+    char *tokstr;
+    char *bptr;
+    int bsiz;
 
     unsigned char *cstack;
     int csp;
@@ -65,6 +71,7 @@ void lexsave()
     ls->incond = incond;
     ls->incasepat = incasepat;
     ls->dbparens = dbparens;
+    ls->in_brace_param = in_brace_param;
     ls->alstat = alstat;
     ls->alstackind = alstackind;
     ls->isfirstln = isfirstln;
@@ -81,6 +88,10 @@ void lexsave()
     ls->cstack = cmdstack;
     ls->csp = cmdsp;
     cmdstack = (unsigned char *)zalloc(256);
+    ls->tokstr = tokstr;
+    ls->bptr = bptr;
+    ls->bsiz = bsiz;
+    in_brace_param = 0;
     cmdsp = 0;
     inredir = 0;
 
@@ -102,6 +113,7 @@ void lexrestore()
     incond = lstack->incond;
     incasepat = lstack->incasepat;
     dbparens = lstack->dbparens;
+    in_brace_param = lstack->in_brace_param;
     alstat = lstack->alstat;
     isfirstln = lstack->isfirstln;
     isfirstch = lstack->isfirstch;
@@ -117,6 +129,9 @@ void lexrestore()
 	free(cmdstack);
     cmdstack = lstack->cstack;
     cmdsp = lstack->csp;
+    tokstr = lstack->tokstr;
+    bptr = lstack->bptr;
+    bsiz = lstack->bsiz;
     clearalstack();
     alstackind = lstack->alstackind;
     hlinesz = lstack->hlinesz;
@@ -253,7 +268,7 @@ void initlextabs()
 void lexinit()
 {				/**/
     incond = incasepat = nocorrect =
-    dbparens = alstat = lexstop = 0;
+    dbparens = alstat = lexstop = in_brace_param = 0;
     incmdpos = 1;
     tok = ENDINPUT;
     if (isset(EXTENDEDGLOB)) {
@@ -265,9 +280,6 @@ void lexinit()
     }
 }
 
-int len = 0, bsiz = 256;
-char *bptr;
-
 /* add a char to the string buffer */
 
 void add(c)			/**/
@@ -297,7 +309,7 @@ int gettok()
 {				/**/
     int bct = 0, pct = 0, brct = 0;
     int c, d, intpos = 1;
-    int peekfd = -1, peek, ninbracks;
+    int peekfd = -1, peek, ninbracks, intick;
 
   beginning:
     hlastw = NULL;
@@ -324,9 +336,8 @@ int gettok()
 
     if (c == hashchar &&
 	(isset(INTERACTIVECOMMENTS) ||
-	 (!zleparse && (!interact || unset(SHINSTDIN) || strin))))
-    {
-		/* History is handled here to prevent extra newlines
+	 (!zleparse && (!interact || unset(SHINSTDIN) || strin)))) {
+    /* History is handled here to prevent extra newlines
 		 * being inserted into the history.
 		 *
 		 * Also ignore trailing spaces to prevent history from
@@ -444,15 +455,15 @@ int gettok()
 	    int tbs = 256, n = 0, nc;
 	    char *tbuf, *tbp, *ntb;
 
-	    tbuf = tbp = (char *) zalloc(tbs);
+	    tbuf = tbp = (char *)zalloc(tbs);
 	    hungetc(d);
 
 	    while ((nc = hgetc()) && !lexstop) {
 		if (!idigit(nc) && nc != '-')
 		    break;
-		*tbp++ = (char) nc;
+		*tbp++ = (char)nc;
 		if (++n == tbs) {
-		    ntb = (char *) realloc(tbuf, tbs *= 2);
+		    ntb = (char *)realloc(tbuf, tbs *= 2);
 		    tbp += ntb - tbuf;
 		    tbuf = ntb;
 		}
@@ -561,7 +572,9 @@ int gettok()
 	}
 	switch (act) {
 	case LX2_BREAK:
-	    goto brk;
+	    if (!in_brace_param)
+		goto brk;
+	    break;
 	case LX2_OUTPAR:
 	    if (!pct)
 		goto brk;
@@ -597,8 +610,11 @@ int gettok()
 		    goto brk;
 		}
 		c = Outpar;
-	    } else
+	    } else {
+		if (e == '{')
+		    in_brace_param = 1;
 		hungetc(e);
+	    }
 	    break;
 	case LX2_INBRACK:
 	    brct++;
@@ -627,7 +643,8 @@ int gettok()
 	case LX2_OUTBRACE:
 	    if (!bct)
 		goto brk;
-	    bct--;
+	    if (!--bct && in_brace_param)
+		in_brace_param = 0;
 	    c = Outbrace;
 	    break;
 	case LX2_OUTANG:
@@ -740,8 +757,9 @@ int gettok()
 	    break;
 	case LX2_DQUOTE:
 	    add(Dnull);
-	    cmdpush(CS_DQUOTE);
+	    cmdpush(dbparens ? CS_MATH : CS_DQUOTE);	/* math or dquote */
 	    endchar = dbparens ? ')' : '"';
+	    intick = 0;
 	    while (((c = hgetc()) != endchar || (dbparens && pct > 0))
 		   && !lexstop)
 		if (c == '\\') {
@@ -758,7 +776,7 @@ int gettok()
 			else
 			    break;
 		    }
-		    if (c == '$') {
+		    if (c == '$' && !intick) {
 			e = hgetc();
 			if (e == '(') {
 			    add(Qstring);
@@ -778,15 +796,14 @@ int gettok()
 			    c = Qstring;
 			    hungetc(e);
 			}
-		    } else if (c == '`')
+		    } else if (c == '`') {
 			c = Qtick;
-		    else if (dbparens) {
+			intick = !intick;
+		    } else if (dbparens) {
 			if (c == '(')
 			    pct++;
 			else if (c == ')')
 			    pct--;
-			else if (c == '\n')
-			    break;
 		    }
 		    add(c);
 		}
