@@ -23,7 +23,6 @@
 vptr(*alloc) DCLPROTO((int));
 vptr(*ncalloc) DCLPROTO((int));
 
-
 /*
 
 	There are two ways to allocate memory in zsh.  The first way is
@@ -62,7 +61,6 @@ int h_m[1025], h_push, h_pop, h_free;
 
 #endif
 
-
 #define H_ISIZE  sizeof(long)
 
 #define HEAPSIZE (8192 - H_ISIZE)
@@ -149,8 +147,7 @@ void freeheap()
 	if (h->pos.next) {
 	    h->pos.free = h->pos.next->free;
 	    h->pos.ptr = h->pos.next->ptr;
-	}
-	else {
+	} else {
 	    h->pos.free += (int)(h->pos.ptr - h->arena);
 	    h->pos.ptr = h->arena;
 	}
@@ -175,9 +172,9 @@ void popheap()
 	    h->pos.free = hp->free;
 	    h->pos.ptr = hp->ptr;
 	    zfree(hp, sizeof(struct heappos));
+
 	    hl = h;
-	}
-	else {
+	} else {
 	    for (hp = h->pos.next; hp; hp = hpn) {
 		hpn = hp->next;
 		free(hp);
@@ -214,8 +211,7 @@ int size;
 	ret = h->pos.ptr;
 	h->pos.ptr += size;
 	h->pos.free -= size;
-    }
-    else {
+    } else {
 	int n = (size > HEAPSIZE) ? size : HEAPSIZE;
 
 #ifdef USE_ZSH_MALLOC
@@ -229,7 +225,7 @@ int size;
 	for (hp = NULL, h = heaps; h; hp = h, h = h->next);
 
 	h = (Heap) zalloc(sizeof *h);
-	h->arena = (char *) zalloc(n);
+	h->arena = (char *)zalloc(n);
 
 #ifdef USE_ZSH_MALLOC
 	if (called)
@@ -296,7 +292,6 @@ int l;
 	zerr("fatal error: out of memory", NULL, 0);
 	exit(1);
     }
-    
     zigsafe();
 
     return z;
@@ -336,15 +331,14 @@ const char *s;
     return t;
 }
 
-
 #ifdef USE_ZSH_MALLOC
 
 /*
-   Below is a simple segment oriented memory allocator for systems on 
+   Below is a simple segment oriented memory allocator for systems on
    which it is better than the system's one. Memory id given in blocks
-   aligned to an integer multiple of sizeof(long) (4 bytes on most machines, 
+   aligned to an integer multiple of sizeof(long) (4 bytes on most machines,
    but 8 bytes on e.g. a dec alpha). Each block is preceded by a header
-   which contains the length of the data part (in bytes). In allocated 
+   which contains the length of the data part (in bytes). In allocated
    blocks only this field of the structure m_hdr is senseful. In free
    blocks the second field (next) is a pointer to the next free segment
    on the free list.
@@ -358,7 +352,7 @@ const char *s;
    (free) contains a pointer to the first free block in the array. The
    last field (used) gives the number of already used blocks in the array.
 
-   If the macro name MEM_DEBUG is defined, some information about the memory 
+   If the macro name MEM_DEBUG is defined, some information about the memory
    usage is stored. This information can than be viewed by calling the
    builtin `mem' (which is only available if MEM_DEBUG is set).
 
@@ -410,7 +404,6 @@ struct m_hdr {
     long used;
 };
 
-
 #define M_ALIGN (sizeof(long))
 
 #define M_HSIZE (sizeof(struct m_hdr))
@@ -421,7 +414,6 @@ struct m_hdr *m_lfree, *m_free;
 long m_pgsz = 0;
 char *m_high, *m_low;
 
-
 #define M_SIDX(S)  ((S) / M_ISIZE)
 #define M_SNUM     50
 #define M_SLEN(M)  ((M)->len / M_SNUM)
@@ -431,11 +423,8 @@ char *m_high, *m_low;
 		     sizeof(long) - sizeof(struct m_hdr *)) / M_SNUM)
 #define M_NSMALL 8
 
-
 struct m_hdr *m_small[M_NSMALL];
 
-
-
 #ifdef MEM_DEBUG
 
 int m_s = 0, m_b = 0;
@@ -473,7 +462,6 @@ MALLOC_ARG_T size;
 #endif
 	m_free = m_lfree = NULL;
     }
-
     size = (size + M_ALIGN - 1) & ~(M_ALIGN - 1);
 
     if ((s = M_SIDX(size)) && s < M_NSMALL) {
@@ -487,7 +475,7 @@ MALLOC_ARG_T size;
 
 	    if (m->used == M_SNUM && m->next) {
 		for (mt = m; mt->next; mt = mt->next);
-		
+
 		mt->next = m;
 		if (mp)
 		    mp->next = m->next;
@@ -495,29 +483,26 @@ MALLOC_ARG_T size;
 		    m_small[s] = m->next;
 		m->next = NULL;
 	    }
-
 #ifdef MEM_DEBUG
 	    m_m[size / M_ISIZE]++;
 #endif
 
 	    return (MALLOC_RET_T) sh;
 	}
-
 	os = size;
 	size = M_SBLEN(size);
-    }
-    else
+    } else
 	s = 0;
 
     for (mp = NULL, m = m_free; m && m->len < size; mp = m, m = m->next);
 
-    /* If there is an empty zsh heap at a lower address we steel it and take
+ /* If there is an empty zsh heap at a lower address we steel it and take
        the memory from it, putting the rest on the free list. */
-    
+
     for (hp = NULL, h = heaps; h; hp = h, h = h->next)
 	if (h->pos.ptr == h->arena &&
 	    (!hf || h < hf) &&
-	    (!m || ((char *) m) > ((char *) h)))
+	    (!m || ((char *)m) > ((char *)h)))
 	    hf = h, hfp = hp;
 
     if (hf) {
@@ -536,19 +521,17 @@ MALLOC_ARG_T size;
 
 	for (mp = NULL, m = m_free; m && m->len < size; mp = m, m = m->next);
     }
-
     if (!m) {
 	n = (size + M_HSIZE + m_pgsz - 1) & ~(m_pgsz - 1);
 
-	if (((char *) (m = (struct m_hdr *) sbrk(n))) == ((char *) -1)) {
+	if (((char *)(m = (struct m_hdr *)sbrk(n))) == ((char *)-1)) {
 #ifdef MEM_WARNING
 	    zerr("allocation error at sbrk: %e", NULL, 0);
 #endif
 	    return NULL;
 	}
-
 	if (!m_low)
-	    m_low = (char *) m;
+	    m_low = (char *)m;
 
 #ifdef MEM_DEBUG
 	m_s += n;
@@ -557,7 +540,7 @@ MALLOC_ARG_T size;
 	    m_l = m;
 #endif
 
-	m_high = ((char *) m) + n;
+	m_high = ((char *)m) + n;
 
 	m->len = n - M_ISIZE;
 	m->next = NULL;
@@ -566,9 +549,8 @@ MALLOC_ARG_T size;
 	    m_lfree->next = m;
 	m_lfree = m;
     }
-
     if ((n = m->len - size) > M_MIN) {
-	struct m_hdr *mtt = (struct m_hdr *) (((char *) m) + M_ISIZE + size);
+	struct m_hdr *mtt = (struct m_hdr *)(((char *)m) + M_ISIZE + size);
 
 	mtt->len = n - M_ISIZE;
 	mtt->next = m->next;
@@ -582,13 +564,11 @@ MALLOC_ARG_T size;
 	    mp->next = mtt;
 	else
 	    m_free = mtt;
-    }
-    else if (mp) {
+    } else if (mp) {
 	if (m == m_lfree)
 	    m_lfree = mp;
 	mp->next = m->next;
-    }
-    else {
+    } else {
 	m_free = m->next;
 	if (m == m_lfree)
 	    m_lfree = m_free;
@@ -597,8 +577,9 @@ MALLOC_ARG_T size;
     if (s) {
 	struct m_shdr *sh, *shn;
 
-	m->free = sh = (struct m_shdr *) (((char *) m) + 
-					  sizeof(struct m_hdr) + os);
+	m->free = sh = (struct m_shdr *)(((char *)m) +
+					 sizeof(struct m_hdr) + os);
+
 	m->used = 1;
 
 	for (n = M_SNUM - 2; n--; sh = shn)
@@ -612,21 +593,20 @@ MALLOC_ARG_T size;
 	m_m[os / M_ISIZE]++;
 #endif
 
-	return (MALLOC_RET_T) (((char *) m) + sizeof(struct m_hdr));
+	return (MALLOC_RET_T) (((char *)m) + sizeof(struct m_hdr));
     }
-
 #ifdef MEM_DEBUG
     m_m[m->len < (1024 * M_ISIZE) ? (m->len / M_ISIZE) : 1024]++;
 #endif
 
-    return (MALLOC_RET_T) &m->next;
+    return (MALLOC_RET_T) & m->next;
 }
 
 void zfree(p, sz)		/**/
 vptr p;
 int sz;
 {
-    struct m_hdr *m = (struct m_hdr *) (((char *) p) - M_ISIZE), *mp, *mt;
+    struct m_hdr *m = (struct m_hdr *)(((char *)p) - M_ISIZE), *mp, *mt;
     int i;
 
 #ifdef SECURE_FREE
@@ -638,30 +618,29 @@ int sz;
     if (!p)
 	return;
 
-    if (((char *) p) < m_low || ((char *) p) > m_high ||
-	((long) p) & (M_ALIGN - 1)) {
+    if (((char *)p) < m_low || ((char *)p) > m_high ||
+	((long)p) & (M_ALIGN - 1)) {
 
 #ifdef MEM_WARNING
-	    zerr("attempt to free storage at invalid address", NULL, 0);
-#endif	
+	zerr("attempt to free storage at invalid address", NULL, 0);
+#endif
 
 	return;
     }
-
- fr_rec:
+  fr_rec:
 
     if ((i = sz / M_ISIZE) < M_NSMALL || !sz)
 	for (; i < M_NSMALL; i++) {
-	    for (mp = NULL, mt = m_small[i]; 
-		 mt && (((char *) mt) > ((char *) p) || 
-			(((char *) mt) + mt->len) < ((char *) p));
+	    for (mp = NULL, mt = m_small[i];
+		 mt && (((char *)mt) > ((char *)p) ||
+			(((char *)mt) + mt->len) < ((char *)p));
 		 mp = mt, mt = mt->next);
 
 	    if (mt) {
-		struct m_shdr *sh = (struct m_shdr *) p, *sh2;
+		struct m_shdr *sh = (struct m_shdr *)p, *sh2;
 
 #ifdef SECURE_FREE
-		if ((((char *) p) - (((char *) mt) + sizeof(struct m_hdr))) %
+		if ((((char *)p) - (((char *)mt) + sizeof(struct m_hdr))) %
 		    M_BSLEN(mt->len)) {
 
 #ifdef MEM_WARNING
@@ -669,12 +648,11 @@ int sz;
 #endif
 		    return;
 		}
-
 		for (sh2 = mt->free; sh2; sh2 = sh2->next)
-		    if (((char *) p) == ((char *) sh2)) {
+		    if (((char *)p) == ((char *)sh2)) {
 
 #ifdef MEM_WARNING
-		    zerr("attempt to free already free storage", NULL, 0);
+			zerr("attempt to free already free storage", NULL, 0);
 #endif
 			return;
 		    }
@@ -686,7 +664,7 @@ int sz;
 #ifdef MEM_DEBUG
 		m_f[M_BSLEN(mt->len) / M_ISIZE]++;
 #endif
-		
+
 		if (--mt->used) {
 		    if (mp) {
 			mp->next = mt->next;
@@ -695,36 +673,33 @@ int sz;
 		    }
 		    return;
 		}
-
 		if (mp)
 		    mp->next = mt->next;
 		else
 		    m_small[i] = mt->next;
 
 		m = mt;
-		p = (vptr) &m->next;
+		p = (vptr) & m->next;
 
 		break;
-	    }
-	    else if (sz) {
+	    } else if (sz) {
 		sz = 0;
 		goto fr_rec;
 	    }
 	}
-
 #ifdef MEM_DEBUG
     if (!mt)
 	m_f[m->len < (1024 * M_ISIZE) ? (m->len / M_ISIZE) : 1024]++;
 #endif
 
 #ifdef SECURE_FREE
-    for (mt = (struct m_hdr *) m_low; 
-	 ((char *) mt) < m_high;
-	 mt = (struct m_hdr *) (((char *) mt) + M_ISIZE + mt->len))
-	if (((char *) p) == ((char *) &mt->next))
+    for (mt = (struct m_hdr *)m_low;
+	 ((char *)mt) < m_high;
+	 mt = (struct m_hdr *)(((char *)mt) + M_ISIZE + mt->len))
+	if (((char *)p) == ((char *)&mt->next))
 	    break;
 
-    if (((char *) mt) >= m_high) {
+    if (((char *)mt) >= m_high) {
 
 #ifdef MEM_WARNING
 	zerr("attempt to free storage at invalid address", NULL, 0);
@@ -742,30 +717,27 @@ int sz;
 #endif
 	return;
     }
-
-    if (mt && ((char *) mt) == (((char *) m) + M_ISIZE + m->len)) {
+    if (mt && ((char *)mt) == (((char *)m) + M_ISIZE + m->len)) {
 	m->len += mt->len + M_ISIZE;
 	m->next = mt->next;
-	
+
 	if (mt == m_lfree)
 	    m_lfree = m;
-    }
-    else
+    } else
 	m->next = mt;
 
-    if (mp && ((char *) m) == (((char *) mp) + M_ISIZE + mp->len)) {
+    if (mp && ((char *)m) == (((char *)mp) + M_ISIZE + mp->len)) {
 	mp->len += m->len + M_ISIZE;
 	mp->next = m->next;
 
 	if (m == m_lfree)
 	    m_lfree = mp;
-    }
-    else if(mp)
+    } else if (mp)
 	mp->next = m;
     else
 	m_free = m;
 
-    if ((((char *) m_lfree) + M_ISIZE + m_lfree->len) == m_high &&
+    if ((((char *)m_lfree) + M_ISIZE + m_lfree->len) == m_high &&
 	m_lfree->len >= m_pgsz + M_MIN) {
 	long n = (m_lfree->len - M_MIN) & ~(m_pgsz - 1);
 
@@ -775,7 +747,6 @@ int sz;
 	    zerr("allocation error at brk: %e", NULL, 0);
 #endif
 	}
-
 #ifdef MEM_DEBUG
 	m_b += n;
 #endif
@@ -789,10 +760,10 @@ FREE_ARG_T p;
 
 #ifdef FREE_DO_RET
     return 0;
-#endif    
+#endif
 }
 
-void zsfree(p)		/**/
+void zsfree(p)			/**/
 char *p;
 {
     if (p)
@@ -803,17 +774,19 @@ MALLOC_RET_T realloc(p, size)
 MALLOC_RET_T p;
 MALLOC_ARG_T size;
 {
-    struct m_hdr *m = (struct m_hdr *) (((char *) p) - M_ISIZE), *mp, *mt;
+    struct m_hdr *m = (struct m_hdr *)(((char *)p) - M_ISIZE), *mp, *mt;
     char *r;
     int i, l = 0;
 
+    if (!p && size)
+	return (MALLOC_RET_T) malloc(size);
     if (!p || !size)
 	return (MALLOC_RET_T) p;
 
     for (i = 0; i < M_NSMALL; i++) {
-	for (mp = NULL, mt = m_small[i]; 
-	     mt && (((char *) mt) > ((char *) p) || 
-		    (((char *) mt) + mt->len) < ((char *) p));
+	for (mp = NULL, mt = m_small[i];
+	     mt && (((char *)mt) > ((char *)p) ||
+		    (((char *)mt) + mt->len) < ((char *)p));
 	     mp = mt, mt = mt->next);
 
 	if (mt) {
@@ -825,13 +798,12 @@ MALLOC_ARG_T size;
 	l = m->len;
 
     r = malloc(size);
-    memcpy(r, (char *) p, (size > l) ? l : size);
+    memcpy(r, (char *)p, (size > l) ? l : size);
     free(p);
 
     return (MALLOC_RET_T) r;
 }
 
-
 MALLOC_RET_T calloc(n, size)
 MALLOC_ARG_T n;
 MALLOC_ARG_T size;
@@ -839,7 +811,7 @@ MALLOC_ARG_T size;
     long l;
     char *r;
 
-    if(!(l = n * size))
+    if (!(l = n * size))
 	return (MALLOC_RET_T) m_high;
 
     r = malloc(l);
@@ -849,7 +821,6 @@ MALLOC_ARG_T size;
     return (MALLOC_RET_T) r;
 }
 
-
 FREE_RET_T cfree(p)
 FREE_ARG_T p;
 {
@@ -860,10 +831,9 @@ FREE_ARG_T p;
 #endif
 }
 
-
 #ifdef MEM_DEBUG
 
-int bin_mem(name, argv, ops, func)			/**/
+int bin_mem(name, argv, ops, func)	/**/
 char *name;
 char **argv;
 char *ops;
@@ -879,16 +849,14 @@ int func;
 	printf("The lower and the upper addresses of the heap. Diff gives\n");
 	printf("the difference between them, i.e. the size of the heap.\n\n");
     }
-
-    printf("low mem %ld\t high mem %ld\t diff %ld\n", 
-	   (long) m_l, (long) m_high, (long) (m_high - ((char *) m_l)));
+    printf("low mem %ld\t high mem %ld\t diff %ld\n",
+	   (long)m_l, (long)m_high, (long)(m_high - ((char *)m_l)));
 
     if (ops['v']) {
 	printf("\nThe number of bytes that were allocated using sbrk() and\n");
 	printf("the number of bytes that were given back to the system\n");
 	printf("via brk().\n");
     }
-
     printf("\nsbrk %d\tbrk %d\n", m_s, m_b);
 
     if (ops['v']) {
@@ -900,11 +868,10 @@ int func;
 	printf("i.e. the number of bytes that are allocated for blocks of\n");
 	printf("this size.\n");
     }
-
     printf("\nsize\tmalloc\tfree\tdiff\ttotal\n");
     for (i = 0; i < 1024; i++)
 	if (m_m[i] || m_f[i])
-	    printf("%d\t%d\t%d\t%d\t%d\n", i * M_ISIZE, m_m[i], m_f[i], 
+	    printf("%d\t%d\t%d\t%d\t%d\n", i * M_ISIZE, m_m[i], m_f[i],
 		   m_m[i] - m_f[i], i * M_ISIZE * (m_m[i] - m_f[i]));
 
     if (m_m[i] || m_f[i])
@@ -929,10 +896,9 @@ int func;
 	printf("blocks is shown. For otherwise used blocks the first few\n");
 	printf("bytes are shown as an ASCII dump.\n");
     }
-
     printf("\nblock list:\nnum\ttnum\taddr\tlen\tstate\tcum\n");
-    for (m = m_l, mf = m_free, ii = fi = ui = 1; ((char *) m) < m_high; 
-	 m = (struct m_hdr *) (((char *) m) + M_ISIZE + m->len), ii++) {
+    for (m = m_l, mf = m_free, ii = fi = ui = 1; ((char *)m) < m_high;
+	 m = (struct m_hdr *)(((char *)m) + M_ISIZE + m->len), ii++) {
 	for (j = 0, ms = NULL; j < M_NSMALL && !ms; j++)
 	    for (ms = m_small[j]; ms; ms = ms->next)
 		if (ms == m)
@@ -943,16 +909,17 @@ int func;
 	else if (m == ms)
 	    sprintf(buf, "%d %d %d", M_SNUM - ms->used, ms->used,
 		    (m->len - sizeof(struct m_hdr)) / M_SNUM + 1);
+
 	else {
-	    for (i = 0, b = buf, c = (char *) &m->next; i < 20 && i < m->len;
+	    for (i = 0, b = buf, c = (char *)&m->next; i < 20 && i < m->len;
 		 i++, c++)
 		*b++ = (*c >= ' ' && *c < 127) ? *c : '.';
 	    *b = '\0';
 	}
 
-	printf("%d\t%d\t%ld\t%ld\t%s\t%ld\t%s\n", ii, 
+	printf("%d\t%d\t%ld\t%ld\t%s\t%ld\t%s\n", ii,
 	       (m == mf) ? fi++ : ui++,
-	       (long) m, m->len,
+	       (long)m, m->len,
 	       (m == mf) ? "free" : ((m == ms) ? "small" : "used"),
 	       (m == mf) ? (f += m->len) : (u += m->len),
 	       buf);
@@ -966,13 +933,12 @@ int func;
 	printf("For each size the arrays with the number of free and the\n");
 	printf("number of used blocks are shown.\n");
     }
-
     printf("\nsmall blocks:\nsize\tblocks (free/used)\n");
 
     for (i = 0; i < M_NSMALL; i++)
 	if (m_small[i]) {
-	    printf ("%d\t", i*M_ISIZE);
-	    
+	    printf("%d\t", i * M_ISIZE);
+
 	    for (ii = 0, m = m_small[i]; m; m = m->next) {
 		printf("(%d/%d) ", M_SNUM - m->used, m->used);
 		if (!((++ii) & 7))
@@ -980,13 +946,11 @@ int func;
 	    }
 	    putchar('\n');
 	}
-
     if (ops['v']) {
 	printf("\n\nBelow is some information about the allocation\n");
 	printf("behaviour of the zsh heaps. First the number of times\n");
 	printf("pushheap(), popheap(), and freeheap() were called.\n");
     }
-
     printf("\nzsh heaps:\n\n");
 
     printf("push %d\tpop %d\tfree %d\n\n", h_push, h_pop, h_free);
@@ -995,7 +959,6 @@ int func;
 	printf("\nThe next list shows for several sizes the number of times\n");
 	printf("memory of this size were taken from heaps.\n\n");
     }
-
     printf("size\tmalloc\ttotal\n");
     for (i = 0; i < 1024; i++)
 	if (h_m[i])
@@ -1008,7 +971,7 @@ int func;
 
 #endif
 
-#else /* not USE_ZSH_MALLOC */
+#else				/* not USE_ZSH_MALLOC */
 
 void zfree(p, sz)		/**/
 vptr p;
