@@ -145,8 +145,8 @@ void setupparams()
 	IPDEF8("MANPATH", "manpath", &manpath),
 	IPDEF8("WATCH", "watch", &watch),
 	IPDEF8("PSVAR", "psvar", &psvar),
-	{NULL,"PATH",NULL,IFN(colonpathsetfn),IFN(colonpathgetfn),0,
-	     PMFLAG_SPECIAL,(vptr)NULL,NULL,"path",NULL,0},
+	{NULL, "PATH", NULL, IFN(colonpathsetfn), IFN(colonpathgetfn), 0,
+	 PMFLAG_SPECIAL, (vptr) NULL, NULL, "path", NULL, 0},
 
 #define IPDEF9(A,B,C,D) {NULL,A,NULL,IFN(arrvarsetfn),IFN(arrvargetfn),0,\
 		PMFLAG_A|PMFLAG_SPECIAL|C,(vptr)B,NULL,D,NULL,0}
@@ -162,8 +162,8 @@ void setupparams()
 	IPDEF9("*", &pparams, 0, NULL),
 	IPDEF9("@", &pparams, 0, NULL),
 
-	{NULL,"path",NULL,IFN(pathsetfn),IFN(pathgetfn),0,
-	     PMFLAG_A|PMFLAG_SPECIAL,NULL,NULL,"PATH",NULL,0},
+	{NULL, "path", NULL, IFN(pathsetfn), IFN(pathgetfn), 0,
+	 PMFLAG_A | PMFLAG_SPECIAL, NULL, NULL, "PATH", NULL, 0},
 
 	{NULL,}
     };
@@ -189,19 +189,21 @@ int flags;
 	pm = oldpm;
 	pm->ct = 0;
 	oldpm = pm->old;
-	pm->flags = (flags & (PMFLAG_x|PMFLAG_L|PMFLAG_R|PMFLAG_Z|
-			      PMFLAG_l|PMFLAG_u|PMFLAG_r|PMFLAG_t)) |
-		    (pm->flags & (PMFLAG_s|PMFLAG_i|PMFLAG_A|PMFLAG_SPECIAL));
+	pm->flags = (flags & (PMFLAG_x | PMFLAG_L | PMFLAG_R | PMFLAG_Z |
+			      PMFLAG_l | PMFLAG_u | PMFLAG_r | PMFLAG_t)) |
+	    (pm->flags & (PMFLAG_s | PMFLAG_i | PMFLAG_A | PMFLAG_SPECIAL));
 	if (pm->ename) {
 	    Param altpm = (Param) gethnode(pm->ename, paramtab);
-	    if (altpm) altpm->flags &= ~(PMFLAG_UNSET|PMFLAG_x|PMFLAG_L|
-					 PMFLAG_R|PMFLAG_Z|PMFLAG_l|
-					 PMFLAG_u|PMFLAG_r|PMFLAG_t);
+
+	    if (altpm)
+		altpm->flags &= ~(PMFLAG_UNSET | PMFLAG_x | PMFLAG_L |
+				  PMFLAG_R | PMFLAG_Z | PMFLAG_l |
+				  PMFLAG_u | PMFLAG_r | PMFLAG_t);
 	}
     } else {
 	pm = (struct param *)zcalloc(sizeof *pm);
 	if ((pm->old = oldpm))
-	    remhnode(name, paramtab);  /* needed to avoid freeing oldpm */
+	    remhnode(name, paramtab);	/* needed to avoid freeing oldpm */
 	addhnode(ztrdup(name), pm, paramtab, freepm);
     }
     if (isset(ALLEXPORT) && !oldpm)
@@ -222,7 +224,6 @@ int flags;
     return pm;
 }
 
-
 int isident(s)			/**/
 char *s;
 {
@@ -251,7 +252,7 @@ char *s;
 
 char **garr;
 
-long getarg(str, inv, v, a2, w)		/**/
+long getarg(str, inv, v, a2, w)	/**/
 char **str;
 int *inv;
 Value v;
@@ -628,10 +629,10 @@ Value v;
 	    v->a += strlen(s);
 	if (v->b < 0)
 	    v->b += strlen(s);
-	s = (v->a > (int) strlen(s)) ? dupstring("") : dupstring(s + v->a);
+	s = (v->a > (int)strlen(s)) ? dupstring("") : dupstring(s + v->a);
 	if (v->b < v->a)
 	    s[0] = '\0';
-	else if (v->b - v->a < (int) strlen(s))
+	else if (v->b - v->a < (int)strlen(s))
 	    s[v->b - v->a + 1] = '\0';
 	return s;
     }
@@ -792,7 +793,7 @@ char *val;
 	break;
     }
     if ((!v->pm->env && !(v->pm->flags & PMFLAG_x) &&
-	!(isset(ALLEXPORT) && !v->pm->old)) ||
+	 !(isset(ALLEXPORT) && !v->pm->old)) ||
 	(v->pm->flags & PMFLAG_A) || v->pm->ename)
 	return;
     if (pmtype(v->pm) == PMFLAG_i)
@@ -963,11 +964,11 @@ char *val;
 	    createparam(t, PMFLAG_A);
 	*ss = '[';
 	v = NULL;
-    } else{
+    } else {
 	if (!(v = getvalue(&s, 1)))
 	    createparam(t, PMFLAG_s);
 	else if ((v->pm->flags & PMTYPE) == PMFLAG_A &&
-	    !(v->pm->flags & PMFLAG_SPECIAL)) {
+		 !(v->pm->flags & PMFLAG_SPECIAL)) {
 	    unsetparam(t);
 	    createparam(t, PMFLAG_s);
 	    v = NULL;
@@ -1003,7 +1004,7 @@ char **val;
 	if (!(v = getvalue(&s, 1)))
 	    createparam(t, PMFLAG_A);
 	else if ((v->pm->flags & PMTYPE) != PMFLAG_A &&
-	    !(v->pm->flags & PMFLAG_SPECIAL)) {
+		 !(v->pm->flags & PMFLAG_SPECIAL)) {
 	    unsetparam(t);
 	    createparam(t, PMFLAG_A);
 	    v = NULL;
@@ -1127,7 +1128,6 @@ char **x;
     pm->u.arr = x;
 }
 
-
 char **arrgetfn(pm)		/**/
 Param pm;
 {
@@ -1163,6 +1163,7 @@ char *strvargetfn(pm)		/**/
 Param pm;
 {
     char *s = *((char **)pm->data);
+
     if (!s)
 	return "";
     return s;
@@ -1199,7 +1200,11 @@ char *x;
     char ***dptr = (char ***)pm->data;
 
     freearray(*dptr);
-    *dptr = x ? colonfix(x, pm->ename ? pm->nam : NULL) : mkarray(NULL);
+    if (pm->data == (vptr) & mailpath && x && !*x) {
+	zsfree(x);
+	*dptr = mkarray(NULL);
+    } else
+	*dptr = x ? colonfix(x, pm->ename ? pm->nam : NULL) : mkarray(NULL);
 }
 
 void colonpathsetfn(pm, x)	/**/
@@ -1217,7 +1222,7 @@ Param pm;
     return join(*(char ***)pm->data, ':');
 }
 
-char *colonpathgetfn(pm)		/**/
+char *colonpathgetfn(pm)	/**/
 Param pm;
 {
     return join(path, ':');
@@ -1234,6 +1239,7 @@ Param pm;
 char **x;
 {
     char ***dptr = (char ***)pm->data;
+
     if (*dptr != x)
 	freearray(*dptr);
     *dptr = x ? x : mkarray(NULL);
@@ -1255,7 +1261,8 @@ char **x;
 	freearray(path);
     path = x ? x : mkarray(NULL);
     newcmdnamtab();
-    if (x) arrfixenv("PATH", x);
+    if (x)
+	arrfixenv("PATH", x);
 }
 
 long poundgetfn(pm)		/**/
@@ -1306,7 +1313,7 @@ Param pm;
 long gidgetfn(pm)		/**/
 Param pm;
 {
-    return getegid();
+    return getgid();
 }
 
 long egidgetfn(pm)		/**/
@@ -1578,7 +1585,7 @@ void setparams()
     pm = (struct param *)gethnode("SHLVL", paramtab);
     if (!(pm->flags & PMFLAG_x))
 	pm->flags |= PMFLAG_x;
-    sprintf(buf, "%d", (int) ++shlvl);
+    sprintf(buf, "%d", (int)++shlvl);
     pm->env = addenv("SHLVL", buf);
     noerrs = 0;
 }
@@ -1598,7 +1605,7 @@ char *y;
     return z;
 }
 
-void arrfixenv(s, t)	/**/
+void arrfixenv(s, t)		/**/
 char *s;
 char **t;
 {
@@ -1613,7 +1620,8 @@ char **t;
 	}
     if (isset(ALLEXPORT))
 	pm->flags |= PMFLAG_x;
-    if (pm->flags & PMFLAG_x) pm->env = addenv(s, u);
+    if (pm->flags & PMFLAG_x)
+	pm->env = addenv(s, u);
 }
 
 char *replenv(e, value)		/**/
