@@ -105,6 +105,7 @@ List par_event()
 List parse_list()
 {				/**/
     List ret;
+
     tok = ENDINPUT;
     incmdpos = 1;
     yylex();
@@ -211,6 +212,7 @@ Pline par_pline()
     if (!(c = par_cmd()))
 	return NULL;
     if (tok == BAR) {
+	c->flags &= ~CFLAG_EXEC;
 	cmdpush(CS_PIPE);
 	yylex();
 	while (tok == SEPER)
@@ -225,6 +227,7 @@ Pline par_pline()
     } else if (tok == BARAMP) {
 	struct redir *rdr = (struct redir *)allocnode(N_REDIR);
 
+	c->flags &= ~CFLAG_EXEC;
 	rdr->type = MERGEOUT;
 	rdr->fd1 = 2;
 	rdr->fd2 = 1;
@@ -256,6 +259,7 @@ Cmd par_cmd()
     Cmd c;
 
     c = (Cmd) make_cmd();
+    c->lineno = lineno;
     c->args = newlist();
     c->redir = newlist();
     c->vars = newlist();
@@ -350,13 +354,13 @@ void par_for(c)			/**/
 Cmd c;
 {
     struct forcmd *f;
-    int csh = (tok == FOREACH || isset(CSHJUNKIELOOPS));
+    int csh = (tok == FOREACH);
 
     f = (struct forcmd *)make_forcmd();
     c->type = (tok == SELECT) ? CSELECT : CFOR;
     incmdpos = 0;
     yylex();
-    if (tok != STRING)
+    if (tok != STRING || !isident(tokstr))
 	YYERRORV;
     f->name = tokstr;
     incmdpos = 1;
@@ -368,7 +372,7 @@ Cmd c;
 	c->args = par_wordlist();
 	if (tok != SEPER)
 	    YYERRORV;
-    } else if (tok == INPAR) {
+    } else if (tok == INPAR && (csh || isset(CSHJUNKIEPAREN))) {
 	f->inflag = 1;
 	incmdpos = 0;
 	yylex();
@@ -393,7 +397,7 @@ Cmd c;
 	if (tok != OUTBRACE)
 	    YYERRORV;
 	yylex();
-    } else if (csh) {
+    } else if (csh || isset(CSHJUNKIELOOPS)) {
 	f->list = par_list();
 	if (tok != ZEND)
 	    YYERRORV;
@@ -489,10 +493,11 @@ Cmd c;
 	incmdpos = 0;
 	yylex();
     }
-    
-    cc->pats = (char **) alloc((n + 1) * sizeof(char *));
+
+    cc->pats = (char **)alloc((n + 1) * sizeof(char *));
+
     for (pp = cc->pats, no = firstnode(pats); no; incnode(no))
-	*pp++ = (char *) getdata(no);
+	*pp++ = (char *)getdata(no);
     *pp = NULL;
     cc->lists = (List *) alloc((n + 1) * sizeof(List));
     for (ll = cc->lists, no = firstnode(lists); no; incnode(no), ll++)
@@ -536,7 +541,7 @@ Cmd c;
 	    cmdpop();
 	    YYERRORV;
 	}
-	if (tok == INPAR) {
+	if (tok == INPAR && isset(CSHJUNKIEPAREN)) {
 	    yylex();
 	    l = par_list();
 	    if (tok != OUTPAR) {
@@ -564,33 +569,38 @@ Cmd c;
 	    nt++;
 	    incmdpos = 1;
 	    cmdpop();
-	} else if (tok == INBRACE) {
-	    cmdpop();
-	    cmdpush(nc);
-	    yylex();
-	    l = par_list();
-	    if (tok != OUTBRACE) {
+	} else if (isset(CSHJUNKIEPAREN)) {
+	    if (tok == INBRACE) {
+		cmdpop();
+		cmdpush(nc);
+		yylex();
+		l = par_list();
+		if (tok != OUTBRACE) {
+		    cmdpop();
+		    YYERRORV;
+		}
+		addnode(thensl, l);
+		nt++;
+		yylex();
+		incmdpos = 1;
+		if (tok == SEPER)
+		    break;
+		cmdpop();
+	    } else if (isset(NOSHORTLOOPS)) {
 		cmdpop();
 		YYERRORV;
-	    }
-	    addnode(thensl, l);
-	    nt++;
-	    yylex();
-	    incmdpos = 1;
-	    if (tok == SEPER)
+	    } else {
+		cmdpop();
+		cmdpush(nc);
+		addnode(thensl, par_list1());
+		nt++;
+		cmdpop();
+		incmdpos = 1;
 		break;
-	    cmdpop();
-	} else if (isset(NOSHORTLOOPS)) {
-	    cmdpop();
-	    YYERRORV;
+	    }
 	} else {
 	    cmdpop();
-	    cmdpush(nc);
-	    addnode(thensl, par_list1());
-	    nt++;
-	    cmdpop();
-	    incmdpos = 1;
-	    break;
+	    YYERRORV;
 	}
     }
     cmdpop();
@@ -620,8 +630,7 @@ Cmd c;
 	}
 	cmdpop();
     }
-
-    i = (struct ifcmd *) make_ifcmd();
+    i = (struct ifcmd *)make_ifcmd();
     i->ifls = (List *) alloc((ni + 1) * sizeof(List));
     i->thenls = (List *) alloc((nt + 1) * sizeof(List));
 
@@ -630,7 +639,7 @@ Cmd c;
 	    *ll = &dummy_list;
     *ll = NULL;
     for (ll = i->thenls, no = firstnode(thensl); no; incnode(no), ll++)
-	if(!(*ll = (List) getdata(no)))
+	if (!(*ll = (List) getdata(no)))
 	    *ll = &dummy_list;
     *ll = NULL;
 
@@ -650,7 +659,7 @@ Cmd c;
     w = c->u.whilecmd = (struct whilecmd *)make_whilecmd();
     w->cond = (tok == UNTIL);
     yylex();
-    if (tok == INPAR) {
+    if (tok == INPAR && isset(CSHJUNKIEPAREN)) {
 	yylex();
 	w->cont = par_list();
 	if (tok != OUTPAR)
@@ -1009,9 +1018,9 @@ Cond par_cond_2()
     if (tok != STRING)
 	YYERROR;
     s2 = tokstr;
-    incond++;  /* parentheses do globbing */
+    incond++;			/* parentheses do globbing */
     yylex();
-    incond--;  /* parentheses do grouping */
+    incond--;			/* parentheses do grouping */
     if (tok == STRING) {
 	s3 = tokstr;
 	yylex();
@@ -1030,6 +1039,7 @@ Lklist l;
     struct redir *fn = (struct redir *)allocnode(N_REDIR);
     int mergerror = 0;
     int oldcmdpos, oldnc;
+    unsigned char bc = bangchar;
 
     oldcmdpos = incmdpos;
     incmdpos = 0;
@@ -1038,7 +1048,10 @@ Lklist l;
 	nocorrect = 1;
     fn->type = redirtab[tok - OUTANG];
     fn->fd1 = tokfd;
+    if (fn->type == HEREDOC || fn->type == HEREDOCDASH)
+	bangchar = '\0';
     yylex();
+    bangchar = bc;
     if (tok != STRING && tok != ENVSTRING)
 	YYERRORV;
     toks = tokstr;
@@ -1218,7 +1231,7 @@ char *c;
     else if (b[0] == '!' && (b[1] == Equals || b[1] == '=') && !b[2])
 	n->type = COND_STRNEQ;
     else if (b[0] == '-') {
-	if ((t0 = get_cond_num(b+1)) > -1)
+	if ((t0 = get_cond_num(b + 1)) > -1)
 	    n->type = t0 + COND_NT;
 	else
 	    zerr("unrecognized condition: %s", b, 0);
