@@ -30,7 +30,7 @@ int flags;
     Lknode node = firstnode(list);
     int qt;
 
-    /* bits 0 and 1 of flags are flags to filesub (treat as assignment)
+ /* bits 0 and 1 of flags are flags to filesub (treat as assignment)
      * bit 2 is a flag to paramsubst (single word sub)
      */
     while (node) {
@@ -101,7 +101,7 @@ int flags;
     Lknode node = firstnode(list);
     int glb = 1;
 
-    /* bit 0 of flags is to do glob, bit 1 is flag for single substitution */
+ /* bit 0 of flags is to do glob, bit 1 is flag for single substitution */
     badcshglob = 0;
     if (isset(NOGLOBOPT) || !(flags & 1))
 	glb = 0;
@@ -114,7 +114,7 @@ int flags;
 		*str == Tick || *str == Qtick) {
 		Lknode n = prevnode(node);
 
-		/* `...`,$(...) */
+	    /* `...`,$(...) */
 		commsubst(list, node, str, str3,
 			  (*str == Qstring || *str == Qtick || flags > 1));
 		if (errflag)
@@ -295,18 +295,17 @@ int len;
     user[len] = '\0';
     if ((t0 = findname(user)) != -1)
 	ret_val = dupstring(namdirs[t0].dir);
-    else if ((pm = (struct param *) gethnode(user, paramtab)) &&
+    else if ((pm = (struct param *)gethnode(user, paramtab)) &&
 	     !(pm->flags & (PMFLAG_i | PMFLAG_A)) &&
 	     (str = getsparam(user)) && *str == '/') {
 	adduserdir(user, str, 0, 1);
 	ret_val = str;
-    }
-    else if ((pw = getpwnam(user))) {
+    } else if ((pw = getpwnam(user))) {
 	str = xsymlink(pw->pw_dir);
 	adduserdir(user, str, 1, 1);
 	ret_val = dupstring(str);
 	zsfree(str);
-    }     
+    }
     user[len] = sav;
     return ret_val;
 }
@@ -347,21 +346,14 @@ int qt;
 	insnode(l, where, dyncat(str, str3));
 }
 
-void strcatsub(dest, src, cshtilde)	/**/
+void strcatsub(dest, src, globsubst)	/**/
 char *dest;
 char *src;
-int cshtilde;
+int globsubst;
 {
-    if (!cshtilde) {
-	strcat(dest, src);
-	return;
-    }
-    while (*dest)
-	dest++;
-    do {
-	*dest++ = (*src == '~') ? Tilde : (*src == '=') ? Equals : *src;
-    }
-    while (*src++);
+    strcat(dest, src);
+    if (globsubst)
+	tokenize(dest);
 }
 
 int strpcmp(a, b)		/**/
@@ -434,7 +426,7 @@ char *postmul;
     if (lr == ls)
 	return str;
 
-    r = ret = (char *) halloc(lr + 1);
+    r = ret = (char *)halloc(lr + 1);
 
     if (prenum) {
 	if (postnum) {
@@ -565,7 +557,7 @@ Lklist l;
 Lknode n;
 char *aptr;
 char *bptr;
-int qt;
+int qt;				/* if bit 0 set, real quote, else single word substitution */
 int sp;
 {
     char *s = aptr, *u, *idbeg, *idend, *ostr = bptr;
@@ -574,7 +566,7 @@ int sp;
     int doub = 0;		/* != 0 means we have %%, not %, or ##, not # */
     int isarr = 0;
     int plan9 = isset(RCEXPANDPARAM);
-    int cshtilde = isset(CSHJUNKIETILDE);
+    int globsubst = isset(GLOBSUBST);
     int getlen = 0;
     int whichlen = 0;
     int chkset = 0;
@@ -776,13 +768,13 @@ int sp;
 						|| (isstring(s[1]) && isbrace(s[2]) && iident(s[3]))))
 	    getlen = 1 + whichlen, s++;
 	else if (*s == '~' || *s == Tilde)
-	    cshtilde ^= 1, s++;
+	    globsubst ^= 1, s++;
 	else if (*s == '+' && iident(s[1]))
 	    chkset = 1, s++;
 	else
 	    break;
     }
-    cshtilde = cshtilde && !(qt & 1);
+    globsubst = globsubst && !(qt & 1);
 
     idbeg = s;
     if (isstring(*s) && isbrace(s[1])) {
@@ -805,12 +797,13 @@ int sp;
     } else if ((isarr = v->isarr)) {
 	aval = getarrvalue(v);
 	if (qt && ((qt & 1) || !getlen) && isarr > 0) {
-	    val = sepjoin(aval,sep);
+	    val = sepjoin(aval, sep);
 	    isarr = 0;
 	}
     } else {
 	if (v->pm->flags & PMFLAG_A) {
 	    int tmplen = arrlen(v->pm->gets.afn(v->pm));
+
 	    if (v->a < 0)
 		v->a += tmplen + v->inv;
 	    if (!v->inv && (v->a >= tmplen || v->a < 0))
@@ -888,7 +881,7 @@ int sp;
 	if (*s == '%')
 	    flags |= 1;
 
-	if (*s == s[1]) {
+	if ((*s == '%' || *s == '#' || *s == Pound) && *s == s[1]) {
 	    s++;
 	    doub = 1;
 	}
@@ -931,7 +924,7 @@ int sp;
 
 		*idend = '\0';
 		val = dupstring(u);
-		filesub(&val, 3);
+		singsub(&val);
 		setsparam(idbeg, ztrdup(val));
 		*idend = sav;
 		isarr = 0;
@@ -941,6 +934,7 @@ int sp;
 	case (int)STOUC(Quest):
 	    if (vunset) {
 		char *msg;
+
 		*idend = '\0';
 		msg = tricat(idbeg, ": ", *u ? u : "parameter not set");
 		zerr("%s", msg, 0);
@@ -962,6 +956,7 @@ int sp;
 	case (int)STOUC(Pound):
 	    if (qt & 1)
 		tokenize(u);
+
 	    if (!vunset && v && v->isarr) {
 		char **ap = aval;
 		char **pp = aval = (char **)ncalloc(sizeof(char *) * (arrlen(aval) + 1));
@@ -971,6 +966,8 @@ int sp;
 		    getmatch(pp, u, flags, flnum);
 		    pp++;
 		}
+		if (!isarr)
+		    val = sepjoin(aval, sep);
 	    } else {
 		if (vunset)
 		    val = dupstring("");
@@ -1051,18 +1048,21 @@ int sp;
 	val = aval[0];
 	isarr = 0;
     }
-    if (!qt && spbreak) {
+    if (!qt && (spbreak || spsep || sep)) {
 	if (isarr)
 	    val = sepjoin(aval, sep);
-	isarr = 1;
-	aval = sepsplit(val, spsep);
-	if (!aval || !aval[0]) {
-	    val = dupstring("");
+	if (spbreak || spsep) {
+	    isarr = 1;
+	    aval = sepsplit(val, spsep);
+	    if (!aval || !aval[0]) {
+		val = dupstring("");
+		isarr = 0;
+	    } else if (!aval[1]) {
+		val = aval[0];
+		isarr = 0;
+	    }
+	} else
 	    isarr = 0;
-	} else if (!aval[1]) {
-	    val = aval[0];
-	    isarr = 0;
-	}
     }
     if (casmod) {
 	if (isarr) {
@@ -1101,7 +1101,7 @@ int sp;
 	if (!aval[0]) {
 	    if (plan9)
 		return 0;
-	    y = (char *) ncalloc((aptr - bptr) + strlen(s) + 1);
+	    y = (char *)ncalloc((aptr - bptr) + strlen(s) + 1);
 	    strcpy(y, ostr);
 	    strcat(y, s);
 	    remnulargs(y);
@@ -1112,21 +1112,23 @@ int sp;
 		return 1;
 	    }
 	}
-
 	if (sortit && !copied)
 	    aval = arrdup(aval);
 	if (sortit == 1)
-	    qsort(aval, arrlen(aval), sizeof(char *), (int (*)()) strpcmp);
+	    qsort(aval, arrlen(aval), sizeof(char *), (int (*)())strpcmp);
+
 	else if (sortit == 2)
-	    qsort(aval, arrlen(aval), sizeof(char *), (int (*)()) invstrpcmp);
+	    qsort(aval, arrlen(aval), sizeof(char *), (int (*)())invstrpcmp);
+
 	else if (sortit == 3)
-	    qsort(aval, arrlen(aval), sizeof(char *), (int (*)()) cstrpcmp);
+	    qsort(aval, arrlen(aval), sizeof(char *), (int (*)())cstrpcmp);
+
 	else if (sortit)
-	    qsort(aval, arrlen(aval), sizeof(char *), (int (*)()) invcstrpcmp);
+	    qsort(aval, arrlen(aval), sizeof(char *), (int (*)())invcstrpcmp);
+
+	if (plan9) {
+	    int dlen;
 
-  	if (plan9) {
-  	    int dlen;
-  
 	    dlen = (aptr - bptr) + strlen(s) + 1;
 	    i = 0;
 	    while (aval[i]) {
@@ -1139,9 +1141,9 @@ int sp;
 		    xlen = nulstrlen;
 		} else
 		    xlen = strlen(x);
-		y = (char *) ncalloc(dlen + xlen);
+		y = (char *)ncalloc(dlen + xlen);
 		strcpy(y, ostr);
-		strcatsub(y, x, cshtilde);
+		strcatsub(y, x, globsubst);
 		strcat(y, s);
 		if (i == 1)
 		    setdata(n, (vptr) y);
@@ -1151,16 +1153,16 @@ int sp;
 	} else {
 	    x = aval[0];
 	    if (prenum || postnum)
-		x = dopadding(x, prenum, postnum, preone, postone, 
+		x = dopadding(x, prenum, postnum, preone, postone,
 			      premul, postmul);
 	    if (qt && !*x) {
 		x = nulstring;
 		xlen = nulstrlen;
 	    } else
 		xlen = strlen(x);
-	    y = (char *) ncalloc((aptr - bptr) + xlen + 1);
+	    y = (char *)ncalloc((aptr - bptr) + xlen + 1);
 	    strcpy(y, ostr);
-	    strcatsub(y, x, cshtilde);
+	    strcatsub(y, x, globsubst);
 	    setdata(n, (vptr) y);
 
 	    i = 1;
@@ -1171,8 +1173,8 @@ int sp;
 				  premul, postmul);
 		if (qt && !*x)
 		    y = dupstring(nulstring);
-		else if (cshtilde) {
-		    y = (char *) ncalloc(strlen(x) + 1);
+		else if (globsubst) {
+		    y = (char *)ncalloc(strlen(x) + 1);
 		    *y = '\0';
 		    strcatsub(y, x, 1);
 		} else
@@ -1182,14 +1184,14 @@ int sp;
 
 	    x = aval[i];
 	    if (prenum || postnum)
-		x = dopadding(x, prenum, postnum, preone, postone, 
+		x = dopadding(x, prenum, postnum, preone, postone,
 			      premul, postmul);
 	    if (qt && !*x) {
 		x = nulstring;
 		xlen = nulstrlen;
 	    } else
 		xlen = strlen(x);
-	    y = (char *) ncalloc(xlen + strlen(s) + 1);
+	    y = (char *)ncalloc(xlen + strlen(s) + 1);
 	    strcpy(y, x);
 	    strcat(y, s);
 	    insnode(l, n, (vptr) y);
@@ -1208,9 +1210,9 @@ int sp;
 	    xlen = nulstrlen;
 	} else
 	    xlen = strlen(x);
-	y = (char *) ncalloc((aptr - bptr) + xlen + strlen(s) + 1);
+	y = (char *)ncalloc((aptr - bptr) + xlen + strlen(s) + 1);
 	strcpy(y, ostr);
-	strcatsub(y, x, cshtilde);
+	strcatsub(y, x, globsubst);
 	strcat(y, s);
 	setdata(n, (vptr) y);
     }
@@ -1287,8 +1289,12 @@ char **ptr;
 		for (ptr3 = ptr2; *ptr3 != del && *ptr3; ptr3++);
 		if ((sav = *ptr3))
 		    *ptr3++ = '\0';
-		hsubl = ztrdup(ptr1);
-		hsubr = ztrdup(ptr2);
+		for (tt = hsubl = ztrdup(ptr1); *tt; tt++)
+		    if (INULL(*tt))
+			chuck(tt);
+		for (tt = hsubr = ztrdup(ptr2); *tt; tt++)
+		    if (INULL(*tt))
+			chuck(tt);
 		ptr2[-1] = del;
 		if (sav)
 		    ptr3[-1] = sav;
@@ -1388,7 +1394,7 @@ char **ptr;
 		    tc = *tt;
 		    *tt = '\0';
 		    nl = al + strlen(t) + strlen(copy);
-		    ptr1 = tmp = (char *) halloc(nl + 1);
+		    ptr1 = tmp = (char *)halloc(nl + 1);
 		    if (all)
 			for (ptr2 = all; *ptr2;)
 			    *ptr1++ = *ptr2++;
