@@ -76,13 +76,14 @@ char *s;
 
 /* print an error */
 
-void zwarnnam(cmd, fmt, str, num) /**/
+void zwarnnam(cmd, fmt, str, num)	/**/
 char *cmd;
 char *fmt;
 char *str;
 int num;
 {
     int waserr = errflag;
+
     zerrnam(cmd, fmt, str, num);
     errflag = waserr;
 }
@@ -204,7 +205,7 @@ int sig;
 
     struct sigvec vec;
 
-    vec.sv_handler = (SIGVEC_HANDTYPE)handler;
+    vec.sv_handler = (SIGVEC_HANDTYPE) handler;
     vec.sv_mask = sigmask(sig);
     vec.sv_flags = SV_INTERRUPT;
     sigvec(sig, &vec, (struct sigvec *)NULL);
@@ -249,8 +250,7 @@ void nointr()
 
 void holdintr()
 {				/**/
-    if (interact)
-    {
+    if (interact) {
 #ifdef SIGNAL_MASKS
 	sig_block(sig_mask(SIGINT));
 #else
@@ -263,8 +263,7 @@ void holdintr()
 
 void noholdintr()
 {				/**/
-    if (interact)
-    {
+    if (interact) {
 #ifdef SIGNAL_MASKS
 	sig_unblock(sig_mask(SIGINT));
 #else
@@ -299,6 +298,7 @@ int sig;
 
 #if !defined(POSIX) && !defined(SIGNAL_MASKS)
 static sigset_t heldset;
+
 #endif
 
 sigset_t sig_block(set)		/**/
@@ -373,8 +373,7 @@ sigset_t set;
 	if (sigismember(&set, i) && !sigismember(&heldset, i)) {
 	    sigaddset(&heldset, i);
 	    sighold(i);
-	}
-	else if (!sigismember(&set, i) && sigismember(&heldset, i)) {
+	} else if (!sigismember(&set, i) && sigismember(&heldset, i)) {
 	    sigdelset(&heldset, i);
 	    sigrelse(i);
 	}
@@ -385,8 +384,9 @@ sigset_t set;
     return oset;
 }
 
-int sig_suspend(sig)		/**/
+int sig_suspend(sig, sig2)		/**/
 int sig;
+int sig2;
 {
     int ret;
 
@@ -396,12 +396,21 @@ int sig;
     if (sig) {
 	set = sig_notmask(sig);
 	sigdelset(&set, SIGHUP);
-    }
-    else
+	sigdelset(&set, sig2);
+    } else
 	sigemptyset(&set);
     ret = sigsuspend(&set);
 #else
+#ifdef HAS_SIGRELSE
+    /* Use System V version of sigpause.  */
     ret = sigpause(sig);
+#else
+    /* Use BSD version of sigpause.  */
+    sigset_t set = sig_notmask(sig);
+    if (sig2)
+      sigdelset(&set, sig2);
+    ret = sigpause(set);
+#endif
 #endif
 
     return ret;
@@ -624,7 +633,7 @@ char *s;
 
 /* add a named directory */
 
-void adduserdir(s, t, ishomedir, always) /**/
+void adduserdir(s, t, ishomedir, always)	/**/
 char *s;
 char *t;
 int ishomedir;
@@ -650,8 +659,7 @@ int always;
 	}
 	return;
     }
-    if ((!always && !isset(AUTONAMEDIRS)) ||
-	(always == 2 && findname(s) == -1))
+    if (unset(AUTONAMEDIRS) && findname(s) == -1 && always == 2)
 	return;
 
     t2 = strlen(t);
@@ -801,6 +809,7 @@ void preprompt()
 	    schl->next = sch->next;
 	    zsfree(sch->cmd);
 	    zfree(sch, sizeof(struct schedcmd));
+
 	    sch = schl;
 	}
 	if (errflag)
@@ -936,7 +945,8 @@ vptr a;
     if (c->flags & SHFUNC) {
 	if (c->u.list)
 	    freestruct(c->u.list);
-    }
+    } else if ((c->flags & HASHCMD) == HASHCMD)
+	zsfree(c->u.cmd);
 
     zfree(c, sizeof(struct cmdnam));
 }
@@ -1021,7 +1031,7 @@ vptr a;
     zfree(cc, sizeof(struct compctl));
 }
 
-void freecompctlp(a)          /**/
+void freecompctlp(a)		/**/
 vptr a;
 {
     Compctlp ccp = (Compctlp) a;
@@ -1076,7 +1086,7 @@ struct ttyinfo *ti;
 #ifdef TIOCGWINSZ
 /*	if (ioctl(SHTTY, TIOCGWINSZ, &ti->winsize) == -1)
 	    	zerr("bad tiocgwinsz: %e",NULL,errno);*/
-	ioctl(SHTTY, TIOCGWINSZ, (char *) &ti->winsize);
+	ioctl(SHTTY, TIOCGWINSZ, (char *)&ti->winsize);
 #endif
     }
 }
@@ -1091,12 +1101,12 @@ struct ttyinfo *ti;
 #define TCSADRAIN 1		/* XXX Princeton's include files are screwed up */
 #endif
 	tcsetattr(SHTTY, TCSADRAIN, &ti->tio);
-	/* if (tcsetattr(SHTTY, TCSADRAIN, &ti->tio) == -1) */
+    /* if (tcsetattr(SHTTY, TCSADRAIN, &ti->tio) == -1) */
 #else
 	ioctl(SHTTY, TCSETS, &ti->tio);
-	/* if (ioctl(SHTTY, TCSETS, &ti->tio) == -1) */
+    /* if (ioctl(SHTTY, TCSETS, &ti->tio) == -1) */
 #endif
-	    /*	zerr("settyinfo: %e",NULL,errno)*/ ;
+	/*	zerr("settyinfo: %e",NULL,errno)*/ ;
 #else
 #ifdef HAS_TERMIO
 	ioctl(SHTTY, TCSETA, &ti->tio);
@@ -1120,7 +1130,7 @@ void adjustwinsize()
     if (SHTTY == -1)
 	return;
 
-    ioctl(SHTTY, TIOCGWINSZ, (char *) &shttyinfo.winsize);
+    ioctl(SHTTY, TIOCGWINSZ, (char *)&shttyinfo.winsize);
     if (shttyinfo.winsize.ws_col)
 	columns = shttyinfo.winsize.ws_col;
     if (shttyinfo.winsize.ws_row)
@@ -1342,15 +1352,17 @@ char *s;
 
 int getquery()
 {				/**/
-    char c;
-    int val;
+    char c, d;
+    int val, isem = !strcmp(term, "emacs");
 
     attachtty(mypgrp);
-    setcbreak();
+    if (!isem)
+	setcbreak();
 #ifdef FIONREAD
-    ioctl(SHTTY, FIONREAD, (char *) &val);
+    ioctl(SHTTY, FIONREAD, (char *)&val);
     if (val) {
-	settyinfo(&shttyinfo);
+	if (!isem)
+	    settyinfo(&shttyinfo);
 	write(2, "n\n", 2);
 	return 'n';
     }
@@ -1358,9 +1370,14 @@ int getquery()
     if (read(SHTTY, &c, 1) == 1)
 	if (c == 'y' || c == 'Y' || c == '\t')
 	    c = 'y';
-    settyinfo(&shttyinfo);
-    if (c != '\n')
-	write(2, "\n", 1);
+    if (isem) {
+	if (c != '\n')
+	    while (read(SHTTY, &d, 1) == 1 && d != '\n');
+    } else {
+	settyinfo(&shttyinfo);
+	if (c != '\n')
+	    write(2, "\n", 1);
+    }
     return (int)c;
 }
 
@@ -1373,7 +1390,7 @@ char *junk;
 {
     int nd;
 
-    nd = spdist(s, guess, (int) strlen(guess) / 4 + 1);
+    nd = spdist(s, guess, (int)strlen(guess) / 4 + 1);
     if (nd <= d) {
 	best = s;
 	d = nd;
@@ -1445,7 +1462,7 @@ int ask;
     }
     if (errflag)
 	return;
-    if (best && (int) strlen(best) > 1 && strcmp(best, guess)) {
+    if (best && (int)strlen(best) > 1 && strcmp(best, guess)) {
 	if (ask) {
 	    char *pp;
 	    int junk;
@@ -1770,21 +1787,31 @@ char **s;
 char *sep;
 {
     char *r, *p, **t;
-    int l, sl;
+    int l, sl, elide = 0;
     static char *lastmem = NULL;
+    char sepbuf[2];
 
-    if (!sep)
-	return spacejoin(s);
-
+    if (!*s)
+	return "";
+    if (!sep) {
+	elide = 1;
+	sep = sepbuf;
+	sepbuf[0] = *ifs;
+	sepbuf[1] = '\0';
+    }
     sl = strlen(sep);
     for (t = s, l = 1 - sl; *t; l += strlen(*t) + sl, t++);
     if (l == 1)
 	return "";
     zsfree(lastmem);
     lastmem = r = p = (char *)zalloc(l);
-    for (t = s; *t; t++) {
+    t = s;
+    if (elide)
+	while (*t && !**t)
+	    t++;
+    for (; *t; t++) {
 	strucpy(&p, *t);
-	if (t[1])
+	if (t[1] && (!elide || t[1][0]))
 	    strucpy(&p, sep);
     }
     return r;
@@ -1806,7 +1833,7 @@ char *sep;
 
     for (t = s; n--;) {
 	tt = t;
-	f = findsep(&t, sep);/* f set not but not used? ++jhi; */
+	f = findsep(&t, sep);	/* f set not but not used? ++jhi; */
 	*p = (char *)zalloc(t - tt + 1);
 	strncpy(*p, tt, t - tt);
 	(*p)[t - tt] = '\0';
@@ -1843,7 +1870,8 @@ char *nam;
 
 /* allocate a tree element */
 
-static int sizetab[N_COUNT] = {
+static int sizetab[N_COUNT] =
+{
     sizeof(struct list),
     sizeof(struct sublist),
     sizeof(struct pline),
@@ -1854,9 +1882,10 @@ static int sizetab[N_COUNT] = {
     sizeof(struct casecmd),
     sizeof(struct ifcmd),
     sizeof(struct whilecmd),
-    sizeof(struct varasg) };
+    sizeof(struct varasg)};
 
-static int flagtab[N_COUNT] = {
+static int flagtab[N_COUNT] =
+{
     NT_SET(N_LIST, 1, NT_NODE, NT_NODE, 0, 0),
     NT_SET(N_SUBLIST, 2, NT_NODE, NT_NODE, 0, 0),
     NT_SET(N_PLINE, 1, NT_NODE, NT_NODE, 0, 0),
@@ -1867,14 +1896,14 @@ static int flagtab[N_COUNT] = {
     NT_SET(N_CASE, 0, NT_STR | NT_ARR, NT_NODE | NT_ARR, 0, 0),
     NT_SET(N_IF, 0, NT_NODE | NT_ARR, NT_NODE | NT_ARR, 0, 0),
     NT_SET(N_WHILE, 1, NT_NODE, NT_NODE, 0, 0),
-    NT_SET(N_VARASG, 1, NT_STR, NT_STR, NT_STR | NT_LIST, 0) };
+    NT_SET(N_VARASG, 1, NT_STR, NT_STR, NT_STR | NT_LIST, 0)};
 
 vptr allocnode(type)		/**/
 int type;
 {
     struct node *n = (struct node *)alloc(sizetab[type]);
 
-    memset((char *) n, 0, sizetab[type]);
+    memset((char *)n, 0, sizetab[type]);
     n->type = flagtab[type];
     if (useheap)
 	n->type |= NT_HEAP;
@@ -1885,19 +1914,18 @@ int type;
 vptr dupstruct(a)		/**/
 vptr a;
 {
-    struct node *n = (struct node *) a, *r;
+    struct node *n = (struct node *)a, *r;
 
     if (!a || ((List) a) == &dummy_list)
 	return (vptr) a;
 
     if ((n->type & NT_HEAP) && !useheap) {
 	heapalloc();
-	n = (struct node *) dupstruct2((vptr) n);
+	n = (struct node *)dupstruct2((vptr) n);
 	permalloc();
 	n = simplifystruct(n);
     }
-
-    r = (struct node *) dupstruct2((vptr) n);
+    r = (struct node *)dupstruct2((vptr) n);
 
     if (!(n->type & NT_HEAP) && useheap)
 	r = expandstruct(r, N_LIST);
@@ -1916,27 +1944,27 @@ struct node *n;
 	{
 	    List l = (List) n;
 
-	    l->left = (Sublist) simplifystruct((struct node *) l->left);
+	    l->left = (Sublist) simplifystruct((struct node *)l->left);
 	    if (l->type == SYNC && !l->right)
-		return (struct node *) l->left;
+		return (struct node *)l->left;
 	}
 	break;
     case N_SUBLIST:
 	{
 	    Sublist sl = (Sublist) n;
 
-	    sl->left = (Pline) simplifystruct((struct node *) sl->left);
+	    sl->left = (Pline) simplifystruct((struct node *)sl->left);
 	    if (sl->type == END && !sl->flags && !sl->right)
-		return (struct node *) sl->left;
+		return (struct node *)sl->left;
 	}
 	break;
     case N_PLINE:
 	{
 	    Pline pl = (Pline) n;
 
-	    pl->left = (Cmd) simplifystruct((struct node *) pl->left);
+	    pl->left = (Cmd) simplifystruct((struct node *)pl->left);
 	    if (pl->type == END && !pl->right)
-		return (struct node *) pl->left;
+		return (struct node *)pl->left;
 	}
 	break;
     case N_CMD:
@@ -1951,46 +1979,46 @@ struct node *n;
 	    if (empty(c->vars))
 		c->vars = NULL, i++;
 
-	    c->u.list = (List) simplifystruct((struct node *) c->u.list);
+	    c->u.list = (List) simplifystruct((struct node *)c->u.list);
 	    if (i == 3 && !c->flags &&
 		(c->type == CWHILE || c->type == CIF ||
 		 c->type == COND))
-		return (struct node *) c->u.list;
+		return (struct node *)c->u.list;
 	}
 	break;
     case N_FOR:
 	{
-	    struct forcmd *f = (struct forcmd *) n;
+	    struct forcmd *f = (struct forcmd *)n;
 
-	    f->list = (List) simplifystruct((struct node *) f->list);
+	    f->list = (List) simplifystruct((struct node *)f->list);
 	}
 	break;
     case N_CASE:
 	{
-	    struct casecmd *c = (struct casecmd *) n;
+	    struct casecmd *c = (struct casecmd *)n;
 	    List *l;
 
 	    for (l = c->lists; *l; l++)
-		*l = (List) simplifystruct((struct node *) *l);
+		*l = (List) simplifystruct((struct node *)*l);
 	}
 	break;
     case N_IF:
 	{
-	    struct ifcmd *i = (struct ifcmd *) n;
+	    struct ifcmd *i = (struct ifcmd *)n;
 	    List *l;
 
 	    for (l = i->ifls; *l; l++)
-		*l = (List) simplifystruct((struct node *) *l);
+		*l = (List) simplifystruct((struct node *)*l);
 	    for (l = i->thenls; *l; l++)
-		*l = (List) simplifystruct((struct node *) *l);
+		*l = (List) simplifystruct((struct node *)*l);
 	}
 	break;
     case N_WHILE:
 	{
-	    struct whilecmd *w = (struct whilecmd *) n;
+	    struct whilecmd *w = (struct whilecmd *)n;
 
-	    w->cont = (List) simplifystruct((struct node *) w->cont);
-	    w->loop = (List) simplifystruct((struct node *) w->loop); 
+	    w->cont = (List) simplifystruct((struct node *)w->cont);
+	    w->loop = (List) simplifystruct((struct node *)w->loop);
 	}
     }
 
@@ -2012,40 +2040,40 @@ int exp;
 	    {
 		List l;
 
-		m = (struct node *) allocnode(N_LIST);
+		m = (struct node *)allocnode(N_LIST);
 		l = (List) m;
 		l->type = SYNC;
 		l->left = (Sublist) expandstruct(n, N_SUBLIST);
 
-		return (struct node *) l;
+		return (struct node *)l;
 	    }
 	case N_SUBLIST:
 	    {
 		Sublist sl;
 
-		m = (struct node *) allocnode(N_SUBLIST);
+		m = (struct node *)allocnode(N_SUBLIST);
 		sl = (Sublist) m;
 		sl->type = END;
 		sl->left = (Pline) expandstruct(n, N_PLINE);
 
-		return (struct node *) sl;
+		return (struct node *)sl;
 	    }
 	case N_PLINE:
 	    {
 		Pline pl;
 
-		m = (struct node *) allocnode(N_PLINE);
+		m = (struct node *)allocnode(N_PLINE);
 		pl = (Pline) m;
 		pl->type = END;
 		pl->left = (Cmd) expandstruct(n, N_CMD);
 
-		return (struct node *) pl;
+		return (struct node *)pl;
 	    }
 	case N_CMD:
 	    {
 		Cmd c;
 
-		m = (struct node *) allocnode(N_CMD);
+		m = (struct node *)allocnode(N_CMD);
 		c = (Cmd) m;
 		switch (NT_TYPE(n->type)) {
 		case N_WHILE:
@@ -2062,19 +2090,18 @@ int exp;
 		c->vars = newlist();
 		c->redir = newlist();
 
-		return (struct node *) c;
+		return (struct node *)c;
 	    }
 	}
-    }
-    else
+    } else
 	switch (NT_TYPE(n->type)) {
 	case N_LIST:
 	    {
 		List l = (List) n;
-		
-		l->left = (Sublist) expandstruct((struct node *) l->left, 
+
+		l->left = (Sublist) expandstruct((struct node *)l->left,
 						 N_SUBLIST);
-		l->right = (List) expandstruct((struct node *) l->right,
+		l->right = (List) expandstruct((struct node *)l->right,
 					       N_LIST);
 	    }
 	    break;
@@ -2082,9 +2109,9 @@ int exp;
 	    {
 		Sublist sl = (Sublist) n;
 
-		sl->left = (Pline) expandstruct((struct node *) sl->left,
+		sl->left = (Pline) expandstruct((struct node *)sl->left,
 						N_PLINE);
-		sl->right = (Sublist) expandstruct((struct node *) sl->right,
+		sl->right = (Sublist) expandstruct((struct node *)sl->right,
 						   N_SUBLIST);
 	    }
 	    break;
@@ -2092,9 +2119,9 @@ int exp;
 	    {
 		Pline pl = (Pline) n;
 
-		pl->left = (Cmd) expandstruct((struct node *) pl->left,
-					       N_CMD);
-		pl->right = (Pline) expandstruct((struct node *) pl->right,
+		pl->left = (Cmd) expandstruct((struct node *)pl->left,
+					      N_CMD);
+		pl->right = (Pline) expandstruct((struct node *)pl->right,
 						 N_PLINE);
 	    }
 	    break;
@@ -2112,70 +2139,70 @@ int exp;
 		switch (c->type) {
 		case CFOR:
 		case CSELECT:
-		    c->u.list = (List) expandstruct((struct node *) c->u.list,
+		    c->u.list = (List) expandstruct((struct node *)c->u.list,
 						    N_FOR);
 		    break;
 		case CWHILE:
-		    c->u.list = (List) expandstruct((struct node *) c->u.list,
+		    c->u.list = (List) expandstruct((struct node *)c->u.list,
 						    N_WHILE);
 		    break;
 		case CIF:
-		    c->u.list = (List) expandstruct((struct node *) c->u.list,
+		    c->u.list = (List) expandstruct((struct node *)c->u.list,
 						    N_IF);
 		    break;
 		case CCASE:
-		    c->u.list = (List) expandstruct((struct node *) c->u.list,
+		    c->u.list = (List) expandstruct((struct node *)c->u.list,
 						    N_CASE);
 		    break;
 		case COND:
-		    c->u.list = (List) expandstruct((struct node *) c->u.list,
+		    c->u.list = (List) expandstruct((struct node *)c->u.list,
 						    N_COND);
 		    break;
 		case ZCTIME:
-		    c->u.list = (List) expandstruct((struct node *) c->u.list,
+		    c->u.list = (List) expandstruct((struct node *)c->u.list,
 						    N_SUBLIST);
 		    break;
 		default:
-		    c->u.list = (List) expandstruct((struct node *) c->u.list,
+		    c->u.list = (List) expandstruct((struct node *)c->u.list,
 						    N_LIST);
 		}
 	    }
 	    break;
 	case N_FOR:
 	    {
-		struct forcmd *f = (struct forcmd *) n;
+		struct forcmd *f = (struct forcmd *)n;
 
-		f->list = (List) expandstruct((struct node *) f->list,
+		f->list = (List) expandstruct((struct node *)f->list,
 					      N_LIST);
 	    }
 	    break;
 	case N_CASE:
 	    {
-		struct casecmd *c = (struct casecmd *) n;
+		struct casecmd *c = (struct casecmd *)n;
 		List *l;
 
 		for (l = c->lists; *l; l++)
-		    *l = (List) expandstruct((struct node *) *l, N_LIST);
+		    *l = (List) expandstruct((struct node *)*l, N_LIST);
 	    }
 	    break;
 	case N_IF:
 	    {
-		struct ifcmd *i = (struct ifcmd *) n;
+		struct ifcmd *i = (struct ifcmd *)n;
 		List *l;
 
 		for (l = i->ifls; *l; l++)
-		    *l = (List) expandstruct((struct node *) *l, N_LIST);
+		    *l = (List) expandstruct((struct node *)*l, N_LIST);
 		for (l = i->thenls; *l; l++)
-		    *l = (List) expandstruct((struct node *) *l, N_LIST);
+		    *l = (List) expandstruct((struct node *)*l, N_LIST);
 	    }
 	    break;
 	case N_WHILE:
 	    {
-		struct whilecmd *w = (struct whilecmd *) n;
+		struct whilecmd *w = (struct whilecmd *)n;
 
-		w->cont = (List) expandstruct((struct node *) w->cont,
+		w->cont = (List) expandstruct((struct node *)w->cont,
 					      N_LIST);
-		w->loop = (List) expandstruct((struct node *) w->loop,
+		w->loop = (List) expandstruct((struct node *)w->loop,
 					      N_LIST);
 	    }
 	}
@@ -2185,7 +2212,7 @@ int exp;
 
 /* duplicate a syntax tree node of given type, argument number */
 
-vptr dupnode(type, a, argnum)		/**/
+vptr dupnode(type, a, argnum)	/**/
 int type;
 vptr a;
 int argnum;
@@ -2193,33 +2220,31 @@ int argnum;
     if (!a)
 	return NULL;
     switch (NT_N(type, argnum)) {
-	    case NT_NODE:
+    case NT_NODE:
 	return (vptr) dupstruct2(a);
-	    case NT_STR:
+    case NT_STR:
 	return (useheap) ? ((vptr) dupstring(a)) :
 	    ((vptr) ztrdup(a));
-	    case NT_LIST | NT_NODE:
+    case NT_LIST | NT_NODE:
 	if (type & NT_HEAP)
-		    if (useheap)
+	    if (useheap)
 		return (vptr) duplist(a, (VFunc) dupstruct2);
 	    else
 		return (vptr) list2arr(a, (VFunc) dupstruct2);
-		else
-		    if (useheap)
-		return (vptr) arr2list(a, (VFunc) dupstruct2);
-		    else
-		return (vptr) duparray(a, (VFunc) dupstruct2);
-	    case NT_LIST | NT_STR:
+	else if (useheap)
+	    return (vptr) arr2list(a, (VFunc) dupstruct2);
+	else
+	    return (vptr) duparray(a, (VFunc) dupstruct2);
+    case NT_LIST | NT_STR:
 	if (type & NT_HEAP)
-		    if (useheap)
+	    if (useheap)
 		return (vptr) duplist(a, (VFunc) dupstring);
-		    else
+	    else
 		return (vptr) list2arr(a, (VFunc) ztrdup);
-		else
-		    if (useheap)
-		return (vptr) arr2list(a, (VFunc) dupstring);
-		    else
-		return (vptr) duparray(a, (VFunc) ztrdup);
+	else if (useheap)
+	    return (vptr) arr2list(a, (VFunc) dupstring);
+	else
+	    return (vptr) duparray(a, (VFunc) ztrdup);
     case NT_NODE | NT_ARR:
 	return (vptr) duparray(a, (VFunc) dupstruct2);
     case NT_STR | NT_ARR:
@@ -2231,7 +2256,7 @@ int argnum;
 
 /* Free a syntax tree node of given type, argument number */
 
-void freenode(type, a, argnum)		/**/
+void freenode(type, a, argnum)	/**/
 int type;
 vptr a;
 int argnum;
@@ -2241,22 +2266,22 @@ int argnum;
     switch (NT_N(type, argnum)) {
     case NT_NODE:
 	freestruct(a);
-		break;
+	break;
     case NT_STR:
 	zsfree(a);
 	break;
     case NT_LIST | NT_NODE:
-	    case NT_NODE | NT_ARR:
+    case NT_NODE | NT_ARR:
 	{
-	    char **p = (char **) a;
-	    
+	    char **p = (char **)a;
+
 	    while (*p)
 		freestruct(*p++);
 	    free(a);
 	}
-		break;
+	break;
     case NT_LIST | NT_STR:
-	    case NT_STR | NT_ARR:
+    case NT_STR | NT_ARR:
 	freearray(a);
 	break;
     default:
@@ -2269,9 +2294,9 @@ int argnum;
 vptr *dupstruct2(a)		/**/
 vptr a;
 {
-    struct node *n = (struct node *) a, *m;
+    struct node *n = (struct node *)a, *m;
     int type;
-    
+
     if (!n || ((List) n) == &dummy_list)
 	return a;
     type = n->type;
@@ -2281,10 +2306,10 @@ vptr a;
 	m->type |= NT_HEAP;
     switch (NT_TYPE(type)) {
     case N_LIST:
-        {
+	{
 	    List nl = (List) n;
 	    List ml = (List) m;
-	    
+
 	    ml->type = nl->type;
 	    ml->left = (Sublist) dupnode(type, nl->left, 0);
 	    ml->right = (List) dupnode(type, nl->right, 1);
@@ -2318,6 +2343,7 @@ vptr a;
 
 	    mc->type = nc->type;
 	    mc->flags = nc->flags;
+	    mc->lineno = nc->lineno;
 	    mc->args = (Lklist) dupnode(type, nc->args, 0);
 	    mc->u.generic = (vptr) dupnode(type, nc->u.generic, 1);
 	    mc->redir = (Lklist) dupnode(type, nc->redir, 2);
@@ -2332,7 +2358,7 @@ vptr a;
 	    mr->type = nr->type;
 	    mr->fd1 = nr->fd1;
 	    mr->fd2 = nr->fd2;
-	    mr->name = (char *) dupnode(type, nr->name, 0);
+	    mr->name = (char *)dupnode(type, nr->name, 0);
 	}
 	break;
     case N_COND:
@@ -2347,37 +2373,37 @@ vptr a;
 	break;
     case N_FOR:
 	{
-	    struct forcmd *nf = (struct forcmd *) n;
-	    struct forcmd *mf = (struct forcmd *) m;
+	    struct forcmd *nf = (struct forcmd *)n;
+	    struct forcmd *mf = (struct forcmd *)m;
 
 	    mf->inflag = nf->inflag;
-	    mf->name = (char *) dupnode(type, nf->name, 0);
+	    mf->name = (char *)dupnode(type, nf->name, 0);
 	    mf->list = (List) dupnode(type, nf->list, 1);
 	}
 	break;
     case N_CASE:
 	{
-	    struct casecmd *ncc = (struct casecmd *) n;
-	    struct casecmd *mcc = (struct casecmd *) m;
+	    struct casecmd *ncc = (struct casecmd *)n;
+	    struct casecmd *mcc = (struct casecmd *)m;
 
-	    mcc->pats = (char **) dupnode(type, ncc->pats, 0);
+	    mcc->pats = (char **)dupnode(type, ncc->pats, 0);
 	    mcc->lists = (List *) dupnode(type, ncc->lists, 1);
 	}
 	break;
     case N_IF:
 	{
-	    struct ifcmd *nic = (struct ifcmd *) n;
-	    struct ifcmd *mic = (struct ifcmd *) m;
+	    struct ifcmd *nic = (struct ifcmd *)n;
+	    struct ifcmd *mic = (struct ifcmd *)m;
 
 	    mic->ifls = (List *) dupnode(type, nic->ifls, 0);
 	    mic->thenls = (List *) dupnode(type, nic->thenls, 1);
-	    
+
 	}
 	break;
     case N_WHILE:
 	{
-	    struct whilecmd *nwc = (struct whilecmd *) n;
-	    struct whilecmd *mwc = (struct whilecmd *) m;
+	    struct whilecmd *nwc = (struct whilecmd *)n;
+	    struct whilecmd *mwc = (struct whilecmd *)m;
 
 	    mwc->cond = nwc->cond;
 	    mwc->cont = (List) dupnode(type, nwc->cont, 0);
@@ -2390,10 +2416,10 @@ vptr a;
 	    Varasg mva = (Varasg) m;
 
 	    mva->type = nva->type;
-	    mva->name = (char *) dupnode(type, nva->name, 0);
-	    mva->str = (char *) dupnode(type, nva->str, 1);
+	    mva->name = (char *)dupnode(type, nva->name, 0);
+	    mva->str = (char *)dupnode(type, nva->str, 1);
 	    mva->arr = (Lklist) dupnode(type, nva->arr, 2);
-	    }
+	}
 	break;
     }
     return (vptr *) m;
@@ -2413,12 +2439,13 @@ vptr a;
     type = n->type;
     switch (NT_TYPE(type)) {
     case N_LIST:
-        {
+	{
 	    List nl = (List) n;
+
 	    freenode(type, nl->left, 0);
 	    freenode(type, nl->right, 1);
 	}
-		break;
+	break;
     case N_SUBLIST:
 	{
 	    Sublist nsl = (Sublist) n;
@@ -2426,9 +2453,9 @@ vptr a;
 	    freenode(type, nsl->left, 0);
 	    freenode(type, nsl->right, 1);
 	}
-		break;
+	break;
     case N_PLINE:
-		{
+	{
 	    Pline npl = (Pline) n;
 
 	    freenode(type, npl->left, 0);
@@ -2443,8 +2470,8 @@ vptr a;
 	    freenode(type, nc->u.generic, 1);
 	    freenode(type, nc->redir, 2);
 	    freenode(type, nc->vars, 3);
-		}
-		break;
+	}
+	break;
     case N_REDIR:
 	{
 	    Redir nr = (Redir) n;
@@ -2462,7 +2489,7 @@ vptr a;
 	break;
     case N_FOR:
 	{
-	    struct forcmd *nf = (struct forcmd *) n;
+	    struct forcmd *nf = (struct forcmd *)n;
 
 	    freenode(type, nf->name, 0);
 	    freenode(type, nf->list, 1);
@@ -2470,7 +2497,7 @@ vptr a;
 	break;
     case N_CASE:
 	{
-	    struct casecmd* ncc = (struct casecmd *) n;
+	    struct casecmd *ncc = (struct casecmd *)n;
 
 	    freenode(type, ncc->pats, 0);
 	    freenode(type, ncc->lists, 1);
@@ -2478,16 +2505,16 @@ vptr a;
 	break;
     case N_IF:
 	{
-	    struct ifcmd *nic = (struct ifcmd *) n;
+	    struct ifcmd *nic = (struct ifcmd *)n;
 
 	    freenode(type, nic->ifls, 0);
 	    freenode(type, nic->thenls, 1);
-	    
+
 	}
 	break;
     case N_WHILE:
 	{
-	    struct whilecmd *nwc = (struct whilecmd *) n;
+	    struct whilecmd *nwc = (struct whilecmd *)n;
 
 	    freenode(type, nwc->cont, 0);
 	    freenode(type, nwc->loop, 1);
@@ -2501,8 +2528,8 @@ vptr a;
 	    freenode(type, nva->str, 1);
 	    freenode(type, nva->arr, 2);
 	}
-		break;
-	    }
+	break;
+    }
     zfree(n, sizetab[NT_TYPE(type)]);
 }
 
@@ -2523,10 +2550,10 @@ char **duparray(arr, func)	/**/
 char **arr;
 VFunc func;
 {
-    char **ret = (char **) alloc((arrlen(arr) + 1) * sizeof(char *)), **rr;
+    char **ret = (char **)alloc((arrlen(arr) + 1) * sizeof(char *)), **rr;
 
     for (rr = ret; *arr;)
-	*rr++ = (char *) func(*arr++);
+	*rr++ = (char *)func(*arr++);
     *rr = NULL;
 
     return ret;
@@ -2539,10 +2566,10 @@ VFunc func;
     char **arr, **r;
     Lknode n;
 
-    arr = r = (char **) alloc((countnodes(l) + 1) * sizeof(char *));
+    arr = r = (char **)alloc((countnodes(l) + 1) * sizeof(char *));
 
     for (n = firstnode(l); n; incnode(n))
-	*r++ = (char *) func(getdata(n));
+	*r++ = (char *)func(getdata(n));
     *r = NULL;
 
     return arr;
@@ -2722,8 +2749,8 @@ char *mindistbest;
     if (!(dd = opendir(dir)))
 	return mindistd;
     while ((de = readdir(dd))) {
-	nd = spdist(de->d_name, mindistguess, 
-		    (int) strlen(mindistguess) / 4 + 1);
+	nd = spdist(de->d_name, mindistguess,
+		    (int)strlen(mindistguess) / 4 + 1);
 	if (nd <= mindistd) {
 	    strcpy(mindistbest, de->d_name);
 	    mindistd = nd;
@@ -2873,7 +2900,7 @@ long pgrp;
 	if (SHTTY != -1 && tcsetpgrp(SHTTY, pgrp) == -1 && !ep)
 #else
 #if ardent
-	if (SHTTY != -1 && setpgrp() == -1 && !ep) 
+	if (SHTTY != -1 && setpgrp() == -1 && !ep)
 #else
 	int arg = pgrp;
 
@@ -2909,3 +2936,11 @@ long gettygrp()
 #endif
     return arg;
 }
+
+#if defined(SCO)
+void gettimeofday(struct timeval *tv, struct timezone *tz)
+{
+    tv->tv_usec = 0;
+    tv->tv_sec = (long)time((time_t) 0);
+}
+#endif
