@@ -28,7 +28,7 @@
 #endif
 
 static int emacs_cur_bindtab[256], eofchar, eofsent;
-int viins_cur_bindtab[256], ungetok; /* needed in zle_hist */
+int viins_cur_bindtab[256], ungetok;	/* needed in zle_hist */
 
 /* hash table containing the zle multi-character bindings */
 
@@ -44,7 +44,8 @@ void setterm()
 
 #if defined(CLOBBERS_TYPEAHEAD) && defined(FIONREAD)
     int val;
-    ioctl(SHTTY, FIONREAD, (char *) &val);
+
+    ioctl(SHTTY, FIONREAD, (char *)&val);
     if (val)
 	return;
 #endif
@@ -55,7 +56,7 @@ void setterm()
 #ifdef FLUSHO
     shttyinfo.tio.c_lflag &= ~FLUSHO;
 #endif
-#else /* not HAS_TIO */
+#else				/* not HAS_TIO */
     shttyinfo.sgttyb.sg_flags = (shttyinfo.sgttyb.sg_flags & ~CBREAK) | ECHO;
     shttyinfo.lmodes &= ~LFLUSHO;
 #endif
@@ -105,7 +106,7 @@ void setterm()
     ti.tio.c_cc[VMIN] = 1;
     ti.tio.c_cc[VTIME] = 0;
     ti.tio.c_iflag |= (INLCR | ICRNL);
-     /* this line exchanges \n and \r; it's changed back in getkey
+ /* this line exchanges \n and \r; it's changed back in getkey
 	so that the net effect is no change at all inside the shell.
 	This double swap is to allow typeahead in common cases, eg.
 
@@ -131,7 +132,7 @@ void setterm()
 	change to \r :-(
 	*/
 
-#else /* not HAS_TIO */
+#else				/* not HAS_TIO */
     ti.sgttyb.sg_flags = (ti.sgttyb.sg_flags | CBREAK) & ~ECHO & ~XTABS;
     ti.lmodes &= ~LFLUSHO;
     eofchar = ti.tchars.t_eofc;
@@ -143,10 +144,10 @@ void setterm()
 
 #if defined(TTY_NEEDS_DRAINING) && defined(TIOCOUTQ) && defined(HAS_SELECT)
 /* this is mostly stolen from bash's draino() */
-    if (baud) {				/**/
+    if (baud) {			/**/
 	int n = 0;
 
-	while ((ioctl(SHTTY, TIOCOUTQ, (char *) &n) >= 0) && n) {
+	while ((ioctl(SHTTY, TIOCOUTQ, (char *)&n) >= 0) && n) {
 	    struct timeval tv;
 
 	    tv.tv_sec = n / baud;
@@ -188,14 +189,14 @@ char *buf;
 
     FD_ZERO(&f);
     FD_SET(fd, &f);
-    return (select(fd + 1, (SELECT_ARG_2_T) &f, NULL, NULL, NULL) == -1 ? 
+    return (select(fd + 1, (SELECT_ARG_2_T) & f, NULL, NULL, NULL) == -1 ?
 	    EOF : read(fd, buf, n));
 }
 
 #define read    breakread
 #endif
 
-int getkey(keytmout)	/**/
+int getkey(keytmout)		/**/
 int keytmout;
 {
     char cc;
@@ -203,11 +204,14 @@ int keytmout;
     long exp100ths;
     int die = 0, r;
     int old_errno = errno;
+
 #ifdef HAS_SELECT
     fd_set foofd;
+
 #else
 #ifdef HAS_TIO
     struct ttyinfo ti;
+
 #endif
 #endif
 
@@ -229,7 +233,7 @@ int keytmout;
 		expire_tv.tv_usec = (exp100ths % 100) * 10000L;
 		FD_ZERO(&foofd);
 		FD_SET(0, &foofd);
-		if (select(1, (SELECT_ARG_2_T)&foofd, NULL, NULL, &expire_tv) <= 0)
+		if (select(1, (SELECT_ARG_2_T) & foofd, NULL, NULL, &expire_tv) <= 0)
 		    return EOF;
 	    }
 #else
@@ -245,8 +249,10 @@ int keytmout;
 #endif
 #endif
 	}
-        while ((r = read(0, &cc, 1)) != 1) {
+	while ((r = read(0, &cc, 1)) != 1) {
 	    if (r == 0) {
+		if (isset(IGNOREEOF))
+		    continue;
 		stopmsg = 1;
 		zexit(1);
 	    }
@@ -272,8 +278,8 @@ int keytmout;
 		zexit(1);
 	    }
 	}
-	if (cc == '\r')	/* undo the exchange of \n and \r determined by */
-	    cc = '\n';	/* setterm() */
+	if (cc == '\r')		/* undo the exchange of \n and \r determined by */
+	    cc = '\n';		/* setterm() */
 	else if (cc == '\n')
 	    cc = '\r';
 
@@ -299,10 +305,12 @@ int p2len;
     int z;
     unsigned char *s;
     int old_errno = errno;
+
 #ifdef HAS_SELECT
-    long costmult = (baud) ? 3840000L /  baud : 0;
+    long costmult = (baud) ? 3840000L / baud : 0;
     struct timeval tv;
     fd_set foofd;
+
     tv.tv_sec = 0;
 #endif
 
@@ -396,16 +404,16 @@ int p2len;
 	    break;
 	}
 #ifdef HAS_SELECT
-	if (baud) {
+	if (baud && (!lastcmd & ZLE_MENUCMP)) {
 	    FD_SET(0, &foofd);
 	    if ((tv.tv_usec = cost * costmult) > 500000)
 		tv.tv_usec = 500000;
-	    if (!kungetct && select(1, (SELECT_ARG_2_T)&foofd, NULL, NULL, &tv) <= 0)
+	    if (!kungetct && select(1, (SELECT_ARG_2_T) & foofd, NULL, NULL, &tv) <= 0)
 		refresh();
 	} else
 #endif
-	    if (!kungetct)
-		refresh();
+	if (!kungetct)
+	    refresh();
     }
     statusline = NULL;
     trashzle();
@@ -420,15 +428,16 @@ int p2len;
 	free(line);
 	line = NULL;
     }
-    free(lastline);	/* freeundo */
+    free(lastline);		/* freeundo */
     errno = old_errno;
     return line;
 }
 
-int getkeycmd()		/**/
-{
+int getkeycmd()
+{				/**/
     int ret;
     static int hops = 0;
+
     cky = NULL;
 
     if ((c = getkey(0)) < 0)
@@ -445,11 +454,11 @@ int getkeycmd()		/**/
 	    cky = NULL;
 	else
 	    lastlen = 1;
-	if(!c)
-	    buf[0] = (char) 0x80;
+	if (!c)
+	    buf[0] = (char)0x80;
 	for (;;) {
 	    if ((c = getkey(cky ? 1 : 0)) >= 0) {
-		if (t0 == buflen -1)
+		if (t0 == buflen - 1)
 		    buf = (char *)realloc(buf, buflen *= 2);
 		buf[t0++] = (c) ? c : 0x80;
 		buf[t0] = '\0';
@@ -475,7 +484,7 @@ int getkeycmd()		/**/
 	}
 	zfree(buf, buflen);
     } else
-	ungetok = 1; /* for doisearch() */
+	ungetok = 1;		/* for doisearch() */
     if (ret == z_vidigitorbeginningofline)
 	ret = (lastcmd & ZLE_ARG) ? z_digitargument : z_beginningofline;
     else if (ret == z_executenamedcmd && !statusline)
@@ -543,8 +552,7 @@ void initxbindtab()
     if (vi) {
 	mainbindtab = viins_cur_bindtab;
 	xbindtab = vi_xbindtab;
-    }
-    else {
+    } else {
 	mainbindtab = emacs_cur_bindtab;
 	xbindtab = em_xbindtab;
     }
@@ -558,40 +566,40 @@ void initxbindtab()
     if (!kungetbuf)
 	kungetbuf = (char *)zalloc(kungetsz = 32);
 
-    addhnode("\33\133", ky = makefunckey(z_undefinedkey), em_xbindtab, 0);
+    addhnode(ztrdup("\33\133"), ky = makefunckey(z_undefinedkey), em_xbindtab, 0);
     ky->prefixct = 4;
-    addhnode("\33\133C", makefunckey(z_forwardchar), em_xbindtab, 0);
-    addhnode("\33\133D", makefunckey(z_backwardchar), em_xbindtab, 0);
-    addhnode("\33\133A", makefunckey(z_uplineorhistory), em_xbindtab, 0);
-    addhnode("\33\133B", makefunckey(z_downlineorhistory), em_xbindtab, 0);
-    addhnode("\33", ky = makefunckey(z_vicmdmode), vi_xbindtab, 0);
+    addhnode(ztrdup("\33\133C"), makefunckey(z_forwardchar), em_xbindtab, 0);
+    addhnode(ztrdup("\33\133D"), makefunckey(z_backwardchar), em_xbindtab, 0);
+    addhnode(ztrdup("\33\133A"), makefunckey(z_uplineorhistory), em_xbindtab, 0);
+    addhnode(ztrdup("\33\133B"), makefunckey(z_downlineorhistory), em_xbindtab, 0);
+    addhnode(ztrdup("\33"), ky = makefunckey(z_vicmdmode), vi_xbindtab, 0);
     ky->prefixct = 4;
-    addhnode("\33\133", ky = makefunckey(z_undefinedkey), vi_xbindtab, 0);
+    addhnode(ztrdup("\33\133"), ky = makefunckey(z_undefinedkey), vi_xbindtab, 0);
     ky->prefixct = 4;
-    addhnode("\33\133C", makefunckey(z_forwardchar), vi_xbindtab, 0);
-    addhnode("\33\133D", makefunckey(z_backwardchar), vi_xbindtab, 0);
-    addhnode("\33\133A", makefunckey(z_uplineorhistory), vi_xbindtab, 0);
-    addhnode("\33\133B", makefunckey(z_downlineorhistory), vi_xbindtab, 0);
-    addhnode("\30", ky = makefunckey(z_undefinedkey), em_xbindtab, 0);
+    addhnode(ztrdup("\33\133C"), makefunckey(z_forwardchar), vi_xbindtab, 0);
+    addhnode(ztrdup("\33\133D"), makefunckey(z_backwardchar), vi_xbindtab, 0);
+    addhnode(ztrdup("\33\133A"), makefunckey(z_uplineorhistory), vi_xbindtab, 0);
+    addhnode(ztrdup("\33\133B"), makefunckey(z_downlineorhistory), vi_xbindtab, 0);
+    addhnode(ztrdup("\30"), ky = makefunckey(z_undefinedkey), em_xbindtab, 0);
     ky->prefixct = 15;
-    addhnode("\30*", makefunckey(z_expandword), em_xbindtab, (FFunc) 0);
-    addhnode("\30g", makefunckey(z_listexpand), em_xbindtab, (FFunc) 0);
-    addhnode("\30G", makefunckey(z_listexpand), em_xbindtab, (FFunc) 0);
-    addhnode("\30\16", makefunckey(z_infernexthistory), em_xbindtab, 0);
-    addhnode("\30\13", makefunckey(z_killbuffer), em_xbindtab, (FFunc) 0);
-    addhnode("\30\6", makefunckey(z_vifindnextchar), em_xbindtab, (FFunc) 0);
-    addhnode("\30\17", makefunckey(z_overwritemode), em_xbindtab, (FFunc) 0);
-    addhnode("\30\25", makefunckey(z_undo), em_xbindtab, (FFunc) 0);
-    addhnode("\30\26", makefunckey(z_vicmdmode), em_xbindtab, (FFunc) 0);
-    addhnode("\30\12", makefunckey(z_vijoin), em_xbindtab, (FFunc) 0);
-    addhnode("\30\2", makefunckey(z_vimatchbracket), em_xbindtab, (FFunc) 0);
-    addhnode("\30s", makefunckey(z_historyincrementalsearchforward),
-             em_xbindtab, (FFunc) 0);
-    addhnode("\30r", makefunckey(z_historyincrementalsearchbackward),
-             em_xbindtab, (FFunc) 0);
-    addhnode("\30u", makefunckey(z_undo), em_xbindtab, (FFunc) 0);
-    addhnode("\30\30", makefunckey(z_exchangepointandmark), em_xbindtab, 0);
-    addhnode("\33", ky = makefunckey(z_undefinedkey), em_xbindtab, 0);
+    addhnode(ztrdup("\30*"), makefunckey(z_expandword), em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30g"), makefunckey(z_listexpand), em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30G"), makefunckey(z_listexpand), em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30\16"), makefunckey(z_infernexthistory), em_xbindtab, 0);
+    addhnode(ztrdup("\30\13"), makefunckey(z_killbuffer), em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30\6"), makefunckey(z_vifindnextchar), em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30\17"), makefunckey(z_overwritemode), em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30\25"), makefunckey(z_undo), em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30\26"), makefunckey(z_vicmdmode), em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30\12"), makefunckey(z_vijoin), em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30\2"), makefunckey(z_vimatchbracket), em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30s"), makefunckey(z_historyincrementalsearchforward),
+	     em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30r"), makefunckey(z_historyincrementalsearchbackward),
+	     em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30u"), makefunckey(z_undo), em_xbindtab, (FFunc) 0);
+    addhnode(ztrdup("\30\30"), makefunckey(z_exchangepointandmark), em_xbindtab, 0);
+    addhnode(ztrdup("\33"), ky = makefunckey(z_undefinedkey), em_xbindtab, 0);
     ky->prefixct = 4;
 
     strcpy(buf, "\33q");
@@ -619,56 +627,56 @@ int *misc;
     for (; *s; s++) {
 	if (*s == '\\' && s[1])
 	    switch (*++s) {
-	      case 'a':
+	    case 'a':
 #ifdef __STDC__
 		*t++ = '\a';
 #else
 		*t++ = '\07';
 #endif
 		break;
-	      case 'n':
+	    case 'n':
 		*t++ = '\n';
 		break;
-	      case 'b':
+	    case 'b':
 		*t++ = '\010';
 		break;
-	      case 't':
+	    case 't':
 		*t++ = '\t';
 		break;
-	      case 'v':
+	    case 'v':
 		*t++ = '\v';
 		break;
-	      case 'f':
+	    case 'f':
 		*t++ = '\f';
 		break;
-	      case 'r':
+	    case 'r':
 		*t++ = '\r';
 		break;
-	      case 'E':
+	    case 'E':
 		if (!fromwhere) {
-		    *t++ ='\\', s--;
+		    *t++ = '\\', s--;
 		    continue;
 		}
-	      case 'e':
+	    case 'e':
 		*t++ = '\033';
 		break;
-	      case 'M':
+	    case 'M':
 		if (fromwhere) {
 		    if (s[1] == '-')
 			s++;
-		    meta = 1 + control; /* preserve the order of ^ and meta */
+		    meta = 1 + control;	/* preserve the order of ^ and meta */
 		} else
-		    *t++ ='\\', s--;
+		    *t++ = '\\', s--;
 		continue;
-	      case 'C':
+	    case 'C':
 		if (fromwhere) {
 		    if (s[1] == '-')
 			s++;
 		    control = 1;
 		} else
-		    *t++ ='\\', s--;
+		    *t++ = '\\', s--;
 		continue;
-	      case 'c':
+	    case 'c':
 		if (fromwhere < 2) {
 		    *misc = 1;
 		    break;
@@ -696,7 +704,7 @@ int *misc;
 		    s--;
 		} else {
 		    if (!fromwhere && *s != '\\')
-			*t++ ='\\';
+			*t++ = '\\';
 		    *t++ = *s;
 		}
 		break;
@@ -795,8 +803,7 @@ int junc;
 	if (ops['e']) {
 	    mainbindtab = emacs_cur_bindtab;
 	    xbindtab = em_xbindtab;
-	}
-	else if (ops['v']) {
+	} else if (ops['v']) {
 	    mainbindtab = viins_cur_bindtab;
 	    xbindtab = vi_xbindtab;
 	}
@@ -806,7 +813,6 @@ int junc;
 		    mainbindtab[i] = emacsbind[i];
 	return 0;
     }
-
     tab = (ops['a']) ? altbindtab : mainbindtab;
     if (!*argv) {
 	char buf[2];
@@ -828,7 +834,6 @@ int junc;
 	listhtable(xbindtab, (HFunc) printbinding);
 	return 0;
     }
-
     while (*argv) {
 	Key ky = NULL, cur = NULL;
 	char *s;
@@ -842,15 +847,14 @@ int junc;
 		firstzero = 1;
 	    for (i = 0; i < len; i++)
 		if (!s[i])
-		    s[i] = (char) 0x80;
+		    s[i] = (char)0x80;
 	}
-
 	if (!*argv || ops['r']) {
 	    if (len == 1)
 		func = tab[(unsigned char)*s];
 	    else
 		func = (ky = (Key) gethnode(s, xbindtab)) ? ky->func
-			    : z_undefinedkey;
+		    : z_undefinedkey;
 	    if (func == z_undefinedkey) {
 		zerrnam(name, "in-string is not bound", NULL, 0);
 		zfree(s, len);
@@ -897,7 +901,6 @@ int junc;
 	    zfree(s, len);
 	    return 0;
 	}
-
 	if (!ops['s']) {
 	    for (i = 0; i != ZLECMDCOUNT; i++)
 		if (!strcmp(*argv, zlecmds[i].name))
@@ -919,28 +922,28 @@ int junc;
 		free(remhnode(s, xbindtab));
 	    tab[(unsigned char)*s] = func;
 	} else {
-	    if (!  (cur = (Key) gethnode(s,xbindtab))
+	    if (!(cur = (Key) gethnode(s, xbindtab))
 		|| (cur->func == z_undefinedkey))
-                for (i = len - 1; i > 0; i--) {
-                    char sav;
+		for (i = len - 1; i > 0; i--) {
+		    char sav;
 
-                    sav = s[i];
-                    s[i] = '\0';
+		    sav = s[i];
+		    s[i] = '\0';
 		    if (i == 1 && firstzero)
 			*s = '\0';
 		    if (!(ky = (Key) gethnode(s, xbindtab)))
 			addhnode(ztrdup(s), ky = makefunckey(z_undefinedkey),
 				 xbindtab, freekey);
 		    ky->prefixct++;
-                    s[i] = sav;
+		    s[i] = sav;
 		    if (i == 1 && firstzero)
-			*s = (char) 0x80;
+			*s = (char)0x80;
 		}
 	    if (cur) {
 		cur->func = func;
 		zfree(cur->str, cur->len);
 	    } else
-		addhnode(ztrdup(s), cur = makefunckey(func), xbindtab,freekey);
+		addhnode(ztrdup(s), cur = makefunckey(func), xbindtab, freekey);
 	    if (firstzero)
 		*s = 0;
 	    if (tab[(unsigned char)*s] != z_sequenceleadin) {
