@@ -55,7 +55,7 @@ void selfinsertunmeta()
 {				/**/
     c &= 0x7f;
     if (c == '\r')
-        c = '\n';
+	c = '\n';
     selfinsert();
 }
 
@@ -432,10 +432,10 @@ void copyprevword()
 }
 
 void sendbreak()
-{                             /**/
+{				/**/
     errflag = 1;
 }
- 
+
 void undo()
 {				/**/
     char *s;
@@ -525,33 +525,33 @@ int executenamedcommand()
     len = 0;
     statusline = buf;
     refresh();
-    for (; ; refresh()) {
+    for (;; refresh()) {
 	if ((cmd = getkeycmd()) < 0 || cmd == z_sendbreak) {
 	    statusline = NULL;
 	    return z_undefinedkey;
 	}
 	switch (cmd) {
-	  case z_backwarddeletechar:
-	  case z_vibackwarddeletechar:
+	case z_backwarddeletechar:
+	case z_vibackwarddeletechar:
 	    if (len) {
 		len--;
 		*--ptr = '\0';
 	    }
 	    break;
-	  case z_killregion:
-	  case z_backwardkillword:
-	  case z_vibackwardkillword:
+	case z_killregion:
+	case z_backwardkillword:
+	case z_vibackwardkillword:
 	    while (len && (len--, *--ptr != '-'))
 		*ptr = '\0';
 	    break;
-	  case z_killwholeline:
-	  case z_vikillline:
-	  case z_backwardkillline:
+	case z_killwholeline:
+	case z_vikillline:
+	case z_backwardkillline:
 	    len = 0;
 	    ptr = buf + 9;
 	    *ptr = '\0';
 	    break;
-	  case z_acceptline:
+	case z_acceptline:
 	    for (t0 = 0; t0 != ZLECMDCOUNT; t0++)
 		if (!strcmp(buf + 9, zlecmds[t0].name))
 		    break;
@@ -563,7 +563,7 @@ int executenamedcommand()
 		feep();
 		break;
 	    }
-	  default:
+	default:
 	    if (cmd == z_listchoices || cmd == z_deletecharorlist ||
 		c == ' ' || c == '\t') {
 		Lklist cmdll;
@@ -686,24 +686,25 @@ int d;
     return 0;
 }
 
-void tstradd(X)	/**/
+void tstradd(X)			/**/
 char *X;
 {
     int t0;
 
     if (termok && unset(SINGLELINEZLE)) {
 #ifdef _IBMR2
-	/* AIX tgetstr() ignores second argument */
+    /* AIX tgetstr() ignores second argument */
 	char *tbuf;
 
 	if (tbuf = tgetstr(X, &tbuf))
 #else
 	char tbuf[2048], *tptr = tbuf;
+
 	if (tgetstr(X, &tptr))
 #endif
 	    tputs(tbuf, 1, putstr);
 	if (*X == 's' && (X[1] == 'o' || X[1] == 'e') &&
-	    (t0 = tgetnum("sg")) > - 1)
+	    (t0 = tgetnum("sg")) > -1)
 	    lensb -= t0;
     }
 }
@@ -744,18 +745,17 @@ int isspell;
     clearerr(stdin);
 
     putpromptchar(isspell == -1 ? 0 : isspell, 1, '\0');
-  
-    if ( isspell != -1 ){
-      *lenp = (bp - bl0) - lensb;
-      *lenp %= columns;
-      if (*lenp == columns - 1) {
-        *lenp = 0;
-        *bp++ = ' ';
-        }
-     }
-    else
-    *lenp = (bp - bl0);
-      *bp = '\0';
+
+    if (isspell != -1) {
+	*lenp = (bp - bl0) - lensb;
+	*lenp %= columns;
+	if (*lenp == columns - 1) {
+	    *lenp = 0;
+	    *bp++ = ' ';
+	}
+    } else
+	*lenp = (bp - bl0);
+    *bp = '\0';
 
     *bp = '\0';
     return buf;
@@ -1057,7 +1057,7 @@ int endchar;
 		    stradd("()");
 		break;
 	    case '?':
-		sprintf(bp, "%ld", (long) lastval);
+		sprintf(bp, "%ld", (long)lastval);
 		bp += strlen(bp);
 		break;
 	    case '%':
@@ -1073,7 +1073,7 @@ int endchar;
 	       sizeof(buf), 220 is from the overflow test made at the
 	       top of the loop, and 1 is for the \0 byte at the end. */
 
-		if (arrlen(psvar) >= arg && (int) strlen(psvar[arg - 1]) < 35)
+		if (arrlen(psvar) >= arg && (int)strlen(psvar[arg - 1]) < 35)
 		    stradd(psvar[arg - 1]);
 		else
 		    stradd("");
