@@ -22,7 +22,7 @@
 #include "zsh.h"
 
 char **obuf = NULL, **nbuf = NULL;
-static int olnct, vcs, vln, vmaxln, winw, winh, winpos;
+static int olnct, vcs, vln, vmaxln, winw, winh, winpos, ovln;
 
 void resetvideo()
 {				/**/
@@ -97,6 +97,7 @@ int scrollwindow()
 
 #ifdef TIOCGWINSZ
 int winchanged;
+
 #endif
 
 int hasam;
@@ -107,7 +108,7 @@ void refresh()
 {				/**/
     unsigned char *s, *t, *sen, *scs = line + cs;
     char **qbuf;
-    int ln = 0, nvcs = 0, nvln = -1, t0 = -1, put_rprompt;
+    int ln = 0, nvcs = 0, nvln = -1, t0 = -1, put_rprompt, res = 0;
 
 #ifdef HAS_SELECT
     cost = 0;
@@ -124,6 +125,11 @@ void refresh()
 	resetvideo();
 	resetneeded = 0;
 	oput_rprompt = 0;
+	if (!clearflag)
+	    if (tccan(TCCLEAREOD))
+		tcout(TCCLEAREOD);
+	    else
+		res = 1;
 	if (t0 > -1)
 	    olnct = t0;
 	if (isset(SINGLELINEZLE) || !termok)
@@ -162,7 +168,8 @@ void refresh()
 	    } else if ((char)*t == '\t') {
 		t0 = (char *)s - nbuf[ln];
 		if (t == scs) {
-		    nvln = ln; scs = NULL;
+		    nvln = ln;
+		    scs = NULL;
 		    if ((nvcs = t0) == columns)
 			nvcs = 0, nvln++;
 		}
@@ -173,7 +180,8 @@ void refresh()
 			    *s++ = ' ';	/* make tab in first column visible */
 		} else {
 		    if (t0 == winw) {
-			t0 = 0; nextline;
+			t0 = 0;
+			nextline;
 		    }
 		    do
 			*s++ = ' ';
@@ -185,7 +193,7 @@ void refresh()
 		*s++ = '^';
 		if (t == scs)
 		    nvcs = s - (unsigned char *)(nbuf[nvln = ln]) - 1,
-		    scs = NULL;
+			scs = NULL;
 		if (s == sen)
 		    nextline;
 		*s++ = (*t == 127) ? '?' : (*t | '@');
@@ -206,7 +214,7 @@ void refresh()
 
     if (statusline) {
 	if (!nbuf[(nlnct == winh) ? winh - 1 : nlnct++])
-	    nbuf[nlnct -1] = (char*)zalloc(winw + 1);
+	    nbuf[nlnct - 1] = (char *)zalloc(winw + 1);
 	s = (unsigned char *)nbuf[nlnct - 1];
 	t = (unsigned char *)statusline;
 	sen = (unsigned char *)(*nbuf + winw);
@@ -226,7 +234,6 @@ void refresh()
 	}
 	*s = '\0';
     }
-
     for (ln = nlnct; ln < winh; ln++)
 	zfree(nbuf[ln], winw + 1), nbuf[ln] = NULL;
 
@@ -238,7 +245,6 @@ void refresh()
 	    nbuf[0][t0] = ' ';
 	nbuf[0][t0] = '\0';
     }
-
     for (ln = 0; ln < nlnct; ln++) {
 
     /* if old line and new line are different,
@@ -247,7 +253,7 @@ void refresh()
 	if (ln < olnct - 1 && !(hasam && vcs == columns) &&
 	    nbuf[ln] && obuf[ln] &&
 	    strncmp(nbuf[ln], obuf[ln], 16)) {
-	    if (tccan(TCDELLINE) && obuf[ln + 1] && obuf[ln + 1][0] && 
+	    if (tccan(TCDELLINE) && obuf[ln + 1] && obuf[ln + 1][0] &&
 		nbuf[ln] && !strncmp(nbuf[ln], obuf[ln + 1], 16)) {
 		moveto(ln, 0);
 		tcout(TCDELLINE);
@@ -270,7 +276,12 @@ void refresh()
 		olnct++;
 	    }
 	}
-	refreshline(ln, put_rprompt);
+	if (res && tccan(TCCLEAREOL)) {
+	    moveto(ln, 0);
+	    tcout(TCCLEAREOL);
+	    refreshline(ln, put_rprompt, 0);
+	} else
+	    refreshline(ln, put_rprompt, res);
 	if (!ln && put_rprompt && !oput_rprompt) {
 	    moveto(0, winw - 1 - ppt2len);
 	    fputs(pmpt2, stdout);
@@ -287,16 +298,24 @@ void refresh()
 	    tcout(TCCLEAREOD);
 	} else
 	    for (ln = nlnct; ln < olnct; ln++)
-		refreshline(ln, put_rprompt);
+		if (res && tccan(TCCLEAREOL)) {
+		    moveto(ln, 0);
+		    tcout(TCCLEAREOL);
+		    refreshline(ln, put_rprompt, 0);
+		} else
+		    refreshline(ln, put_rprompt, res);
 
 /* move to the new cursor position */
 
     moveto(nvln, nvcs);
 
-    if (isset(ALWAYSLASTPROMPT) && !nvcs &&
+    if (isset(ALWAYSLASTPROMPT) &&
+	(!nvcs || (nvln == ovln + 1)) &&
+	cs == ll &&
 	tccan(TCCLEAREOL))
 	tcout(TCCLEAREOL);
 
+    ovln = nvln;
     qbuf = nbuf;
     nbuf = obuf;
     obuf = qbuf;
@@ -314,9 +333,10 @@ void refresh()
 #define tc_upcurs(X) tcmultout(TCUP,TCMULTUP,(X))
 #define tc_leftcurs(X) tcmultout(TCLEFT,TCMULTLEFT,(X))
 
-void refreshline(ln, put_rprompt)		/**/
+void refreshline(ln, put_rprompt, res)	/**/
 int ln;
 int put_rprompt;
+int res;
 {
 /* the test in nl below is to prevent a segv if the terminal cannot clear
    either to end of line or display, and nlnct < olnct :-( */
@@ -324,8 +344,16 @@ int put_rprompt;
     char *nl = nbuf[ln] ? nbuf[ln] : obuf[0], *ol = obuf[ln] ? obuf[ln] : "", *p1;
     int ccs = 0;
 
+    if (res) {
+	char *p = hcalloc(winw + 1);
+
+	memset(p, ' ', winw);
+	strcpy(p, nl);
+	p[strlen(p)] = ' ';
+	nl = p;
+    }
     if (hasam && vcs == columns) {	/* must always write another char */
-	if (*nl) {			/* after writing in last column */ 
+	if (*nl) {		/* after writing in last column */
 	    putchar(*nl);
 	    nl++, vcs = ccs = 1;
 	    if (*ol)
@@ -341,8 +369,10 @@ int put_rprompt;
 	if (!*nl && !*ol)
 	    if (!ln && !put_rprompt && oput_rprompt)
 		if (tccan(TCCLEAREOL)) {
-		    moveto(0, ccs);
-		    tcout(TCCLEAREOL);
+		    if (ccs < columns) {
+			moveto(0, ccs);
+			tcout(TCCLEAREOL);
+		    }
 		    return;
 		} else {
 		    int x = winw - 1 - ccs;
@@ -351,11 +381,10 @@ int put_rprompt;
 		    while (x--)
 			*p1++ = ' ';
 		    *p1 = '\0';
-		}
-	    else {
+	    } else {
 		if (hasam && ccs == columns && ln < nlnct - 1 &&
 		    ln < olnct - 1 && *nbuf[ln + 1] &&
-		    !*obuf[ln + 1]) {		/* force join of lines */
+		    !*obuf[ln + 1]) {	/* force join of lines */
 		    moveto(ln, ccs - 1);
 		    putchar(nl[-1]);
 #ifdef HAS_SELECT
@@ -365,7 +394,6 @@ int put_rprompt;
 		}
 		return;
 	    }
-
     /* if this is the end of the new buffer but the old buffer has stuff
        here, clear to end of line if we can, otherwise fill the new buffer
        with blanks and continue. */
@@ -391,7 +419,7 @@ int put_rprompt;
        new buffer. */
 
 	if (!*ol) {
-	    while (*nl == ' ')
+	    while (!res && *nl == ' ')
 		nl++, ccs++;
 	    if (*nl) {
 		moveto(ln, ccs);
@@ -400,16 +428,21 @@ int put_rprompt;
 		cost += strlen(nl);
 #endif
 		vcs += strlen(nl);
-	    } else if (hasam && ccs == columns) { /* must always write */
-		moveto(ln, ccs - 1);		  /* last column */
+
+		if (oput_rprompt && !put_rprompt) {
+		    ccs += strlen(nl);
+		    *nl = '\0';
+		    continue;
+		}
+	    } else if (hasam && ccs == columns) {	/* must always write */
+		moveto(ln, ccs - 1);	/* last column */
 		putchar(' '), vcs++;
 #ifdef HAS_SELECT
-		cost ++;
+		cost++;
 #endif
 	    }
 	    return;
 	}
-
     /* try to insert/delete characters unless there is an rprompt and the old
        line also had it; in this case the length is not changed so that we
        don't have to redraw the rprompt */
@@ -454,7 +487,7 @@ int put_rprompt;
 			moveto(ln, ccs);
 		    }
 #endif
-		    if (ct + ccs + (int) strlen(ol) < winw - 1) {
+		    if (ct + ccs + (int)strlen(ol) < winw - 1) {
 			tc_inschars(ct = p1 - nl);
 			ccs = (vcs += ct);
 #ifdef HAS_SELECT
@@ -491,10 +524,9 @@ int cl;
 	putchar(' '), tcout(TCLEFT);
 	vln++, vcs = 0;
 #ifdef HAS_SELECT
-	cost ++;
+	cost++;
 #endif
     }
-
 /* move up */
 
     if (ln < vln) {
@@ -505,28 +537,28 @@ int cl;
 	instead of TCDOWN */
 
     while (ln > vln) {
-        if (vln < vmaxln - 1)
-            if (ln > vmaxln - 1) {
+	if (vln < vmaxln - 1)
+	    if (ln > vmaxln - 1) {
 		if (tc_downcurs(vmaxln - 1 - vln))
 		    vcs = 0;
-                vln = vmaxln - 1;
-            } else {
-                if (tc_downcurs(ln - vln))
+		vln = vmaxln - 1;
+	    } else {
+		if (tc_downcurs(ln - vln))
 		    vcs = 0;
-                vln = ln;
-                continue;
-            }
+		vln = ln;
+		continue;
+	    }
 	putchar('\r'), vcs = 0;	/* safety precaution */
 #ifdef HAS_SELECT
-	cost ++;
+	cost++;
 #endif
-        while (ln > vln) {
-            putchar('\n');
+	while (ln > vln) {
+	    putchar('\n');
 #ifdef HAS_SELECT
-            cost ++;
+	    cost++;
 #endif
-            vln++;
-        }
+	    vln++;
+	}
     }
     if (cl < vcs / 2) {
 	putchar('\r');
@@ -566,7 +598,6 @@ int cl;
 	tcoutarg(TCMULTRIGHT, ct);
 	return;
     }
-
 /* try to go with tabs if a multright is not feasible/convenient;
    tabs are assumed to be 8 spaces */
 
@@ -580,7 +611,6 @@ int cl;
 	if (vcs == cl)
 	    return;
     }
-
 /* if we're walking over the prompt and we can do a bunch of cursor rights,
    do them, even though they're more expensive.  (We can't redraw the
    prompt very easily in general.)  */
@@ -603,6 +633,7 @@ int tc_downcurs(ct)		/**/
 int ct;
 {
     int ret = 0;
+
     if (ct) {
 	if (tccan(TCMULTDOWN) &&
 	    (!tccan(TCDOWN) || tclen[TCMULTDOWN] < tclen[TCDOWN] * ct))
@@ -690,7 +721,7 @@ void singlerefresh()
     }
     if (winpos)
 	vbuf[winpos] = '<';
-    if ((int) strlen(vbuf + winpos) > winw) {
+    if ((int)strlen(vbuf + winpos) > winw) {
 	vbuf[winpos + winw - 1] = '>';
 	vbuf[winpos + winw] = '\0';
     }
