@@ -34,7 +34,8 @@
 typedef struct {
     int len;
     char *s;
-} dopestring;
+}
+dopestring;
 
 #endif
 
@@ -126,7 +127,7 @@ void spellword()
 
 void deletecharorlist()
 {				/**/
-    char **mc =   menucur;
+    char **mc = menucur;
 
     usemenu = isset(MENUCOMPLETE);
     useglob = isset(GLOBCOMPLETE);
@@ -225,7 +226,7 @@ static char *cmdstr;
 #define HAS_MISC    4
 #define HAS_PATHPAT 8
 
-int checkparams(p)	/**/
+int checkparams(p)		/**/
 char *p;
 {
     int t0, n, l = strlen(p), e = 0;
@@ -238,8 +239,7 @@ char *p;
 		if (strlen(hn->nam) == l)
 		    e = 1;
 	    }
-
-    return (n == 1) ? (getsparam(p) != NULL) : 
+    return (n == 1) ? (getsparam(p) != NULL) :
 	(!menucmp && e && isset(RECEXACT));
 }
 
@@ -250,14 +250,13 @@ char *str;
 	return 0;
     if (str[0] == '%')
 	return 0;
-    for (; *str; str++) {
+    for (; *str;) {
 	if (*str == String || *str == Qstring) {
 	    if (*++str == Inbrace) {
 		for (str++; *str; str++)
 		    if (*str == Outbrace)
 			break;
-	    }
-	    else
+	    } else
 		for (; *str; str++)
 		    if (*str != '^' && *str != Hat &&
 			*str != '=' && *str != Equals &&
@@ -270,6 +269,8 @@ char *str;
 	    *str == Bar || *str == Inbrack || *str == Inang ||
 	    *str == Quest || (*str == Inpar && str[1] == ':'))
 	    return 1;
+	if (*str)
+	    str++;
     }
     return 0;
 }
@@ -277,25 +278,53 @@ char *str;
 void docomplete(lst)		/**/
 int lst;
 {
-    char *s;
-    int olst = lst, dh;
+    char *s, *ol;
+    int olst = lst, dh, chl = 0;
 
     if (menucmp) {
 	do_menucmp(lst);
 	return;
-    } else if ((amenu=(isset(AUTOMENU) &&
-		       (lastcmd & ZLE_MENUCMP) &&
-		       lastambig)))
+    } else if ((amenu = (isset(AUTOMENU) &&
+			 (lastcmd & ZLE_MENUCMP) &&
+			 lastambig)))
 	usemenu = 1;
 
     dh = doexpandhist();
     if (dh == 0 || dh == 1)
 	return;
 
+    if (!isfirstln && chline != NULL) {
+	char *p;
+
+	ol = dupstring((char *) line);
+	*hptr = '\0';
+	chl = strlen(chline);
+	sizeline(ll + chl + 1);
+	strcpy((char *) line, chline);
+	strcat((char *) line, ol);
+	for (p = (char *) line; *p; p++)
+	    if (*p == '\n')
+		*p = ' ';
+	cs += chl;
+	ll = strlen((char *) line);
+    }
+    else
+	ol = NULL;
     inwhat = IN_NOTHING;
     qword = NULL;
     s = get_comp_string();
+    lexrestore();
+    if (ol) {
+	strcpy((char *) line, ol);
+	ll = strlen((char *) line);
+	cs -= chl;
+	wb -= chl;
+	we -= chl;
+	if (wb < 0 || we < 0)
+	    return;
+    }
     freeheap();
+    lexsave();
     if (inwhat == IN_ENV)
 	lincmd = 0;
     if (s) {
@@ -343,7 +372,7 @@ int lst;
 				*q != '#' && *q != Pound &&
 				*q != '~' && *q != Tilde &&
 				*q != '+')
-			break;
+				break;
 			for (t = q; *q && (ialnum(*q) || *q == '_'); q++);
 			sav = *q;
 			*q = '\0';
@@ -384,7 +413,7 @@ int lst;
 	    inststr(*x);
 	} else if (COMP_ISEXPAND(lst)) {
 	    char *ol = (olst == COMP_EXPAND_COMPLETE) ?
-		dupstring((char *) line) : (char *) line;
+	    dupstring((char *)line) : (char *)line;
 	    int ocs = cs, ne = noerrs;
 
 	    noerrs = 1;
@@ -395,29 +424,20 @@ int lst;
 	    noerrs = ne;
 
 	    if (olst == COMP_EXPAND_COMPLETE &&
-		!strcmp(ol, (char *) line)) {
+		!strcmp(ol, (char *)line)) {
+
 		cs = ocs;
 		errflag = 0;
 
-		if (isset(EXTENDEDGLOB)) {
-		    char *b;
-
-		    for (b = s; *b; b++)
-			if (*b == Pound)
-			    *b = '#';
-			else if (*b == Hat)
-			    *b = '^';
-		}
-
+		untokenize(s);
 		docompletion(s, lst, lincmd);
 	    }
-	}
-	else
+	} else
 	    docompletion(s, lst, lincmd);
 	zsfree(s);
     }
-    popheap();
     lexrestore();
+    popheap();
     zsfree(qword);
 }
 
@@ -440,9 +460,9 @@ int instring;			/* 1 if we are completing in a string */
 void addx(ptmp)			/**/
 char **ptmp;
 {
-    if (!line[cs] || inblank(line[cs])) {
+    if (!line[cs] || inblank(line[cs]) || line[cs] == ')') {
 	*ptmp = (char *)line;
-	line = (unsigned char *)ncalloc(strlen((char *)line) + 2);
+	line = (unsigned char *)ncalloc(strlen((char *)line) + 3);
 	memcpy(line, *ptmp, cs);
 	line[cs] = 'x';
 	strcpy((char *)line + cs + 1, (*ptmp) + cs);
@@ -476,6 +496,8 @@ char *get_comp_string()
 	    p++;
     if ((i & 1) || (j & 1) || (k & 1) || l) {
 	instring = (j & 1) ? 2 : (k & 1);
+	if (l)
+	    im = 2;
 	addx(&tmp);
 	if (!addedx) {
 	    tmp = (char *)line;
@@ -493,14 +515,15 @@ char *get_comp_string()
     } else
 	addx(&tmp);
     if (i & 1) {
+	sizeline(ll + 1);
 	if (tmp)
-	    q += (char *)line - tmp;
+	    q = (char *)line + (q - tmp);
 	for (p = q + strlen(q); p > q; p--)
 	    p[1] = *p;
 	*q++ = '$';
 	*q = '(';
-	cs++;
 	ll++;
+	cs++;
 	inc = 1;
     }
     linptr = (char *)line;
@@ -545,7 +568,7 @@ char *get_comp_string()
 	if (!zleparse && !tt0) {
 	    tt = tokstr ? dupstring(tokstr) : NULL;
 	    if (addedx && tt)
-		tt[strlen(tt) - 1] = '\0';
+		chuck(tt + cs - wb - 1);
 	    tt0 = tok;
 	    clwpos = i;
 	    cp = lincmd;
@@ -563,7 +586,7 @@ char *get_comp_string()
 	while (sl && clwords[i][sl - 1] == ' ')
 	    clwords[i][--sl] = '\0';
 	if (clwpos == i++ && addedx)
-	    chuck(&clwords[i - 1][strlen(clwords[i-1]) - 1]);
+	    chuck(&clwords[i - 1][((cs - wb - 1) >= sl) ? (sl - 1) : (cs - wb - 1)]);
     }
     while (tok != LEXERR && tok != ENDINPUT &&
 	   (tok != SEPER || (zleparse && !tt0)));
@@ -605,8 +628,7 @@ char *get_comp_string()
 	    if (addedx)
 		tokstr[cs - wb + 1] = '\0';
 	    s = ztrdup(tokstr);
-	} else
-	    s = ztrdup("");
+	}
 	if (tmp)
 	    line = (unsigned char *)tmp;
 	goto check;
@@ -633,7 +655,7 @@ char *get_comp_string()
     }
     if (inc)
 	cs--, ll--, inc = 0;
-    tt = (char *) line;
+    tt = (char *)line;
     if (tmp)
 	line = (unsigned char *)tmp;
     if (t0 != STRING) {
@@ -649,8 +671,7 @@ char *get_comp_string()
 	noaliases = 0;
 	permalloc();
 	return NULL;
-    }
-    else if (clwords[clwpos] && clwords[clwpos][0]) {
+    } else if (clwords[clwpos] && clwords[clwpos][0]) {
 	for (p = clwords[clwpos]; *p; p++)
 	    if ((*p == String || *p == Equals || *p == Inang || *p == Outang) &&
 		p[1] == Inpar)
@@ -661,21 +682,22 @@ char *get_comp_string()
 		    break;
 	    if (q > p && cs - wb <= q - clwords[clwpos]) {
 		if (!tmp) {
-		    tmp = (char *) line;
-		    tt = dupstring((char *) line);
+		    tmp = (char *)line;
+		    tt = dupstring((char *)line);
 		}
-		line = (unsigned char *) tt;
+		line = (unsigned char *)tt;
 		line[wb + q - clwords[clwpos] + addedx] = '\0';
-		linptr = (char *) line;
+		linptr = (char *)line;
+		ll = strlen((char *)line) - 1;
 		popheap();
 		permalloc();
 		lexrestore();
+		q = NULL;
 		goto start;
 	    }
 	    q = NULL;
 	}
     }
-
   check:
 
     if (q)
@@ -727,12 +749,9 @@ char *get_comp_string()
 	    wb += p - s - ((im & 2) ? 1 : 0);
 	}
 	for (p = s; *p && (ialnum(*p) || *p == '_'); p++);
-	if (p > s && addedx && *p)
-	    p--;
 	*p = '\0';
 	we = wb + p - s;
     }
-
     qword = ztrdup(s);
     for (p = s, tt = qword, i = wb; *p; p++, tt++, i++)
 	if (INULL(*p)) {
@@ -748,8 +767,7 @@ char *get_comp_string()
 			cs--;
 		    we--;
 		}
-	    }
-	    else {
+	    } else {
 		ocs = cs;
 		*tt = '\0';
 		cs = we;
@@ -762,19 +780,18 @@ char *get_comp_string()
 	    }
 	    chuck(p--);
 	}
-
     if (unset(NOBANGHIST)) {
-	q = tt = ncalloc(2*strlen(qword) + 1);
+	q = tt = ncalloc(2 * strlen(qword) + 1);
 	for (p = qword; *p; p++) {
-	    if (*p == (char) bangchar)
-		*q++ = '\\', wb--;
+	    if (*p == (char)bangchar)
+		*q++ = '\\';	/*, wb--, cs--;*/
 	    *q++ = *p;
 	}
 	*q = '\0';
 	zsfree(qword);
 	qword = ztrdup(tt);
     }
-
+    ll = strlen((char *)line);
     permalloc();
 
     return (char *)s;
@@ -787,7 +804,11 @@ int explincmd;
 {
     Lklist vl = newlist();
     char *ss;
+    int ng = opts[NULLGLOB];
 
+    opts[NULLGLOB] = OPT_SET;
+    lexsave();
+    s = dupstring(s);
     pushheap();
     addnode(vl, s);
     prefork(vl, 0);
@@ -830,7 +851,9 @@ int explincmd;
 	line[cs++] = ' ';
     }
   end:
+    opts[NULLGLOB] = ng;
     popheap();
+    lexrestore();
 }
 
 void gotword(s)			/**/
@@ -841,7 +864,7 @@ char *s;
 	wb = ll - wordbeg;
 	zleparse = 0;
     /* major hack ahead */
-	if (wb && line[wb] == '!' && line[wb - 1] == '\\')
+	if (wb > 0 && line[wb] == '!' && line[wb - 1] == '\\')
 	    wb--;
     }
 }
@@ -867,7 +890,7 @@ char **e;
 char **te;
 int *pl;
 {
-    char *tt, *v, *u, buf[MAXPATHLEN*2];
+    char *tt, *v, *u, buf[MAXPATHLEN * 2];
     int sf = 0;
 
     tt = v = buf;
@@ -904,7 +927,7 @@ int *pl;
 	*e += tt - buf;
     if (te && (sf & 2))
 	*te += tt - buf;
-	
+
     if (e && *e == u)
 	*e = v;
     if (te && *te == u)
@@ -948,8 +971,7 @@ char *t;
 	    cp = filecomp;
 	    cc = cp || ispattern;
 	    e = s + sl - fsl;
-	}
-	else {
+	} else {
 	    if ((cp = filecomp)) {
 		if ((test = domatch(s, filecomp, 0)))
 		    cc = 1;
@@ -963,7 +985,7 @@ char *t;
 	}
 	if (test) {
 	    fc = NULL;
-	    if (addwhat == -7 && ISEXCMD(((Cmdnam) t)->flags) && 
+	    if (addwhat == -7 && ISEXCMD(((Cmdnam) t)->flags) &&
 		!(fc = findcmd(s)))
 		return;
 	    if (fc)
@@ -990,18 +1012,19 @@ char *t;
     } else if (addwhat == -2 ||
 	       (addwhat == -3 && !(((Cmdnam) t)->flags & DISABLED)) ||
 	       (addwhat == -4 && (pm = (Param) t) && (pmtype(pm) == PMFLAG_s) &&
-	       (tt = pm->gets.cfn(pm)) && *tt == '/') ||
+		(tt = pm->gets.cfn(pm)) && *tt == '/') ||
 	       (addwhat > 0 &&
 		(((addwhat & CC_ARRAYS) && (((Param) t)->flags & PMFLAG_A)) ||
 		 ((addwhat & CC_INTVARS) && (((Param) t)->flags & PMFLAG_i)) ||
 		 ((addwhat & CC_ENVVARS) && (((Param) t)->flags & PMFLAG_x)) ||
 		 ((addwhat & CC_SCALARS) &&
-		 !(((Param) t)->flags & (PMFLAG_A | PMFLAG_i))) ||
+		  !(((Param) t)->flags & (PMFLAG_A | PMFLAG_i))) ||
 		 ((addwhat & CC_READONLYS) && (((Param) t)->flags & PMFLAG_r)) ||
 		 ((addwhat & CC_SPECIALS) && (((Param) t)->flags & PMFLAG_SPECIAL)) ||
 		 ((addwhat & CC_PARAMS) && !(((Param) t)->flags & PMFLAG_x)) ||
 		 ((addwhat & CC_FUNCS) && (((Cmdnam) t)->flags & SHFUNC)) ||
-		 ((addwhat & CC_BUILTINS) && (((Cmdnam) t)->flags & BUILTIN)) ||
+		 ((addwhat & CC_BUILTINS) && (((Cmdnam) t)->flags & BUILTIN) &&
+		  !(((Cmdnam) t)->flags & EXCMD)) ||
 		 ((addwhat & CC_DISCMDS) && (((Cmdnam) t)->flags & DISABLED)) ||
 		 ((addwhat & CC_EXCMDS) && (((Cmdnam) t)->flags & EXCMD)) ||
 		 ((addwhat & CC_ALREG) && (((Alias) t)->cmd) == 1) ||
@@ -1080,8 +1103,8 @@ dopestring *data;
 
 #endif
 
-void maketildelist()		/**/
-{
+void maketildelist()
+{				/**/
     int i;
 
 #ifdef HAS_NIS
@@ -1096,7 +1119,7 @@ void maketildelist()		/**/
     data.len = fpl;
  /* Get potential matches from NIS and cull those without local accounts */
     if (getdomainname(domain, YPMAXDOMAIN) == 0) {
-	cb.foreach = (int((*)())) match_username;
+	cb.foreach = (int ((*)()))match_username;
 	cb.data = (char *)&data;
 	yp_all(domain, PASSWD_MAP, &cb);
 /*	for (n = firstnode(matches); n; incnode(n))
@@ -1149,7 +1172,7 @@ void maketildelist()		/**/
 	    addmatch(namdirs[i].name, NULL);
 }
 
-char *rembslash(s)	/**/
+char *rembslash(s)		/**/
 char *s;
 {
     char *t = s = dupstring(s);
@@ -1157,9 +1180,9 @@ char *s;
     while (*s)
 	if (*s == '\\') {
 	    chuck(s);
-	    if (*s) s++;
-	}
-	else
+	    if (*s)
+		s++;
+	} else
 	    s++;
 
     return t;
@@ -1209,7 +1232,7 @@ int class;
 
 Compctl ccmain;
 
-Compctl get_ccompctl(occ, compadd, incmd)  /**/
+Compctl get_ccompctl(occ, compadd, incmd)	/**/
 Compctl occ;
 int *compadd;
 int incmd;
@@ -1237,14 +1260,14 @@ int incmd;
 		 !strcmp("-nt", s) ||
 		 !strcmp("-ot", s) ||
 		 !strcmp("-ef", s)) ? CC_FILES :
-		     (CC_FILES | CC_PARAMS);
+		(CC_FILES | CC_PARAMS);
 	    ret = &cc_dummy;
 	    cc_dummy.refc = 10000;
 	} else if (incmd)
 	    ret = &cc_compos;
-	else if (linredir || 
+	else if (linredir ||
 		 !(cmdstr &&
-		   (((ccp = (Compctlp)gethnode(cmdstr, compctltab)) &&
+		   (((ccp = (Compctlp) gethnode(cmdstr, compctltab)) &&
 		     (compc = ret = ccp->cc)) ||
 		    ((s = dupstring(cmdstr)) && remlpaths(&s) &&
 		     (ccp = (Compctlp) gethnode(s, compctltab)) &&
@@ -1254,7 +1277,6 @@ int incmd;
 	ccmain = compc = ret;
 	ccmain->refc++;
     }
-
     if (compc && compc->ext) {
 	compc = compc->ext;
 	for (t = 0; compc && !t; compc = compc->next) {
@@ -1286,7 +1308,7 @@ int incmd;
 			    untokenize(s);
 			    sc = rembslash(cc->u.s.s[i]);
 			    a = strlen(sc);
-			    if (!strncmp(s, sc,a)) {
+			    if (!strncmp(s, sc, a)) {
 				*compadd = (cc->type == CCT_CURSUF ? a : 0);
 				t = 1;
 			    }
@@ -1429,7 +1451,7 @@ int what;
 
 }
 
-char *getreal(str)	/**/
+char *getreal(str)		/**/
 char *str;
 {
     Lklist l = newlist();
@@ -1490,7 +1512,8 @@ int all;
 		    if ((test = !strncmp(n, fpre, fpl)))
 			test = !strcmp(e, fsuf);
 		}
-		if (!test) continue;
+		if (!test)
+		    continue;
 		if (!all) {
 		    strcpy(q, n);
 		    if (stat(p, &buf) < 0)
@@ -1499,8 +1522,8 @@ int all;
 		if (all ||
 		    (dirs && (buf.st_mode & S_IFMT) == S_IFDIR) ||
 		    (execs &&
-		    ((buf.st_mode & (S_IFMT | S_IEXEC))
-		     == (S_IFREG | S_IEXEC)))) {
+		     ((buf.st_mode & (S_IFMT | S_IEXEC))
+		      == (S_IFREG | S_IEXEC)))) {
 		    if (*psuf) {
 			int o = strlen(p), tt;
 
@@ -1542,6 +1565,7 @@ char **ps;
 {
     if (*ps) {
 	char *p = quotename(*ps, NULL, NULL, NULL);
+
 	if (p != *ps) {
 	    zsfree(*ps);
 	    *ps = ztrdup(p);
@@ -1551,7 +1575,6 @@ char **ps;
 
 int clearflag;
 
-
 void docompletion(s, lst, incmd)/**/
 char *s;
 int lst;
@@ -1563,11 +1586,12 @@ int incmd;
     int owe = we, owb = wb, ocs = cs, delit;
     unsigned char *ol = NULL;
 
- xorrec:
+  xorrec:
 
     if (unset(COMPLETEINWORD) && cs != we)
 	cs = we;
-    offs = cs - wb;
+    if ((offs = cs - wb) > (t = strlen(s)))
+	offs = t;
 
     ispattern = haswhat = lastambig = 0;
     patcomp = filecomp = NULL;
@@ -1586,16 +1610,15 @@ int incmd;
     zsfree(mpre);
     zsfree(msuf);
 
-    rpre = rsuf = lpre = lsuf = ppre = psuf = prpre = 
+    rpre = rsuf = lpre = lsuf = ppre = psuf = prpre =
 	fpre = fsuf = mpre = msuf = firstm = NULL;
 
     if (!cc) {
 	heapalloc();
 	pushheap();
 	os = dupstring(s);
-	ol = (unsigned char *) dupstring((char *) line);
+	ol = (unsigned char *)dupstring((char *)line);
     }
-
     matches = newlist();
 
     if (!cc || cc->ext)
@@ -1643,14 +1666,15 @@ int incmd;
 	    !strcmp(sdup, sd + suffixll - sl))
 	    ccsuffix = NULL, haswhat |= HAS_SUFFIX, s[suffixll - sl] = '\0';
     }
-
     if ((ic = *s) != Tilde && ic != Equals)
 	ic = '\0';
     offs = cs - wb;
+    if ((offs = cs - wb) > (t = strlen(s)))
+	offs = t;
 
  /* Check if we have to complete a parameter name... */
 
-    for (p = s + offs; p >= s && *p != String; p--);
+    for (p = s + offs; p > s && *p != String; p--);
     if (*p == String && p[1] != Inpar && p[1] != Inbrack) {
 	char *b = p + 1, *e = b;
 	int n, br = 1;
@@ -1691,14 +1715,12 @@ int incmd;
 	} else
 	    complexpect = 0;
     }
-
     ooffs = offs;
     if (cc->mask & CC_DELETE) {
 	delit = 1;
 	*s = '\0';
 	offs = 0;
-    }
-    else
+    } else
 	delit = 0;
 
  /* compute line prefix/suffix */
@@ -1709,10 +1731,10 @@ int incmd;
     if (strcmp(p, s) && !strpfx(p, qword)) {
 	int l1, l2;
 
-	backdel(l1 = strlen(s));
+	backdel(l1 = cs - wb);
 	untokenize(p);
 	inststrlen(p, 1, l2 = strlen(p));
-	we += l2-1l;
+	we += l2 - 1l;
     }
     lpre = ztrdup(s);
     s[offs] = sav;
@@ -1724,7 +1746,7 @@ int incmd;
 	foredel(l1 = strlen(s + offs));
 	untokenize(p);
 	inststrlen(p, 0, l2 = strlen(p));
-	we += l2-l1;
+	we += l2 - l1;
     }
     lsuf = ztrdup(s + offs);
 
@@ -1776,13 +1798,12 @@ int incmd;
 	ispattern = 0;
 
     if (ispattern) {
-	p = (char *) ncalloc(rpl + rsl + 2);
+	p = (char *)ncalloc(rpl + rsl + 2);
 	strcpy(p, rpre);
 	if (rpl && p[rpl - 1] != Star) {
 	    p[rpl] = Star;
 	    strcpy(p + rpl + 1, rsuf);
-	}
-	else
+	} else
 	    strcpy(p + rpl, rsuf);
 	patcomp = parsereg(p);
     }
@@ -1828,7 +1849,7 @@ int incmd;
 	if (useglob && (ispattern & 2)) {
 	    int t2;
 
-	    p = (char *) ncalloc((t2 = strlen(fpre)) + strlen(fsuf) + 2);
+	    p = (char *)ncalloc((t2 = strlen(fpre)) + strlen(fsuf) + 2);
 	    strcpy(p, fpre);
 	    if ((!t2 || p[t2 - 1] != Star) && *fsuf != Star)
 		p[t2++] = Star;
@@ -1859,12 +1880,12 @@ int incmd;
 		opts[NULLGLOB] = OPT_SET;
 
 		addwhat = 0;
-		p = (char *) ncalloc(lpl + lsl + 3);
+		p = (char *)ncalloc(lpl + lsl + 3);
 		strcpy(p, lpre);
-		if (*lsuf != '*' && *lpre && lpre[lpl-1] != '*')
+		if (*lsuf != '*' && *lpre && lpre[lpl - 1] != '*')
 		    strcat(p, "*");
 		strcat(p, lsuf);
-		if (*lsuf && lsuf[lsl-1] != '*' && lsuf[lsl-1] != ')')
+		if (*lsuf && lsuf[lsl - 1] != '*' && lsuf[lsl - 1] != ')')
 		    strcat(p, "*");
 
 		tokenize(p);
@@ -1907,10 +1928,10 @@ int incmd;
 			char *g = dupstring(cc->glob), pa[MAXPATHLEN];
 			char *p2, *p3;
 			int ne = noerrs, md = opts[MARKDIRS];
-			
+
 			glob_pre = fpre;
 			glob_suf = fsuf;
-			
+
 			noerrs = 1;
 			addwhat = -6;
 			strcpy(pa, prpre);
@@ -1931,7 +1952,7 @@ int incmd;
 			    *p = '\0';
 			    tokenize(g = dupstring(g));
 			    if (*g == Equals || *g == Tilde) {
-				filesub(&g,0);
+				filesub(&g, 0);
 				addnode(l, dupstring(g));
 			    } else if (*g == '/')
 				addnode(l, dupstring(g));
@@ -1972,15 +1993,14 @@ int incmd;
 	    }
 	}
     }
-
-    /* Use tricat() instead of dyncat() to get zalloc()'d memory */
+ /* Use tricat() instead of dyncat() to get zalloc()'d memory */
     if (ic == Tilde || ic == Equals) {
 	char *orpre = rpre;
+
 	rpre = tricat("", (ic == Tilde) ? "~" : "=", rpre);
 	rpl++;
 	zsfree(orpre);
     }
-
     if (!ic && (cc->mask & CC_COMMPATH) && !*ppre && !*psuf) {
 	dumphtable(aliastab, -2);
 	if (isset(HASHLISTALL))
@@ -1991,6 +2011,12 @@ int incmd;
     }
     addwhat = -2;
 
+    if (cc->mask & CC_NAMED) {
+	int t0;
+
+	for (t0 = 0; t0 < userdirct; t0++)
+	    addmatch(namdirs[t0].name, NULL);
+    }
     if (cc->mask & CC_OPTIONS) {
 	struct option *o;
 
@@ -2035,8 +2061,7 @@ int incmd;
 		p = dupstring(os + ooffs);
 		untokenize(p);
 		addnode(args, p);
-	    }
-	    else {
+	    } else {
 		addnode(args, lpre);
 		addnode(args, lsuf);
 	    }
@@ -2079,7 +2104,7 @@ int incmd;
 	zleparse = 1;
 	lexsave();
 	hungets(cc->str);
-	hungets("foo ");		/* KLUDGE! */
+	hungets("foo ");	/* KLUDGE! */
 	strinbeg();
 	noaliases = 1;
 	do {
@@ -2140,7 +2165,7 @@ int incmd;
 	lithist = l;
     }
     if ((t = cc->mask & (CC_ARRAYS | CC_INTVARS | CC_ENVVARS | CC_SCALARS |
-			CC_READONLYS | CC_SPECIALS | CC_PARAMS)))
+			 CC_READONLYS | CC_SPECIALS | CC_PARAMS)))
 	dumphtable(paramtab, t);
     if ((t = cc->mask & (CC_FUNCS | CC_BUILTINS | CC_DISCMDS | CC_EXCMDS)))
 	dumphtable(cmdnamtab, t);
@@ -2160,7 +2185,7 @@ int incmd;
     quotepresuf(&fsuf);
     quotepresuf(&ppre);
     quotepresuf(&psuf);
-    
+
     if (empty(matches) || errflag) {
 	if (cc->xor && !isp) {
 	    errflag = 0;
@@ -2169,9 +2194,10 @@ int incmd;
 	    we = owe;
 	    cs = ocs;
 	    s = dupstring(os);
-	    strcpy((char *) line, (char *) ol);
+	    strcpy((char *)line, (char *)ol);
 	    goto xorrec;
 	}
+	strcpy((char *)line, (char *)ol);
 	feep();
     } else if (lst == COMP_LIST_COMPLETE)
 	listtlist(matches);
@@ -2181,7 +2207,6 @@ int incmd;
 	    strcpy((char *)line + wb, (char *)line + we);
 	    we = cs = wb;
 	}
-
 	if (nextnode(firstnode(matches)))
 	    do_ambiguous();
 	else
@@ -2192,9 +2217,9 @@ int incmd;
 	int up;
 
 	trashzle();
-	
+
 	clearflag = (isset(USEZLE) && termok &&
-	    (isset(ALWAYSLASTPROMPT) && mult == 1)) ||
+		     (isset(ALWAYSLASTPROMPT) && mult == 1)) ||
 	    (unset(ALWAYSLASTPROMPT) && mult != 1);
 
 	up = printfmt(expl, countnodes(matches), 1);
@@ -2203,8 +2228,9 @@ int incmd;
 	    tcmultout(TCUP, TCMULTUP, up + nlnct);
 	fflush(stdout);
     }
-
     ll = strlen((char *)line);
+    if (cs > ll)
+	cs = ll;
   compend:
     if (ccmain != &cc_dummy)
 	freecompctl(ccmain);
@@ -2256,19 +2282,23 @@ char *nam;
 	return getaparam(nam);
 }
 
-int strbpcmp(a, b)	/**/
+int strbpcmp(a, b)		/**/
 const void *a;
 const void *b;
 {
-    char *aa = *((char **) a), *bb = *((char **) b);
+    char *aa = *((char **)a), *bb = *((char **)b);
 
-    for (; *aa; aa++, bb++) {
+    while (*aa && *bb) {
 	if (*aa == '\\')
 	    aa++;
 	if (*bb == '\\')
 	    bb++;
 	if (*aa != *bb)
-	    return (int) (*aa - *bb);
+	    return (int)(*aa - *bb);
+	if (*aa)
+	    aa++;
+	if (*bb)
+	    bb++;
     }
     return 0;
 }
@@ -2290,7 +2320,7 @@ void do_ambiguous()
 				     sizeof(char *));
 
     for (nod = firstnode(matches); nod; incnode(nod))
-	*ap++ = (char *) getdata(nod);
+	*ap++ = (char *)getdata(nod);
     *ap = NULL;
 
     qsort((vptr) amatches, nmatches, sizeof(char *),
@@ -2304,10 +2334,10 @@ void do_ambiguous()
     *cp = NULL;
 
     if ((nmatches = arrlen(amatches)) == 1) {
+	lastambig = 0;
 	do_single(amatches[0]);
 	return;
     }
-
     complexpect = 0;
 
     if ((p = (usemenu || ispattern)) || isset(AUTOMENU)) {
@@ -2326,7 +2356,7 @@ void do_ambiguous()
 		inststrlen(firstm, 1, ab);
 	    if (ae && !atend)
 		inststrlen(firstm + strlen(firstm) - ae, 0, ae);
-	    if (isset(LISTAMBIGUOUS) && (ab || (ae && !atend))) { 
+	    if (isset(LISTAMBIGUOUS) && (ab || (ae && !atend))) {
 		lastambig = 0;
 		return;
 	    }
@@ -2358,7 +2388,7 @@ int ls;
 void do_single(str)		/**/
 char *str;
 {
-    int ccs, l, insc = 0;
+    int ccs, l, insc = 0, inscs = 0;
     char singlec = ' ';
 
     addedsuffix = 0;
@@ -2437,12 +2467,15 @@ char *str;
 		    addedsuffix = isset(AUTOREMOVESLASH) ? 1 : 0;
 	    }
 	}
-	if (menuend && ((char) line[menuend - 1]) != singlec)
+	if (menuend > ll)
+	    menuend = ll;
+	if (menuend && ((char)line[menuend - 1]) != singlec)
 	    if (!menucur || !line[menuend]) {
 		ccs = cs;
 		cs = menuend;
 		inststrlen((char *)&singlec, 1, 1);
 		insc = 1;
+		inscs = cs;
 		if (!menuwe)
 		    cs = ccs;
 	    } else
@@ -2451,8 +2484,13 @@ char *str;
     if (isset(ALWAYSTOEND) || menuwe)
 	cs = menuend + !(haswhat & HAS_SUFFIX);
     if (menucmp && singlec == ' ' && !(haswhat & HAS_SUFFIX)) {
-	if (insc)
+	if (insc) {
+	    ccs = cs;
+	    cs = inscs;
 	    backdel(1);
+	    if (ccs != inscs)
+	      cs = ccs;
+	}
 	else
 	    cs--;
     }
@@ -2505,7 +2543,7 @@ char *t;
 	int i = 0;
 	char *s2 = s + strlen(s) - 1, *t2 = t + strlen(t) - 1;
 
-	while (s2 >= s && *s2 == *t2)
+	while (s2 >= s && t2 >= t && *s2 == *t2)
 	    s2--, t2--, i++;
 
 	return i;
@@ -2513,7 +2551,7 @@ char *t;
 	return 0;
 }
 
-int printfmt(fmt, n, dopr)		/**/
+int printfmt(fmt, n, dopr)	/**/
 char *fmt;
 int n;
 int dopr;
@@ -2521,27 +2559,26 @@ int dopr;
     char *p = fmt, nc[14];
     int l = 1, cc = 0;
 
-    for (;*p;p++) {
+    for (; *p; p++) {
 	if (*p == '%')
-	    if (*++p) 
-		switch(*p) {
-		  case '%':
+	    if (*++p)
+		switch (*p) {
+		case '%':
 		    if (dopr)
 			putchar('%');
 		    cc++;
 		    break;
-		  case 'n':
+		case 'n':
 		    sprintf(nc, "%d", n);
 		    if (dopr)
 			printf(nc);
 		    cc += strlen(nc);
-		}
-	    else
+	    } else
 		break;
 	else {
 	    cc++;
 	    if (*p == '\n') {
-		l += 1 + (cc/columns);
+		l += 1 + (cc / columns);
 		cc = 0;
 	    }
 	    if (dopr)
@@ -2551,7 +2588,7 @@ int dopr;
     if (dopr)
 	putchar('\n');
 
-    return l + (cc/columns);
+    return l + (cc / columns);
 }
 
 void listmatches()
@@ -2562,15 +2599,15 @@ void listmatches()
     char **arr, **ap, sav;
 
     off = ispattern && ppre && *ppre &&
-	!(haswhat & (HAS_MISC|HAS_PATHPAT)) ? strlen(ppre) : 0;
+	!(haswhat & (HAS_MISC | HAS_PATHPAT)) ? strlen(ppre) : 0;
     boff = ispattern && psuf && *psuf &&
-	!(haswhat & (HAS_MISC|HAS_PATHPAT)) ? strlen(psuf) : 0;
+	!(haswhat & (HAS_MISC | HAS_PATHPAT)) ? strlen(psuf) : 0;
 
     trashzle();
     ct = nmatches;
 
     clearflag = (isset(USEZLE) && termok &&
-	(isset(ALWAYSLASTPROMPT) && mult == 1)) ||
+		 (isset(ALWAYSLASTPROMPT) && mult == 1)) ||
 	(unset(ALWAYSLASTPROMPT) && mult != 1);
 
     arr = amatches;
@@ -2621,7 +2658,7 @@ void listmatches()
 	if (of) {
 	    while (*ap) {
 		int t2 = ispattern ? strlen(*ap) :
-		    strlen(*ap + off) - boff + 1 + fpl + fsl;
+		strlen(*ap + off) - boff + 1 + fpl + fsl;
 		char pbuf[MAXPATHLEN], *pb;
 		struct stat buf;
 
@@ -2629,7 +2666,7 @@ void listmatches()
 		    sav = ap[0][t2 - boff];
 		    ap[0][t2 - boff] = '\0';
 		    printf("%s", *ap + off);
-		    ap[0][t2 - boff] = sav;	
+		    ap[0][t2 - boff] = sav;
 		    pb = *ap;
 		    t2 -= off + boff - 1;
 		} else {
@@ -2649,7 +2686,7 @@ void listmatches()
 	} else
 	    while (*ap) {
 		int t2 = ispattern ? strlen(*ap) :
-		    strlen(*ap + off) - boff;
+		strlen(*ap + off) - boff;
 
 		if (ispattern) {
 		    sav = ap[0][t2 - boff];
@@ -2660,8 +2697,7 @@ void listmatches()
 		} else if (!(haswhat & HAS_MISC)) {
 		    printf("%s%s%s", fpre, *ap, fsuf);
 		    t2 += fpl + fsl;
-		}
-		else {
+		} else {
 		    printf("%s%s%s", lpre, *ap, lsuf);
 		    t2 += lpl + lsl;
 		}
@@ -2791,13 +2827,13 @@ Lklist l;
 
 int doexpandhist()
 {				/**/
-    unsigned char *cc, *ce, oc = ' ';
+    unsigned char *cc, *ce, oc = ' ', ooc = ' ';
     int t0, oldcs, oldll;
 
-    for (cc = line, ce = line + ll; cc < ce; oc = *cc++)
+    for (cc = line, ce = line + ll; cc < ce; ooc = oc, oc = *cc++)
 	if (*cc == '\\' && cc[1])
 	    cc++;
-	else if ((*cc == bangchar && oc != '>') ||
+	else if ((*cc == bangchar && oc != '>' && (oc != '&' || ooc != '>')) ||
 		 (*cc == hatchar && *line == hatchar && cc != line))
 	    break;
     if (*cc == bangchar && cc[1] == '"')
