@@ -55,24 +55,33 @@
 #ifdef SYSV
 
 #ifndef SYSVR4
+#if defined(SCO) || defined(_SEQUENT_)
+#include <limits.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#ifdef SCO
+#define MAXPATHLEN 255
+#endif
+#else
 #include <sys/bsdtypes.h>
 #include <sys/limits.h>
 #include <sys/sioctl.h>
 #define MAXPATHLEN PATH_MAX
 #define lstat stat
 #endif
+#endif
 
-#ifdef IRIX5
+#if defined(IRIX5) || defined(SCO)
 #include <sys/signal.h>
 #endif
 
 int gethostname();
 
-#else /* not SYSV */
+#else				/* not SYSV */
 
 #include <sys/types.h>		/* this is the key to the whole thing */
 
-#endif /* not SYSV */
+#endif				/* not SYSV */
 
 #ifdef _IBMR2
 #undef _BSD			/* union wait SUCKS! */
@@ -90,40 +99,44 @@ int gethostname();
 /* this stuff is in <sys/wait.h> but with _POSIX_SOURCE
  * it will not be seen -- this is correct but not nice
  * considering using "union wait" */
-typedef union wait	{
-	int	w_status;		/* used in syscall */
-	struct {
+typedef union wait {
+    int w_status;		/* used in syscall */
+    struct {
 #ifdef _MIPSEL
-		unsigned int	w_Termsig:7,	/* termination signal */
-				w_Coredump:1,	/* core dump indicator */
-				w_Retcode:8,	/* exit code if w_termsig==0 */
-				w_Filler:16;	/* upper bits filler */
+	unsigned int w_Termsig:7,	/* termination signal */
+	    w_Coredump:1,	/* core dump indicator */
+	    w_Retcode:8,	/* exit code if w_termsig==0 */
+	    w_Filler:16;	/* upper bits filler */
 #endif
 #ifdef _MIPSEB
-		unsigned int	w_Filler:16,	/* upper bits filler */
-				w_Retcode:8,	/* exit code if w_termsig==0 */
-				w_Coredump:1,	/* core dump indicator */
-				w_Termsig:7;	/* termination signal */
+	unsigned int w_Filler:16,	/* upper bits filler */
+	    w_Retcode:8,	/* exit code if w_termsig==0 */
+	    w_Coredump:1,	/* core dump indicator */
+	    w_Termsig:7;	/* termination signal */
 #endif
-	} w_T;
-	/*
+    }
+    w_T;
+ /*
 	 * Stopped process status.  Returned
 	 * only for traced children unless requested
 	 * with the WUNTRACED option bit.
 	 */
-	struct {
+    struct {
 #ifdef _MIPSEL
-		unsigned int	w_Stopval:8,	/* == W_STOPPED if stopped */
-				w_Stopsig:8,	/* signal that stopped us */
-				w_Filler:16;	/* upper bits filler */
+	unsigned int w_Stopval:8,	/* == W_STOPPED if stopped */
+	    w_Stopsig:8,	/* signal that stopped us */
+	    w_Filler:16;	/* upper bits filler */
 #endif
 #ifdef _MIPSEB
-		unsigned int	w_Filler:16,	/* upper bits filler */
-				w_Stopsig:8,	/* signal that stopped us */
-				w_Stopval:8;	/* == W_STOPPED if stopped */
+	unsigned int w_Filler:16,	/* upper bits filler */
+	    w_Stopsig:8,	/* signal that stopped us */
+	    w_Stopval:8;	/* == W_STOPPED if stopped */
 #endif
-	} w_S;
-} wait_t;
+    }
+    w_S;
+}
+wait_t;
+
 #endif
 
 #ifdef _CRAY
@@ -131,8 +144,18 @@ typedef union wait	{
 #include <sys/machd.h>		/* HZ definition here */
 #endif
 
-#if defined(HAS_TIME) || defined(_CRAY) || defined(IRIX5)
+#if defined(HAS_TIME) || defined(_CRAY) || defined(IRIX5) || defined(_SEQUENT_)
 #include <time.h>
+/* Sequent DYNIX/ptx gettimeofday() is in the X11 library.  Define timezone
+   here so we don't have to load all of X11/Xos.h.
+   billb@progress.com 4-Jan-93 */
+#ifdef _SEQUENT_
+struct timezone {
+    int tz_minuteswest;
+    int tz_dsttime;
+};
+
+#endif
 #if (defined(_CRAY) && defined(__STDC__)) || defined(IRIX5)
 #include <sys/time.h>
 #endif
@@ -140,8 +163,8 @@ typedef union wait	{
 #include <sys/time.h>
 #endif
 
-#ifdef IRIX5
-#include <sys/select.h>	/* needs timeval */
+#if defined(IRIX5) || defined(SCO)
+#include <sys/select.h>		/* needs timeval */
 #endif
 
 #ifdef SYSV
@@ -154,12 +177,21 @@ typedef union wait	{
 #include <sys/resource.h>
 #endif
 
+#if defined(RLIMIT_OFILE) && defined(RLIMIT_NOFILE)
+#undef RLIMIT_NOFILE
+#endif
+
 #include <signal.h>
 #ifdef USE_SIGSET
 #undef signal
 #define signal sigset
 #endif
 
+#if defined(SCO)
+#include <sys/stream.h>
+#include <sys/ptem.h>
+#endif
+
 #ifdef HAS_TERMIO
 #define VDISABLEVAL -1
 #define HAS_TIO 1
@@ -171,21 +203,21 @@ typedef union wait	{
 #include <termios.h>
 #ifdef __sgi
 #ifndef VSWTCH
-#define VSWTCH 7 /* VSWTCH not defined if _POSIX_SOURCE */
+#define VSWTCH 7		/* VSWTCH not defined if _POSIX_SOURCE */
 #endif
 #ifndef CSWTCH
 #ifndef CTRL
 #define	CTRL(c)	((c)&037)
 #endif
-#define CSWTCH CTRL('z') /* CSWTCH not defined if _POSIX_SOURCE */
+#define CSWTCH CTRL('z')	/* CSWTCH not defined if _POSIX_SOURCE */
 #endif
 #ifndef XTABS
-#define XTABS 0014000 /* XTABS not defined if _POSIX_SOURCE */
+#define XTABS 0014000		/* XTABS not defined if _POSIX_SOURCE */
 #endif
 #ifndef ONLCR
-#define	ONLCR 0000004 /* ONLCR not defined if _POSIX_SOURCE */
+#define	ONLCR 0000004		/* ONLCR not defined if _POSIX_SOURCE */
 #endif
-#endif /* __sgi */
+#endif				/* __sgi */
 #else
 #include <sgtty.h>
 #endif
@@ -201,21 +233,31 @@ typedef union wait	{
 #define OPEN_MAX 64
 #define NOFILE OPEN_MAX
 #endif
- 
+
+#if defined(SCO)
+#define NOFILE _POSIX_OPEN_MAX
+#endif
+
+#if defined(SCO)
+#include <utime.h>
+#endif
 #include <sys/times.h>
 
 #ifdef HAS_DIRENT
-#if defined(ardent)	/* ardent = Titan */ 
+#if defined(ardent)		/* ardent = Titan */
 #include <sys/dirent.h>
 #endif
-#include <dirent.h>	
-#else			
-#include <sys/dir.h>	
+#include <dirent.h>
+#else
+#include <sys/dir.h>
+#if defined(sony) || defined(MACH)
+#define dirent direct
+#endif
 #endif
 #if defined(__NeXT__)
-#define dirent direct	/* if you have absolutely no struct dirent */
-#endif			/* anywhere in system headers, add your */
-			/* system here and pray */
+#define dirent direct		/* if you have absolutely no struct dirent */
+#endif				/* anywhere in system headers, add your */
+ /* system here and pray */
 
 #ifdef __hp9000s800
 #include <sys/bsdtty.h>
@@ -295,8 +337,8 @@ typedef union wait	{
 /* SYSV or POSIX compatible BSD */
 /* ARGH, this is currently broken as the 'BSD' test had to be taken stripped
  * of an additional test BSD >= 199301 as Solaris cpp is broken ++jhi; */
-#if defined(SYSV) || defined(_POSIX_SOURCE) || defined(BSD) || defined(__convex__) || defined(_CRAY) || defined(__osf__) || defined(__linux__)
-#if !defined(__sgi) || !defined(_BSD_COMPAT) /* IRIX */
+#if (defined(SYSV) || defined(_POSIX_SOURCE) || defined(BSD) || defined(__convex__) || defined(_CRAY) || defined(__osf__) || defined(__linux__)) && !defined(NeXT)
+#if !defined(__sgi) || !defined(_BSD_COMPAT)	/* IRIX */
 #ifdef __hpux
 #define GETPGRP()  getpgrp2(0)
 #else
@@ -310,6 +352,7 @@ typedef union wait	{
 #endif
 
 #if defined(__sgi) && defined(__STDC__) && (defined(_BSD_COMPAT) || defined(_BSD_SIGNALS))
+#ifdef _POSIX_SOURCE
 #ifdef IRIX5
 #define gettimeofday(tp, tzp) BSDgettimeofday(tp, tzp)
 #else
@@ -317,7 +360,8 @@ typedef union wait	{
 #define signal(s, h) BSDsignal(s, (HANDTYPE (*)()) h)
 #endif
 #endif
- 
+#endif
+
 #ifndef F_OK
 #define F_OK 00
 #define R_OK 04
@@ -341,6 +385,14 @@ typedef union wait	{
 #endif
 #endif
 
+#if 0
+/* The quad_t should be used for BSD 4.4.  There was an ifdef BSD4_4
+   here, but it broke FreeBSD compilation.  */
+#define RLIM_TYPE quad_t
+#else
+#define RLIM_TYPE long
+#endif
+
 /* math.c */
 typedef int LV;
 
@@ -683,8 +735,8 @@ struct pline {
     Pline right;
 };
 
-#define END	0	/* pnode *right is null */
-#define PIPE	1	/* pnode *right is the rest of the pipeline */
+#define END	0		/* pnode *right is null */
+#define PIPE	1		/* pnode *right is the rest of the pipeline */
 
 /* tree element for commands */
 
@@ -692,6 +744,7 @@ struct cmd {
     int ntype;
     int type;
     int flags;			/* see CFLAGs below */
+    int lineno;			/* lineno of script for command */
     Lklist args;		/* command & argmument List (char *'s) */
     union {
 	List list;		/* for SUBSH/CURSH/SHFUNC */
@@ -800,10 +853,10 @@ struct whilecmd {
 /* one for each fd open */
 
 struct multio {
-    int ct;		/* # of redirections on this fd */
-    int rflag;		/* 0 if open for reading, 1 if open for writing */
-    int pipe;		/* fd of pipe if ct > 1 */
-    int fds[NOFILE];	/* list of src/dests redirected to/from this fd */
+    int ct;			/* # of redirections on this fd */
+    int rflag;			/* 0 if open for reading, 1 if open for writing */
+    int pipe;			/* fd of pipe if ct > 1 */
+    int fds[NOFILE];		/* list of src/dests redirected to/from this fd */
 };
 
 /* node used in command path hash table (cmdnamtab) */
@@ -814,6 +867,7 @@ struct cmdnam {
     int flags;
     union {
 	char **name;		/* full pathname if !(flags & BUILTIN) */
+	char *cmd;		/* file name for hashed commands */
 	int binnum;		/* func to exec if type & BUILTIN */
 	List list;		/* list to exec if type & SHFUNC */
     }
@@ -824,6 +878,8 @@ struct cmdnam {
 #define BUILTIN      0x20000
 #define SHFUNC       0x40000
 #define DISABLED     0x80000
+#define HASHCMD      (EXCMD | BUILTIN)
+
 #define ISEXCMD(X) ((X) & EXCMD)
 
 /* node used in parameter hash table (paramtab) */
@@ -936,7 +992,7 @@ struct job {
 #define STAT_TIMED 4		/* job is being timed */
 #define STAT_DONE 8
 #define STAT_LOCKED 16		/* shell is finished creating this job,	may be
-				   deleted from job table */ 
+				   deleted from job table */
 #define STAT_INUSE 64		/* this job entry is in use */
 
 #define SP_RUNNING -1		/* fake statusp for running jobs */
@@ -974,7 +1030,7 @@ struct alias {
     char *nam;			/* hash data */
     char *text;			/* expansion of alias */
     int cmd;			/* one for regular aliases, zero for global
-				   aliases, negative for reserved words */ 
+				   aliases, negative for reserved words */
     int inuse;			/* alias is being expanded */
 };
 
@@ -986,7 +1042,7 @@ struct schedcmd {
     time_t time;		/* when to run it */
 };
 
-#define MAXAL 20	/* maximum number of aliases expanded at once */
+#define MAXAL 20		/* maximum number of aliases expanded at once */
 
 /* hash table node */
 
@@ -1100,11 +1156,18 @@ struct compctl {		/* the real desc. for compctls */
 #define CC_READONLYS    (1<<21)
 #define CC_SPECIALS	(1<<22)
 #define CC_DELETE       (1<<23)
+#define CC_NAMED        (1<<24)
 
 #define CC_RESERVED     (1<<31)
 
 #include <errno.h>
+#ifdef __NetBSD__
+extern const char *const sys_errlist[];
+
+#else
 extern char *sys_errlist[];
+
+#endif
 extern int sys_nerr;
 extern int errno;
 
@@ -1167,6 +1230,7 @@ struct option {
 #define LOGINSHELL 'l'
 #define MONITOR 'm'
 #define NOEXEC 'n'
+#define KSHPRIV 'p'
 #define SHINSTDIN 's'
 #define NOUNSET 'u'
 #define VERBOSE 'v'
@@ -1202,7 +1266,7 @@ struct option {
 #define APPENDHISTORY '\36'
 #define CSHJUNKIEHISTORY '\037'
 #define MAGICEQUALSUBST '\040'
-#define CSHJUNKIETILDE '\041'
+#define GLOBSUBST '\041'
 #define PROMPTSUBST '\043'
 #define ALWAYSLASTPROMPT '\044'
 #define COMPLETEALIASES '\045'
@@ -1211,7 +1275,7 @@ struct option {
 #define NOFLOWCONTROL '\050'
 #define LISTAMBIGUOUS '\051'
 #define AUTONAMEDIRS '\052'
-
+#define CSHJUNKIEPAREN '\053'
 
 #ifndef GLOBALS
 extern struct option optns[];
@@ -1300,7 +1364,7 @@ struct option optns[] =
     {"appendhistory", APPENDHISTORY},
     {"cshjunkiehistory", CSHJUNKIEHISTORY},
     {"magicequalsubst", MAGICEQUALSUBST},
-    {"cshjunkietilde", CSHJUNKIETILDE},
+    {"globsubst", GLOBSUBST},
     {"promptsubst", PROMPTSUBST},
     {"alwayslastprompt", ALWAYSLASTPROMPT},
     {"completealiases", COMPLETEALIASES},
@@ -1309,6 +1373,7 @@ struct option optns[] =
     {"noflowcontrol", NOFLOWCONTROL},
     {"listambiguous", LISTAMBIGUOUS},
     {"autonamedirs", AUTONAMEDIRS},
+    {"cshjunkieparen", CSHJUNKIEPAREN},
     {NULL, 0}
 };
 
@@ -1387,6 +1452,10 @@ struct option optns[] =
 #undef HAS_FIFOS
 #endif
 
+#if !defined(S_ISFIFO) && !defined(HAS_FIFOS)
+#define S_ISFIFO(m) 0
+#endif
+
 /* buffered shell input for non-interactive shells */
 
 EXTERN FILE *bshin;
@@ -1616,6 +1685,7 @@ EXTERN long baud;		/* $BAUD */
 EXTERN long columns;		/* $COLUMNS */
 EXTERN long lines;		/* $LINES */
 EXTERN long reporttime;		/* $REPORTTIME */
+EXTERN long lastval2;
 
 /* input fd from the coprocess */
 
@@ -1954,6 +2024,7 @@ extern char *tccapnams[TC_COUNT];
 #ifndef POSIX
 typedef unsigned int sigset_t;
 typedef unsigned int mode_t;
+
 #define sigemptyset(s)    (*(s) = 0)
 #if NSIG == 32
 #define sigfillset(s)     ((*(s) = 0xffffffff), 0)
@@ -1961,14 +2032,14 @@ typedef unsigned int mode_t;
 #define sigfillset(s)     ((*(s) = (1 << NSIG) - 1), 0)
 #endif
 #define z_sigmask(n)      (1 << ((n) - 1))
-#define sigaddset(s,n)    ((*(s) |= z_sigmask(n)), 0) 
+#define sigaddset(s,n)    ((*(s) |= z_sigmask(n)), 0)
 #define sigdelset(s,n)    ((*(s) &= ~z_sigmask(n)), 0)
 #define sigismember(s,n)  ((*(s) & z_sigmask(n)) ? 1 : 0)
 #endif
 
 #define blockchld()        sig_block(sig_mask(SIGCHLD))
 #define unblockchld()      sig_unblock(sig_mask(SIGCHLD))
-#define chldpause()        sig_suspend(SIGCHLD)
+#define chldpause(S)       sig_suspend(SIGCHLD, (S))
 
 #ifdef SIGNAL_MASKS
 #define fast_block(s)      (void)(*(s) = sig_block(sig_notmask(0)))
